/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi.c,v 1.97 2002/10/16 20:53:39 smferris Exp $ 
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/config.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/tcp.h>
#include <net/sock.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/timer.h>


#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) )
#  include <linux/init.h>
#  define INIT_MODIFIER __init
#  define EXIT_MODIFIER __exit
#else
#  define INIT_MODIFIER 
#  define EXIT_MODIFIER 
#endif

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
# include <asm/semaphore.h>
#else
# include <asm/spinlock.h>
#endif
#include <asm/uaccess.h>
#include <scsi/sg.h>

#include <sd.h>
#include <scsi.h>
#include <hosts.h>

/* if set, do a TCP Abort when a session drops, instead of (attempting) a graceful TCP Close */
#define TCP_ABORT_ON_DROP 0

#define TEST_ABORTS 0
#define ABORT_FREQUENCY 5000
#define ABORT_COUNT 6

/* requires TEST_ABORTS 1 */
#define TEST_LUN_RESETS 0
#define LUN_RESET_FREQUENCY 2
#define LUN_RESET_COUNT 4

/* requires TEST_LUN_RESETS 1 */
#define TEST_WARM_RESETS 0
#define WARM_RESET_FREQUENCY 2
#define WARM_RESET_COUNT 2

/* requires TEST_WARM_RESETS 1 */
#define TEST_COLD_RESETS 0

/* periodically stall reading data to test data arriving after aborts have started */
#define TEST_DELAYED_DATA 0

/* drop sessions during error recovery */
#define TEST_ERROR_RECOVERY_SESSION_DROP 0

/* periodically fake unit attention sense data to test bugs in Linux */
#define FAKE_DEFERRED_ERRORS 0
#define FAKE_DEFERRED_ERROR_FREQUENCY 100

/* fake sense indicating ILLEGAL_REQUEST for all REPORT_LUNS commands */
#define FAKE_NO_REPORT_LUNS 0

/* fake check conditions on the first 2 attempts for each probe command */
#define FAKE_PROBE_CHECK_CONDITIONS 0

/* fake underflows on the first 4 attempts for each probe command */
#define FAKE_PROBE_UNDERFLOW 0

#define FAKE_PDU_REJECTS 0

/* always fail error recovery in eh_strategy, to test devices going offline */
#define FAIL_ERROR_RECOVERY 0

#define MULTIPATH_SUPPORT 0

#include "iscsi-common.h"
#include "iscsi-protocol.h"
#include "iscsi-ioctl.h"
#include "iscsi-io.h"
#include "iscsi-login.h"
#include "iscsi-trace.h"
#include "iscsi.h"
#include "iscsi-session.h"
#include "iscsi-version.h"
#include "iscsi-probe.h"

/*
 *  IMPORTANT NOTE: to prevent deadlock, when holding multiple locks,
 *  the following locking order must be followed at all times:
 *
 *  hba_list_lock             - access to collection of HBA instances
 *  session->task_lock        - access to a session's collections of tasks
 *  hba->free_task_lock       - for task alloc/free from the HBA's task pool
 *  io_request_lock/host_lock - mid-layer acquires before calling queuecommand, eh_*, 
 *                                we must acquire before done() callback
 *  hba->session_lock         - access to an HBA's collection of sessions   
 *  session->scsi_cmnd_lock   - access to a session's list of Scsi_Cmnds
 *  iscsi_trace_lock          - for the tracing code
 *
 * 
 *  The locking order is somewhat counter-intuitive.  The queue()
 *  function may get called by a bottom-half handler for the SCSI
 *  midlayer, which means it may be called after any interrupt occurs,
 *  while another kernel thread is suspended due to the interrupt.
 *  Since this may be one of our threads which is holding a spinlock,
 *  to prevent deadlocks the spinlocks used by the queue() function must
 *  be last in the locking order.  Also, the bottom-half handler must somehow 
 *  be locally disabled when holding any lock that might be used by queue(), 
 *  to prevent the lock holder being suspended by an interrupt, and then 
 *  the queue() function called (which would deadlock).  While 2.4 kernels
 *  have a spin_lock_bh() function, we don't use it, because spin_unlock_bh()
 *  may immediately run bottom-halves, and the driver sometimes would have
 *  needed to call spin_unlock_bh() while interrupts were off and the 
 *  io_request_lock was already held, which could cause deadlocks.  Instead,
 *  the driver always uses spin_lock_irqsave.
 *
 *  Also, since any interrupt may try to acquire the io_request_lock, we 
 *  want the io_request_lock as late in the lock order as possible, since
 *  interrupts must be disabled when holding any lock that follows the 
 *  io_request_lock in the locking order.  The locks needed in queue()
 *  follow the io_request_lock so that interrupts may call the queue()
 *  entry point.  The eh_*_handlers all release the io_request_lock, since
 *  they all may invoke the scheduler, and that can't be done with a spinlock 
 *  held.  Likewise, since scheduling in an interrupt will panic the kernel,
 *  all of the eh_*_handlers may fail if called from interrupt context.
 *
 *  As of 1-2-2002, various threads may be in the following lock states
 *  (ignoring the trace_lock, since the tracing code is largely unmaintained):
 *
 *  queue: (interrupts off) io_request_lock 
 *         (interrupts off) io_request_lock, hba->session_lock 
 *         (interrupts off) io_request_lock, hba->session_lock, session->scsi_cmnd_lock 
 *
 *  tx: none
 *      (an interrupt acquires) io_request_lock 
 *      hba->free_task_lock,
 *      hba->free_task_lock, (an interrupt acquires) io_request_lock 
 *      session->task_lock, 
 *      session->task_lock, (an interrupt acquires) io_request_lock 
 *      session->task_lock, (interrupts off) session->scsi_cmnd_lock 
 *      (interrupts off) session->scsi_cmnd_lock 
 *
 *  rx: none
 *      (an interrupt acquires) io_request_lock 
 *      session->task_lock
 *      session->task_lock, (an interrupt acquires) io_request_lock 
 *      session->task_lock, (interrupts off) io_request_lock 
 *      hba->free_task_lock
 *      hba->free_task_lock, (an interrupt acquires) io_request_lock 
 *      (interrupts off) session->scsi_cmnd_lock 
 *      session->task_lock, (interrupts off) session->scsi_cmnd_lock 
 *
 *  timer: none
 *         hba_list_lock
 *         hba_list_lock, (an interrupt acquires) io_request_lock
 *         hba_list_lock, (interrupts off) hba->session_lock 
 *
 *  ioctl: none
 *         (an interrupt acquires) io_request_lock 
 *         hba_list_lock
 *         hba_list_lock, (an interrupt acquires) io_request_lock
 *         (interrupts off) hba->session_lock 
 *         session->task_lock
 *         session->task_lock, (an interrupt acquires) io_request_lock
 *         session->task_lock, (interrupts off) session->scsi_cmnd_lock
 *         session->task_lock, (interrupts off) io_request_lock 
 *        (interrupts off) session->scsi_cmnd_lock 
 *         
 *  eh_*_handler: (interrupts off) io_request_lock
 *                none
 *                (an interrupt acquires) io_request_lock
 *                (interrupts off) session->scsi_cmnd_lock
 *                session->task_lock
 *                session->task_lock, (an interrupt acquires) io_request_lock
 *
 *  This driver assumes the eh_*_handler functions can safely release
 *  the io_request_lock and locally enable interrupts, which is true
 *  on 2.4 kernels, but unclear on 2.2 kernels.
 *
 *  The eh_*_handler functions may fail if called from interrupt context,
 *  since they typically need to block and wait for a response from the
 *  target, and scheduling in interrupt context would panic the kernel.
 *
 *  The driver assumes that calling the following kernel primitives may invoke the
 *  scheduler and preempt the caller, and thus no spinlocks can be held when they
 *  are called, nor can interrupts or bottom-half handlers be disabled:
 *
 *  sock_sendmsg
 *  sock_recvmsg
 *  kmalloc
 *  schedule_timeout  (duh)
 *  kernel_thread
 *  waitpid
 *
 *  The following kernel primitives probably don't schedule, but the driver
 *  could handle it even if they did:
 *
 *  signal_pending
 *  get_ds
 *  get_fs
 *  set_fs
 *  fget
 *  fput
 *  
 *  The driver assumes that calling the following kernel primitives WILL NOT invoke the
 *  scheduler, and thus cannot cause a preemption.  If this assumption is violated,
 *  the driver will break badly:
 *
 *  wake_up
 *  kill_proc
 *  printk
 *  kfree
 * 
 *  The following driver functions may invoke the scheduler, and must not be
 *  called while holding any spinlock:
 *
 *  iscsi_sendmsg
 *  iscsi_recvmsg
 *  alloc_task
 *  cold_reset_target
 *  warm_reset_target
 */

MODULE_AUTHOR("Cisco Systems, Inc.");
MODULE_DESCRIPTION("iSCSI driver");
MODULE_LICENSE("GPL");

/* Force tagged command queueing for all devices, regardless of whether they say they support it */
static int force_tcq = 0;
MODULE_PARM(force_tcq, "i");
MODULE_PARM_DESC(force_tcq, "when non-zero, force tagged command queueing for all devices");

/* Queue depth for devices that don't support tagged command queueing.
 * The driver used to use ISCSI_CMDS_PER_LUN, which was probably a bug.
 * Default to 1 now, but let people who want to the old behavior set it higher.
 */
static int untagged_queue_depth = 1;
MODULE_PARM(untagged_queue_depth, "i");
MODULE_PARM_DESC(untagged_queue_depth, "queue depth to use for devices that don't support tagged command queueing");

static int translate_deferred_sense = 1;
MODULE_PARM(translate_deferred_sense, "i");
MODULE_PARM_DESC(translate_deferred_sense, "translate deferred sense data to current sense data in command responses");

#ifndef UINT32_MAX
# define UINT32_MAX 0xFFFFFFFFU
#endif


static int ctl_open(struct inode *inode, struct file *file);
static int ctl_close(struct inode *inode, struct file *file);
static int ctl_ioctl(struct inode *inode,
                      struct file *file,
                      unsigned int cmd,
                      unsigned long arg );

static int control_major;
static const char *control_name = "iscsictl";

static struct file_operations control_fops = {
    owner: THIS_MODULE,
    ioctl: ctl_ioctl,    /* ioctl */
    open: ctl_open,      /* open */
    release: ctl_close,  /* release */
};

spinlock_t iscsi_hba_list_lock = SPIN_LOCK_UNLOCKED;
static iscsi_hba_t *iscsi_hba_list = NULL;

static volatile unsigned long init_module_complete = 0;
static volatile unsigned long iscsi_timer_running = 0;
static volatile pid_t iscsi_timer_pid = 0;

volatile unsigned int iscsi_log_settings = LOG_SET(ISCSI_LOG_ERR);

/* use the otherwise unused Scsi_Pointer Status to record 
 * the stage of error recovery each command is in.
 */
#  define CMND_ERROR_STAGE(c) (c)->SCp.Status

#  define ISCSI_ERROR_STAGE_UNKNOWN     0
#  define ISCSI_ERROR_STAGE_ABORT       1
#  define ISCSI_ERROR_STAGE_LUN_RESET   2
#  define ISCSI_ERROR_STAGE_WARM_RESET  3
#  define ISCSI_ERROR_STAGE_COLD_RESET  4
#  define ISCSI_ERROR_STAGE_OFFLINE     99
#  define ISCSI_ERROR_STAGE_RECOVERED   100


#define is_digit(c)	(((c) >= '0') && ((c) <= '9'))
#define is_hex_lower(c) (((c) >= 'a') && ((c) <= 'f'))
#define is_hex_upper(c) (((c) >= 'A') && ((c) <= 'F'))
#define is_space(c)	((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\0')

#if DEBUG_TRACE
spinlock_t iscsi_trace_lock = SPIN_LOCK_UNLOCKED;
static iscsi_trace_entry_t trace_table[ISCSI_TRACE_COUNT];
static int trace_index=0;

# define ISCSI_TRACE(P_TYPE, P_CMND, P_TASK, P_DATA1, P_DATA2) \
           iscsi_fill_trace((P_TYPE), (P_CMND), (P_TASK), (P_DATA1), (P_DATA2))
#else
# define ISCSI_TRACE(P_TYPE, P_CMND, P_TASK, P_DATA1, P_DATA2) 
#endif

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,27) )
/* note change modeled per linux2.4 drivers/scsi/ips.c */
struct proc_dir_entry proc_dir_iscsi = {
# ifdef PROC_SCSI_ISCSI
    PROC_SCSI_ISCSI,
# else
    PROC_SCSI_NOT_PRESENT,
# endif
    5,
    "iscsi",
    S_IFDIR|S_IRUGO|S_IXUGO,
    2
};
#endif

/* become a daemon kernel thread.  Some kernels provide this functionality
 * already, and some even do it correctly
 */
void iscsi_daemonize(void)
{
    struct task_struct *this_task = current;
    
# if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10) )
    /* use the kernel's daemonize */
    daemonize();

    /* Reparent to init */
    reparent_to_init();

    /* increase priority like the md driver does for it's kernel threads */
    this_task->policy = SCHED_OTHER;
#  ifdef set_user_nice
    set_user_nice(this_task, -20);
#  else
    this_task->nice = -20;
#  endif
    wmb();

# elif ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) )
    /* use the kernel's daemonize */
    daemonize();

    /* We'd like to reparent to init, but don't have a function to do it, and
     * symbols like child_reaper aren't exported to modules
     */ 
    
    /* increase priority like the md driver does for it's kernel threads */
    this_task->policy = SCHED_OTHER;
    this_task->nice = -20;
    wmb();
    
# else
    /* 2.2.18 and later has daemonize(), but it's not always correct, so we do it ourselves. */
    struct fs_struct *fs;

    lock_kernel();
    
    /*
     * If we were started as result of loading a module, close all of the
     * user space pages.  We don't need them, and if we didn't close them
     * they would be locked into memory.
     */
    exit_mm(this_task);
    
    this_task->session = 1;
    this_task->pgrp = 1;
    this_task->tty = NULL;
    
    /* Become as one with the init task */
    exit_files(this_task);
    this_task->files = init_task.files;
    atomic_inc(&this_task->files->count);
    
    exit_fs(this_task);       /* this_task->fs->count--; */
    fs = init_task.fs;
    this_task->fs = fs;
    atomic_inc(&fs->count);                                                 

    /* We'd like to reparent to init, but don't have a function to do it, and
     * symbols like child_reaper aren't exported to modules.
     */ 

    /* increase priority like the md driver does for it's kernel threads */
    this_task->policy = SCHED_OTHER;
    this_task->priority = 40;
    wmb();
    
    unlock_kernel();
# endif
}

/* wake up the tx_thread without ever losing the wakeup event */
static void wake_tx_thread(int control_bit, iscsi_session_t *session)
{
    /* tell the tx thread what to do when it wakes up. */
    set_bit(control_bit, &session->control_bits);
    
    /* We make a condition variable out of a wait queue and atomic test&clear.
     * May get spurious wake-ups, but no wakeups will be lost.
     * this is cv_signal().  wait_event_interruptible is cv_wait().
     */
    set_bit(TX_WAKE, &session->control_bits);
    wake_up(&session->tx_wait_q);
}

/* drop an iscsi session */
static void iscsi_drop_session(iscsi_session_t *session)
{
    pid_t pid;
              
    DEBUG_INIT4("iSCSI: iscsi_drop_session %p, rx %d, tx %d at %lu\n", 
                session, session->rx_pid, session->tx_pid, jiffies);

    clear_bit(SESSION_ESTABLISHED, &session->control_bits);
    set_bit(SESSION_DROPPED, &session->control_bits);  /* so we know whether to abort the connection */

    if ((pid = session->tx_pid))
        kill_proc(pid, SIGHUP, 1);
    if ((pid = session->rx_pid))
        kill_proc(pid, SIGHUP, 1);
}

/* caller must hold session->task_lock */
static void iscsi_request_logout(iscsi_session_t *session, int logout, int logout_response)
{
    if (atomic_read(&session->num_active_tasks) == 0) {
        DEBUG_INIT2("iSCSI: session %p currently has no active tasks, queueing logout at %lu\n", 
                    session, jiffies);
        session->logout_response_deadline = jiffies + (logout_response * HZ);
        if (session->logout_response_deadline == 0)
            session->logout_response_deadline = 1;
        wmb();
        set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
        wake_tx_thread(TX_LOGOUT, session);
    }
    else {
        session->logout_deadline = jiffies + (logout * HZ);
        if (session->logout_deadline == 0)
            session->logout_deadline = 1;
        session->logout_response_deadline = session->logout_deadline + (logout_response * HZ);
        if (session->logout_response_deadline == 0)
            session->logout_response_deadline = 1;
        wmb();
        set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
    }
}

static void iscsi_terminate_session(iscsi_session_t *session)
{
    pid_t pid;
              
    if ((test_and_set_bit(SESSION_TERMINATING, &session->control_bits) == 0) && 
        test_bit(SESSION_ESTABLISHED, &session->control_bits))
    {
        DEBUG_INIT2("iSCSI: iscsi_terminate_session %p, requesting logout at %lu\n",
                    session, jiffies);

        /* on the first terminate request while the session is up, request a logout in the next 3 seconds */
        spin_lock(&session->task_lock);
        iscsi_request_logout(session, 3, session->active_timeout);
        spin_unlock(&session->task_lock);
    }
    else {
        /* either we've already tried to terminate once, or the the session is down.  just kill everything. */
        clear_bit(SESSION_ESTABLISHED, &session->control_bits);

        DEBUG_INIT4("iSCSI: iscsi_terminate_session %p, killing rx %d, tx %d at %lu\n", 
                    session, session->rx_pid, session->tx_pid, jiffies);
        
        /* kill the session's threads */
        if ((pid = session->tx_pid))
            kill_proc(pid, SIGKILL, 1);
        if ((pid = session->rx_pid))
            kill_proc(pid, SIGKILL, 1);
    }
}

/* if a signal is pending, deal with it, and return 1.
 * Otherwise, return 0.
 */
static int iscsi_handle_signals(iscsi_session_t *session)
{
    pid_t pid;
    int ret = 0;
    
    /* if we got SIGHUP, try to establish a replacement session.
     * if we got SIGKILL, terminate this session.
     */
    if (signal_pending(current)) {
        spin_lock_irq(&current->sig->siglock);

        /* iscsi_drop_session and iscsi_terminate_session signal both
         * threads, but someone logged in as root may not.  So, we
         * make sure whichever process gets signalled first propagates
         * the signal when it looks like only one thread got
         * signalled.
         */

        /* on SIGKILL, terminate the session */
        if (SIGNAL_IS_PENDING(SIGKILL)) {
            if (!test_and_set_bit(SESSION_TERMINATING, &session->control_bits)) {
                if ((pid = session->tx_pid) && (pid != current->pid)) {
                    printk("iSCSI: rx thread %d received SIGKILL, killing tx thread %d\n", current->pid, pid);
                    kill_proc(pid, SIGKILL, 1);
                }
                if ((pid = session->rx_pid) && (pid != current->pid)) {
                    printk("iSCSI: tx thread %d received SIGKILL, killing rx thread %d\n", current->pid, pid);
                    kill_proc(pid, SIGKILL, 1);
                }
            }
            ret = 1;
        }
        /* on SIGHUP, drop the session, and try to establish a replacement session */
        if (SIGNAL_IS_PENDING(SIGHUP)) {
            if (test_and_clear_bit(SESSION_ESTABLISHED, &session->control_bits)) {
                if ((pid = session->tx_pid) && (pid != current->pid)) {
                    printk("iSCSI: rx thread %d received SIGHUP, signaling tx thread %d\n", current->pid, pid);
                    kill_proc(pid, SIGHUP, 1);
                }
                if ((pid = session->rx_pid) && (pid != current->pid)) {
                    printk("iSCSI: tx thread %d received SIGHUP, signaling rx thread %d\n", current->pid, pid);
                    kill_proc(pid, SIGHUP, 1);
                }
            }
            ret = 1;
        }
        /* we don't care about any other signals */
        flush_signals(current);
        spin_unlock_irq(&current->sig->siglock);
    }

    return ret;
}


unsigned int iscsi_expected_data_length(Scsi_Cmnd *sc)
{
    unsigned int length = 0;

    switch (sc->cmnd[0]) {
        case INQUIRY:
        case REQUEST_SENSE:
            length = sc->cmnd[4];
            return length;
        case REPORT_LUNS:
            length |= sc->cmnd[6] << 24;
            length |= sc->cmnd[7] << 16;
            length |= sc->cmnd[8] << 8;
            length |= sc->cmnd[9];
            return length;
        default:
            return sc->request_bufflen;
    }
}


/* compare against 2^31 */
#define SNA32_CHECK 2147483648UL

/* Serial Number Arithmetic, 32 bits, less than, RFC1982 */
static int sna_lt(uint32_t n1, uint32_t n2)
{
    return ((n1 != n2) && 
            (((n1 < n2) && ((n2 - n1) < SNA32_CHECK)) || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

/* Serial Number Arithmetic, 32 bits, less than, RFC1982 */
static int sna_lte(uint32_t n1, uint32_t n2)
{
    return ((n1 == n2) ||
            (((n1 < n2) && ((n2 - n1) < SNA32_CHECK)) || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

/* difference isn't really a defined operation in SNA, but we'd like it so that
 * we can determine how many commands can be queued to a session.
 */
static int cmdsn_window_size(uint32_t expected, uint32_t max)
{
    if ((expected <= max) && ((max - expected) < SNA32_CHECK)) {
        return (max - expected + 1);
    }
    else if ((expected > max) && ((expected - max) < SNA32_CHECK)) {
        /* window wraps around */
        return ((UINT32_MAX - expected) + 1 + max + 1);
    }
    else {
        /* window closed, or numbers bogus */
        return 0;
    }
}

/* remember old peak cmdsn window size, and report the largest */
static int max_tasks_for_session(iscsi_session_t *session)
{
    if (session->ExpCmdSn == session->MaxCmdSn + 1)
        /* if the window is closed, report nothing, regardless of what it was in the past */
        return 0;
    else if (session->last_peak_window_size < session->current_peak_window_size)
        /* window increasing, so report the current peak size */
        return MIN(session->current_peak_window_size, ISCSI_CMDS_PER_LUN * session->num_luns);
    else
        /* window decreasing.  report the previous peak size, in case it's
         * a temporary decrease caused by the commands we're sending.
         * we want to keep the right number of commands queued in the driver,
         * ready to go as soon as they can.
         */
        return MIN(session->last_peak_window_size, ISCSI_CMDS_PER_LUN * session->num_luns);
}

/* possibly update the ExpCmdSN and MaxCmdSN, and peak window sizes */
static void updateSN(iscsi_session_t *session, UINT32 expcmdsn, UINT32 maxcmdsn) 
{
    int window_size;

    /* standard specifies this check for when to update expected and max sequence numbers */
    if (!sna_lt(maxcmdsn, expcmdsn - 1)) {
        if ((expcmdsn != session->ExpCmdSn) && !sna_lt(expcmdsn, session->ExpCmdSn)) {
            session->ExpCmdSn = expcmdsn;
        }
        if ((maxcmdsn != session->MaxCmdSn) && !sna_lt(maxcmdsn, session->MaxCmdSn)) {
            
            session->MaxCmdSn = maxcmdsn;

            /* look for the peak window size */
            window_size = cmdsn_window_size(expcmdsn, maxcmdsn);
            if (window_size > session->current_peak_window_size)
                session->current_peak_window_size = window_size;

            /* age peak window size info */
            if (time_before(session->window_peak_check + (15 * HZ), jiffies)) {
                session->last_peak_window_size = session->current_peak_window_size;
                session->current_peak_window_size = window_size;
                session->window_peak_check = jiffies;
            }

            /* memory barrier for all of that */
            mb();

            /* wake the tx thread to try sending more commands */
            wake_tx_thread(TX_SCSI_COMMAND, session);
        }

        /* record whether or not the command window for this session has closed,
         * so that we can ping the target periodically to ensure we eventually
         * find out that the window has re-opened.  
         */
        if (maxcmdsn == expcmdsn - 1) {
            session->current_peak_window_size = 0;
            set_bit(SESSION_WINDOW_CLOSED, &session->control_bits);
        }
        else
            clear_bit(SESSION_WINDOW_CLOSED, &session->control_bits);

        DEBUG_FLOW3("iSCSI: session %p - ExpCmdSN %u, MaxCmdSN %u\n", 
                    session, session->ExpCmdSn, session->MaxCmdSn);
    }
}

/* add a session from an HBA's collection of sessions.
 * caller must hold the HBA's session lock.
 */
static int add_session(iscsi_hba_t *hba, iscsi_session_t *session)
{
    iscsi_session_t *prior, *next;

    if (!hba)
        return 0;

    if (test_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags)) {
        printk("iSCSI: HBA %p is shutting down, can't add session %p\n", hba, session);
        return 0;
    }
    
    prior = NULL;
    next = hba->session_list_head;
    while (next && (next->channel < session->channel)) {
        prior = next;
        next = prior->next;
    }
    while (next && (next->channel == session->channel) && (next->target_id < session->target_id)) {
        prior = next;
        next = prior->next;
    }
    
    /* same Linux SCSI address? */
    if (next && (next->channel == session->channel) && (next->target_id == session->target_id)) {
        if (strcmp(next->TargetName, session->TargetName) == 0) {
            /* already have a session running to this target */
            printk("iSCSI: session to %s already exists\n", session->TargetName);
        }
        else {
            printk("iSCSI: error - TargetName %s cannot claim bus %d id %d, already in use by %s\n", 
                   session->TargetName, session->iscsi_bus, next->target_id, next->TargetName);
        }
        return 0;
    }
    else {
        /* insert the session into the list */
        if ((session->next = next))
            next->prev = session;
        else
            hba->session_list_tail = session;

        if ((session->prev = prior))
            prior->next = session;
        else
            hba->session_list_head = session;

        session->hba = hba;
        wmb();

        DEBUG_INIT2("iSCSI: added session %p to hba %p\n", session, hba);
        
        return 1;
    }
}


/* remove a session from an HBA's collection of sessions.
 * caller must hold the HBA's session lock.
 */
static int remove_session(iscsi_hba_t *hba, iscsi_session_t *session)
{
    if (session->hba && (hba != session->hba)) {
        printk("iSCSI: tried to remove session %p from hba %p, but session->hba is %p\n",
               session, hba, session->hba);
        return 0;
    }

    /* remove the session from the HBA */
    if (session == hba->session_list_head) {
        if ((hba->session_list_head = session->next))
            hba->session_list_head->prev = NULL;
        else
            hba->session_list_tail = NULL;
    }
    else if (session == hba->session_list_tail) {
        hba->session_list_tail = session->prev;
        hba->session_list_tail->next = NULL;
    }
    else {
        /* we should always be in the middle, 
         * but check pointers to make sure we don't crash the kernel 
         * if the function is called for a session not on the hba.
         */
        if (session->next && session->prev) {
            session->next->prev = session->prev;
            session->prev->next = session->next;
        }
        else {
            DEBUG_ERR2("iSCSI: failed to remove session %p from hba %p\n",
                       session, hba);
            return 0;
        }
    }
    session->prev = NULL;
    session->next = NULL;
    
    return 1;
}

static iscsi_session_t *find_session_for_cmnd(Scsi_Cmnd *sc) 
{
    iscsi_session_t *session = NULL;
    iscsi_hba_t *hba;
    DECLARE_NOQUEUE_FLAGS;

    if (!sc->host)
        return NULL;

    if (!sc->host->hostdata)
        return NULL;

    hba = (iscsi_hba_t *)sc->host->hostdata;

    /* find the session for this command */
    SPIN_LOCK_NOQUEUE(&hba->session_lock);
    session = hba->session_list_head;
    while (session && (session->channel != sc->channel || session->target_id != sc->target))
        session = session->next;
    if (session)
        atomic_inc(&session->refcount); /* caller must use drop_reference when it's done with the session */
    SPIN_UNLOCK_NOQUEUE(&hba->session_lock);

    return session;
}

static iscsi_session_t *find_session_by_channel(unsigned int host_no, unsigned int channel, unsigned int target_id)
{
    iscsi_session_t *session = NULL;
    iscsi_hba_t *hba;
    DECLARE_NOQUEUE_FLAGS;

    spin_lock(&iscsi_hba_list_lock);

    hba = iscsi_hba_list;
    while (hba && (hba->host_no != host_no)) {
        hba = hba->next;
    }

    /* find the session for this command */
    if (hba) {
        SPIN_LOCK_NOQUEUE(&hba->session_lock);
        session = hba->session_list_head;
        while (session && (session->channel != channel || session->target_id != target_id))
            session = session->next;
        if (session)
            atomic_inc(&session->refcount); /* caller must use drop_reference when it's done with the session */
        SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
    }

    spin_unlock(&iscsi_hba_list_lock);
        
    return session;
}

static iscsi_session_t *find_session_by_bus(int iscsi_bus, int target_id)
{
    iscsi_session_t *session = NULL;
    iscsi_hba_t *hba;
    unsigned int hba_index;
    unsigned int channel;
    DECLARE_NOQUEUE_FLAGS;

    /* compute the appropriate HBA and channel numbers */
    hba_index = iscsi_bus / ISCSI_MAX_CHANNELS_PER_HBA;
    channel = iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;

    spin_lock(&iscsi_hba_list_lock);

    hba = iscsi_hba_list;
    while (hba && (hba_index-- > 0)) {
        hba = hba->next;
    }

    /* find the session for this command */
    if (hba) {
        SPIN_LOCK_NOQUEUE(&hba->session_lock);
        session = hba->session_list_head;
        while (session && (session->channel != channel || session->target_id != target_id))
            session = session->next;
        if (session)
            atomic_inc(&session->refcount); /* caller must use drop_reference when it's done with the session */
        SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
    }

    spin_unlock(&iscsi_hba_list_lock);

    return session;
}

static void delete_session(iscsi_session_t *session)
{
    DEBUG_INIT1("iSCSI: terminated and deleted session %p\n", session);
    if (session->auth_client) {
        memset(session->auth_client, 0, sizeof(*session->auth_client));
        kfree(session->auth_client);
        session->auth_client = NULL;
    }
    if (session->portals) {
        kfree(session->portals);
        session->portals = NULL;
    }
    memset(session, 0, sizeof(*session));
    kfree(session);
}

/* decrement the session refcount, and remove it and free it if the refcount hit zero */
static void drop_reference(iscsi_session_t *session)
{
    iscsi_hba_t *hba;
    DECLARE_NOQUEUE_FLAGS;

    if (!session) {
        printk("iSCSI: bug - drop_reference(NULL)\n");
        return;
    }

    if ((hba = session->hba)) {
        /* may need to remove it from the HBA's session list */
        SPIN_LOCK_NOQUEUE(&hba->session_lock);
        if (atomic_dec_and_test(&session->refcount)) {
            if (remove_session(hba, session)) {
                delete_session(session);
            }
            else {
                printk("iSCSI: bug - failed to remove unreferenced session %p\n", session);
            }
        }
        SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
    }
    else {
        /* session isn't in an HBA's list at the moment, so just check
         * the refcount, and possibly free it.  
         */
        if (atomic_dec_and_test(&session->refcount)) {
            delete_session(session);
        }
    }
}


/* must hold the task_lock to call this */
static iscsi_task_t *find_task(iscsi_task_collection_t *collection, uint32_t itt)
{
    iscsi_task_t *task = collection->head;
    
    while (task) {
        if (task->itt == itt) {
            DEBUG_FLOW3("iSCSI: found itt %u, task %p, refcount %d\n", itt, task, atomic_read(&task->refcount));
            return task;
        }
        task = task->next;
    }

    return NULL;
}

/* don't actually use this at the moment */
/* must hold the task_lock to call this */
static iscsi_task_t *find_mgmt_task(iscsi_task_collection_t *collection, uint32_t mgmt_itt)
{
    iscsi_task_t *task = collection->head;
    
    while (task) {
        if (task->mgmt_itt == mgmt_itt) {
            DEBUG_FLOW2("iSCSI: found mgmt_itt %u, task %p\n", mgmt_itt, task);
            return task;
        }
        task = task->next;
    }

    return NULL;
}

#if 0
/* don't actually need this at the moment */
/* must hold the task_lock to call this */
static iscsi_task_t *find_task_for_cmnd(iscsi_task_collection_t *collection, Scsi_Cmnd *sc)
{
    iscsi_task_t *task = collection->head;
    
    while (task) {
        if (task->scsi_cmnd == sc) {
            DEBUG_FLOW3("iSCSI: found itt %u, task %p for cmnd %p\n", task->itt, task, sc);
            return task;
        }
        task = task->next;
    }

    return NULL;
}
#endif

/* add a task to the collection.  Must hold the task_lock to do this. */
static void add_task(iscsi_task_collection_t *collection, iscsi_task_t *task)
{
    if (task->prev || task->next)
        printk("iSCSI: bug - adding task %p, prev %p, next %p, to collection %p\n",
               task, task->prev, task->next, collection);

    if (collection->head) {
        task->next = NULL;
        task->prev = collection->tail;
        collection->tail->next = task;
        collection->tail = task;
    }
    else {
        task->prev = task->next = NULL;
        collection->head = collection->tail = task;
    }
}

#define first_iscsi_task(collection_ptr) ((collection_ptr)->head)
#define next_iscsi_task(collection_ptr, task_ptr) ((task_ptr)->next)
#define order_next_task(collection_ptr, task_ptr) ((task_ptr)->order_next)

/* must hold the task_lock when calling this */
static iscsi_task_t *pop_task(iscsi_task_collection_t *collection)
{
    iscsi_task_t *task = NULL;

    if ((task = collection->head)) {
        /* pop the head */
        if ((collection->head = task->next))
            collection->head->prev = NULL;
        else
            collection->tail = NULL;

        /* and return it */
        task->prev = NULL;
        task->next = NULL;
        
        return task;
    }

    return NULL;
}

/* must hold the task_lock when calling this */
static void push_task(iscsi_task_collection_t *collection, iscsi_task_t *task)
{
    if (task) {
        task->prev = NULL;
        task->next = collection->head;
        if (collection->head) {
            collection->head->prev = task;
            collection->head = task;
        }
        else {
            collection->head = collection->tail = task;
        }
    }
}

static void unlink_task(iscsi_task_collection_t *collection, iscsi_task_t *task)
{
    /* unlink the task from the collection */
    if (task == collection->head) {
        if ((collection->head = task->next))
            collection->head->prev = NULL;
        else
            collection->tail = NULL;
    }
    else if (task == collection->tail) {
        collection->tail = task->prev;
        collection->tail->next = NULL;
    }
    else {
        task->next->prev = task->prev;
        task->prev->next = task->next;
    }
    task->next = NULL;
    task->prev = NULL;
}

/* if the task for the itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
static iscsi_task_t *remove_task(iscsi_task_collection_t *collection, uint32_t itt)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *search = collection->head;

    while (search) {
        if (search->itt == itt) {
            task = search;
            unlink_task(collection, task);
            return task;
        }
        search = search->next;
    }

    return NULL;
}

/* if the task for the mgmt_itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
static iscsi_task_t *remove_mgmt_task(iscsi_task_collection_t *collection, uint32_t mgmt_itt)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *search = collection->head;

    while (search) {
        if (search->mgmt_itt == mgmt_itt) {
            task = search;
            unlink_task(collection, task);
            return task;
        }
        search = search->next;
    }

    return NULL;
}

/* if the task for the itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
static iscsi_task_t *remove_task_for_cmnd(iscsi_task_collection_t *collection, Scsi_Cmnd *sc)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *search = collection->head;

    while (search) {
        if (search->scsi_cmnd == sc) {
            task = search;
            unlink_task(collection, task);
            return task;
        }
        search = search->next;
    }

    return NULL;
}

/* 
 * remove all tasks with the specified LUN.  Must hold the task_lock to call this.
 */
static void remove_tasks_for_lun(iscsi_task_collection_t *collection, int lun)
{
    iscsi_task_t *search = collection->head;
    iscsi_task_t *next = NULL;

    while (search) {
        next = search->next;
        if (search->scsi_cmnd && search->scsi_cmnd->lun == lun)
            unlink_task(collection, search);
        
        search = next;
    }
}


/* must be called with no locks held, since it may sleep, and acquires
 * locks on it's own.
 */
static iscsi_task_t *alloc_task(iscsi_session_t *session)
{
    iscsi_task_t *task = NULL;
    iscsi_hba_t *hba = session->hba;

    if (!hba) {
        printk("iSCSI: alloc_task - session %p has NULL HBA\n", session);
        return NULL;
    }

    /* try to get one from the HBA's free task collection */
    spin_lock(&hba->free_task_lock);
    if ((task = pop_task(&hba->free_tasks))) {
        atomic_dec(&hba->num_free_tasks);
        atomic_inc(&hba->num_used_tasks);
        hba->min_free_tasks = MIN(hba->min_free_tasks, atomic_read(&hba->num_free_tasks));
    }
    else {
        hba->min_free_tasks = 0;
    }
    spin_unlock(&hba->free_task_lock);

    /* otherwise, try to dynamically allocate a task */
    if (!task) {
        if ((task = kmalloc(sizeof(iscsi_task_t), GFP_ATOMIC))) {
            atomic_inc(&hba->num_used_tasks);
            DEBUG_ALLOC6("iSCSI: kmalloc task %p (active %u, used %u, free %u) for session %p to %s\n",
                         task, atomic_read(&session->num_active_tasks), 
                         atomic_read(&hba->num_used_tasks), atomic_read(&hba->num_free_tasks),
                         session, session->log_name);
        }
    }

    if (task) {
        memset(task, 0, sizeof(iscsi_task_t) );
        task->itt = RSVD_TASK_TAG;
        task->ttt = RSVD_TASK_TAG;
        task->mgmt_itt = RSVD_TASK_TAG;
        task->next = task->prev = NULL;
        task->order_next = task->order_prev = NULL;
        task->session = session;
        init_waitqueue_head(&task->wait_q);
        wmb();
    }
    else {
        set_bit(SESSION_TASK_ALLOC_FAILED, &session->control_bits);
    }
    return task;
}

/* must be called with the HBA's free_task_lock held */
static void free_task(iscsi_task_t *task)
{
    iscsi_session_t *session = task->session;
    iscsi_hba_t *hba;
    
    if (! task) {
        DEBUG_ERR0("iSCSI: free_task couldn't free NULL task\n");
        return;
    }
    if (! session) {
        DEBUG_ERR1("iSCSI: free_task couldn't find session for task %p\n", task);
        return;
    }
    hba = session->hba;
    if (!hba) {
        DEBUG_ERR1("iSCSI: free_task couldn't find HBA for task %p\n", task);
        return;
    }

    if (task->next || task->prev || task->order_next || task->order_prev) {
        /* this is a memory leak, which is better than memory corruption */
        printk("iSCSI: bug - tried to free task %p with prev %p, next %p, order_prev %p, order_next %p\n",
               task, task->prev, task->next, task->order_prev, task->order_next);
        return;
    }

    DEBUG_QUEUE4("iSCSI: free_task %p, itt %u, session %p, %u currently free\n", 
                 task, task->itt, task->session, atomic_read(&hba->num_free_tasks));

    /* zero out the task settings */
    task->scsi_cmnd = NULL;
    task->session = NULL;
    task->itt = RSVD_TASK_TAG;
    task->mgmt_itt = RSVD_TASK_TAG;
    task->next = task->prev = NULL;
    task->order_next = task->order_prev = NULL;
    atomic_set(&task->refcount, 0);
    
    /* put the task on the session's free list */
    atomic_inc(&hba->num_free_tasks);
    atomic_dec(&hba->num_used_tasks);
    add_task(&hba->free_tasks, task);

    /* If an alloc call has failed, we need to wake up the TX thread
     * now that a task is available, since there are no guarantees
     * that anything else will wake it up.
     */
    if (test_and_clear_bit(SESSION_TASK_ALLOC_FAILED, &session->control_bits))
        wake_tx_thread(TX_SCSI_COMMAND, session);
}


/* As long as the tx thread is the only caller, no locking
 * is required.  If any other thread also needs to call this,
 * then all callers must be changed to agree on some locking
 * protocol.  Currently, some but not all caller's are holding
 * the session->task_lock.
 */
static inline uint32_t allocate_itt(iscsi_session_t *session)
{
    uint32_t itt = 0;

    if (session) {
        itt = session->itt++;
        /* iSCSI reserves 0xFFFFFFFF, this driver reserves 0 */
        if (session->itt == RSVD_TASK_TAG)
            session->itt = 1;
    }
    return itt;
}


/* Caller must hold the session's task_lock.  Associating a task with
 * a session causes it to be completed on a session drop or target
 * reset, along with all other session tasks, in the order they were
 * added to the session.  Preserving the ordering is required by the
 * Linux SCSI architecture.  Tasks that should not be completed to the
 * Linux SCSI layer (because the eh_abort_handler has or will return
 * SUCCESS for it) get removed from the session, though they may still
 * be in various task collections so that PDUs relating to them can be
 * sent or received.
 */
static void add_session_task(iscsi_session_t *session, iscsi_task_t *task)
{
    if (atomic_read(&session->num_active_tasks) == 0) {
        /* session going from idle to active, pretend we just
         * received something, so that the idle period before this doesn't
         * cause an immediate timeout.
         */
        session->last_rx = jiffies;
        wmb();
    }
    atomic_inc(&session->num_active_tasks);

    /* set task info */
    task->session = session;
    task->itt = allocate_itt(session);

    DEBUG_QUEUE5("iSCSI: task %p allocated itt %u for command %p, session %p to %s\n", 
                 task, task->itt, task->scsi_cmnd, session, session->log_name);

    /* add it to the session task ordering list */
    if (session->arrival_order.head) {
        task->order_prev = session->arrival_order.tail;
        task->order_next = NULL;
        session->arrival_order.tail->order_next = task;
        session->arrival_order.tail = task;
    }
    else {
        task->order_prev = NULL;
        task->order_next = NULL;
        session->arrival_order.head = session->arrival_order.tail = task;
    }

    DEBUG_FLOW4("iSCSI: task %p, itt %u, added to session %p to %s\n", task, task->itt, session, session->log_name);
}

static int remove_session_task(iscsi_session_t *session, iscsi_task_t *task)
{
    /* remove the task from the session's arrival_order collection */
    if (task == session->arrival_order.head) {
        if ((session->arrival_order.head = task->order_next))
            session->arrival_order.head->order_prev = NULL;
        else
            session->arrival_order.tail = NULL;
    }
    else if (task == session->arrival_order.tail) {
        session->arrival_order.tail = task->order_prev;
        session->arrival_order.tail->order_next = NULL;
    }
    else {
        /* we should always be in the middle, 
         * but check pointers to make sure we don't crash the kernel 
         * if the function is called for a task not in the session.
         */
        if (task->order_next && task->order_prev) {
            task->order_next->order_prev = task->order_prev;
            task->order_prev->order_next = task->order_next;
        }
        else {
            DEBUG_ERR4("iSCSI: failed to remove itt %u, task %p from session %p to %s\n", 
                       task->itt, task, session, session->log_name);
            return 0;
        }
    }
    task->order_prev = NULL;
    task->order_next = NULL;

    if (atomic_dec_and_test(&session->num_active_tasks)) {
        /* zero active tasks, ready to logout */
        if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
            DEBUG_INIT2("iSCSI: session %p now has no active tasks, queueing logout at %lu\n", session, jiffies);
            wake_tx_thread(TX_LOGOUT, session);
        }
    }

    return 1;
}

/* 
 * move all tasks in the session for the specified LUN into the collection.
 */
static void move_session_tasks_for_lun(iscsi_task_collection_t *collection, iscsi_session_t *session, int lun)
{
    iscsi_task_t *search = session->arrival_order.head;

    while (search) {
        iscsi_task_t *next = search->order_next;

        if (search->scsi_cmnd && search->scsi_cmnd->lun == lun) {
            remove_session_task(session, search);
            add_task(collection, search);
        }
        search = next;
    }
}

/* 
 * remove cmnds for the specified LUN that are in the session's cmnd queue, 
 * or the forced abort queue, and return a list of them.
 */
static Scsi_Cmnd *remove_session_cmnds_for_lun(iscsi_session_t *session, int lun)
{
    Scsi_Cmnd *cmnd = NULL;
    Scsi_Cmnd *prior = NULL;
    Scsi_Cmnd *head = NULL, *tail = NULL;

    /* handle any commands we hid from the tx thread */

    while (session->ignored_cmnd_head && (session->ignored_cmnd_head->lun == lun)) {
        /* move the head */
        cmnd = session->ignored_cmnd_head;
        session->ignored_cmnd_head = (Scsi_Cmnd *)cmnd->host_scribble;
        if (session->ignored_cmnd_head == NULL)
            session->ignored_cmnd_tail = NULL;
        atomic_dec(&session->num_ignored_cmnds);
        if (head) {
            tail->host_scribble = (unsigned char *)cmnd;
            tail = cmnd;
        }
        else {
            cmnd->host_scribble = NULL;
            head = tail = cmnd;
        }
    }
    
    /* we're either out of cmnds, or the head is for a different LUN */
    prior = session->ignored_cmnd_head;
    while (prior && (cmnd = (Scsi_Cmnd *)prior->host_scribble)) {
        if (cmnd->lun == lun) {
            /* splice out cmnd and move it */
            prior->host_scribble = cmnd->host_scribble;
            if (session->ignored_cmnd_tail == cmnd)
                session->ignored_cmnd_tail = prior;
            atomic_dec(&session->num_ignored_cmnds);
            if (head) {
                tail->host_scribble = (unsigned char *)cmnd;
                tail = cmnd;
            }
            else {
                cmnd->host_scribble = NULL;
                head = tail = cmnd;
            }
        }
        else {
            prior = cmnd;
        }
    }

    /* handle cmnds queued for the tx thread to send */

    while (session->scsi_cmnd_head && (session->scsi_cmnd_head->lun == lun)) {
        /* move the head */
        cmnd = session->scsi_cmnd_head;
        session->scsi_cmnd_head = (Scsi_Cmnd *)cmnd->host_scribble;
        if (session->scsi_cmnd_head == NULL)
            session->scsi_cmnd_tail = NULL;
        atomic_dec(&session->num_cmnds);
        if (head) {
            tail->host_scribble = (unsigned char *)cmnd;
            tail = cmnd;
        }
        else {
            cmnd->host_scribble = NULL;
            head = tail = cmnd;
        }
    }
    
    /* we're either out of cmnds, or the head is for a different LUN */
    prior = session->scsi_cmnd_head;

    while (prior && (cmnd = (Scsi_Cmnd *)prior->host_scribble)) {
        if (cmnd->lun == lun) {
            /* splice out cmnd and move it */
            prior->host_scribble = cmnd->host_scribble;
            if (session->scsi_cmnd_tail == cmnd)
                session->scsi_cmnd_tail = prior;
            atomic_dec(&session->num_cmnds);
            if (head) {
                tail->host_scribble = (unsigned char *)cmnd;
                tail = cmnd;
            }
            else {
                cmnd->host_scribble = NULL;
                head = tail = cmnd;
            }
        }
        else {
            prior = cmnd;
        }
    }

    return head;
}

/* mark a Scsi_Cmnd as having a LUN communication failure */
inline void set_lun_comm_failure(Scsi_Cmnd *sc)
{
    sc->sense_buffer[0] = 0x70;
    sc->sense_buffer[2] = NOT_READY;
    sc->sense_buffer[7] = 0x6;
    sc->sense_buffer[12] = 0x08;
    sc->sense_buffer[13] = 0x00;
}

/* mark a Scsi_Cmnd as having a LUN communication failure */
inline void set_data_path_failure(Scsi_Cmnd *sc)
{
    sc->sense_buffer[0] = 0x70;
    sc->sense_buffer[2] = HARDWARE_ERROR;
    sc->sense_buffer[7] = 0x6;
    sc->sense_buffer[12] = 0x41;
    sc->sense_buffer[13] = 0x00;
}

/* decode common network errno values into more useful strings.
 * strerror would be nice right about now.
 */
static char *iscsi_strerror(int errno)
{
    switch (errno) {
        case EIO:
            return "I/O error";
        case EINTR:
            return "Interrupted system call";
        case ENXIO:
            return "No such device or address";
        case EFAULT:
            return "Bad address";
        case EBUSY:
            return "Device or resource busy";
        case EINVAL:
            return "Invalid argument";
        case EPIPE:
            return "Broken pipe";
        case ENONET:
            return "Machine is not on the network";
        case ECOMM:
            return "Communication error on send";
        case EPROTO:
            return "Protocol error";
        case ENOTUNIQ:
            return "Name not unique on network";
        case ENOTSOCK:
            return "Socket operation on non-socket";
        case ENETDOWN:
            return "Network is down";
        case ENETUNREACH:
            return "Network is unreachable";
        case ENETRESET:
            return "Network dropped connection because of reset";
        case ECONNABORTED:
            return "Software caused connection abort";
        case ECONNRESET:
            return "Connection reset by peer";
        case ESHUTDOWN:
            return "Cannot send after shutdown";
        case ETIMEDOUT:
            return "Connection timed out";
        case ECONNREFUSED:
            return "Connection refused";
        case EHOSTDOWN:
            return "Host is down";
        case EHOSTUNREACH:
            return "No route to host";
        default:
            return "";
    }
}


static int iscsi_recvmsg( iscsi_session_t *session, struct msghdr *msg, int len )
{
    int rc = 0;
    mm_segment_t oldfs;

    if (session->socket) {
        oldfs = get_fs();
        set_fs( get_ds() );

        /* Try to avoid memory allocation deadlocks by using GFP_ATOMIC. */
        session->socket->sk->allocation = GFP_ATOMIC;
            
        rc = sock_recvmsg( session->socket, msg, len, MSG_WAITALL);

        set_fs( oldfs ); 
    }

    return rc;
}

static int iscsi_sendmsg( iscsi_session_t *session, struct msghdr *msg, int len )
{
    int rc = 0;
    mm_segment_t oldfs;

    if (session->socket) {
        oldfs = get_fs();
        set_fs( get_ds() );
        
        /* Try to avoid resource acquisition deadlocks by using GFP_ATOMIC. */
        session->socket->sk->allocation = GFP_ATOMIC;

        /* FIXME: ought to loop handling short writes, unless a signal occurs */
        rc = sock_sendmsg(session->socket, msg, len);
        
        set_fs( oldfs );
    }
        
    return rc;
}


/* create and connect a new socket for this session */
int iscsi_connect(iscsi_session_t *session) 
{
    mm_segment_t oldfs;
    struct socket           *socket;
    struct sockaddr_in      addr;
    int window_size;
    int arg = 1, arglen = 0;
    int rc = 0, ret = 0;

    if (session->socket) {
        printk("iSCSI: session %p already has socket %p\n", session, session->socket);
        return 0;
    }

    oldfs = get_fs();
    set_fs( get_ds() );

    if (sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &socket) < 0) {
        printk("iSCSI: failed to create socket\n");
        set_fs(oldfs);
        return 0;
    }

    /* no delay in sending */
    if (socket->ops->setsockopt(socket, IPPROTO_TCP, TCP_NODELAY, (char *)&arg, sizeof(arg)) < 0) {
        printk("iSCSI: failed to setsockopt TCP_NODELAY\n");
        goto done;
    }

    /* try to ensure a reasonably sized TCP window */
    arglen = sizeof(window_size);
    if (sock_getsockopt(socket, SOL_SOCKET, SO_RCVBUF, (char *)&window_size, &arglen) >= 0) {
        DEBUG_FLOW1("iSCSI: TCP recv window size %u\n", window_size);
        
        if (session->tcp_window_size && (window_size < session->tcp_window_size)) {
            window_size = session->tcp_window_size;
            if (sock_setsockopt(socket, SOL_SOCKET, SO_RCVBUF, (char *)&window_size, sizeof(window_size)) < 0) {
                printk("iSCSI: failed to set TCP recv window size to %u\n", window_size);
            }
            else if (sock_getsockopt(socket, SOL_SOCKET, SO_RCVBUF, (char *)&window_size, &arglen) >= 0) {
                DEBUG_INIT2("iSCSI: set TCP recv window size to %u, actually got %u\n", session->tcp_window_size, window_size);
            }
        }
    }
    else {
        printk("iSCSI: getsockopt RCVBUF %p failed\n", socket);
    }
    if (sock_getsockopt(socket, SOL_SOCKET, SO_SNDBUF, (char *)&window_size, &arglen) >= 0) {
        DEBUG_FLOW1("iSCSI: TCP send window size %u\n", window_size);
        
        if (session->tcp_window_size && (window_size < session->tcp_window_size)) {
            window_size = session->tcp_window_size;
            if (sock_setsockopt(socket, SOL_SOCKET, SO_SNDBUF, (char *)&window_size, sizeof(window_size)) < 0) {
                printk("iSCSI: failed to set TCP send window size to %u\n", window_size);
            }
            else if (sock_getsockopt(socket, SOL_SOCKET, SO_SNDBUF, (char *)&window_size, &arglen) >= 0) {
                DEBUG_INIT2("iSCSI: set TCP send window size to %u, actually got %u\n", session->tcp_window_size, window_size);
            }
        }
    }
    else {
        printk("iSCSI: getsockopt SNDBUF %p failed\n", socket);
    }

    /* connect to the target */
    addr.sin_family = AF_INET;
    addr.sin_port = htons(session->port);
    if (session->ip_length == 4) {
        memcpy(&addr.sin_addr.s_addr, session->ip_address, MIN(sizeof(addr.sin_addr.s_addr), session->ip_length));
    }
    else {
        /* FIXME: IPv6 */
        printk("iSCSI: unable to handle IPv6 address, length %u, addr %u.%u.%u.%u\n", 
               session->ip_length, session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3]);
        goto done;
    }
    rc = socket->ops->connect(socket, (struct sockaddr *)&addr, sizeof(addr), 0);

    if (signal_pending(current))
        goto done;

    if (rc < 0) {
        char *error = iscsi_strerror(-rc);
        if (error && error[0] != '\0') {
            printk("iSCSI: session %p to %s failed to connect, rc %d, %s\n", session, session->log_name, rc, error);
        }
        else {
            printk("iSCSI: session %p to %s failed to connect, rc %d\n", session, session->log_name, rc);
        }
    }
    else {
        if (LOG_ENABLED(ISCSI_LOG_LOGIN))
            printk("iSCSI: session %p to %s connected at %lu\n", session, session->log_name, jiffies);
        ret = 1;
    }

 done:
    if (ret) {
        /* save the socket pointer for later */
        session->socket = socket;
    }
    else {
        /* close the socket */
        sock_release(socket);
        session->socket = NULL;
    }
    wmb();
    set_fs(oldfs);
    return ret;
}

void iscsi_disconnect(iscsi_session_t *session) 
{
    if (session->socket) {
#if TCP_ABORT_ON_DROP
        if (test_and_clear_bit(SESSION_DROPPED, &session->control_bits) && 
            !test_bit(SESSION_LOGGED_OUT, &session->control_bits))
        {
            /* setting linger on and lingertime to 0 before closing
             * the socket will trigger a TCP abort (abort all sends
             * and receives, possibly send RST, connection to CLOSED),
             * which is probably what we want if we're dropping and
             * restarting a session.  A TCP Abort will discard TCP
             * data, which is probably a bunch of commands and data
             * we'll resend on a new session anyway.  This frees up
             * skbuffs, and makes the VM livelock less likely.  When
             * we relogin again to the target with the same ISID, the
             * target will kill off the old connections on it's side,
             * so the FIN handshake should be unnecessary, and there
             * are cases where network failures may prevent the FIN
             * handshake from completing, so the connection wouldn't
             * get cleaned up unless the TCP stack has timeouts for
             * some of the TCP states.
             */
            struct linger ling;
            mm_segment_t oldfs;

            memset(&ling, 0, sizeof(ling));
            ling.l_onoff = 1;
            ling.l_linger = 0;
            
            /* we could adjust the socket linger values directly, but using the sockopt call
             * is less likely to break if someone overhauls the socket structure.
             */
            oldfs = get_fs();
            set_fs(get_ds());

            if (sock_setsockopt(session->socket, IPPROTO_TCP, SO_LINGER, (char *)&ling, sizeof(ling)) < 0) {
                printk("iSCSI: session %p couldn't set lingertime to zero after session drop\n", session);
            }
            else {
                DEBUG_INIT1("iSCSI: session %p set lingertime to zero because of session drop\n", session);
            }
            
            set_fs(oldfs);
        }
#endif

        /* close the socket, triggering either a TCP close or a TCP abort */
        sock_release(session->socket);
        
        session->socket = NULL;
        wmb();
    }
}


int iscsi_send_pdu(iscsi_session_t *session, struct IscsiHdr *pdu, char *data, int timeout)
{
    struct msghdr msg;
    struct iovec iov[3];
    char padding[4];
    int pad = 0;
    int rc;
    int pdu_length = 0;
    int data_length;

    if (pdu == NULL) {
        printk("iSCSI: session %p, pdu NULL, can't send PDU header\n", session);
        return 0;
    }

    memset(iov, 0, sizeof(iov));
    memset(&msg, 0, sizeof(msg));
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;
    
    /* pdu header */
    iov[0].iov_base = pdu;
    iov[0].iov_len = sizeof(*pdu);
    pdu_length = sizeof(*pdu);

    /* pdu data */
    data_length = ntoh24(pdu->dlength);
    if (data) {
        iov[msg.msg_iovlen].iov_base = data;
        iov[msg.msg_iovlen].iov_len = data_length;
        msg.msg_iovlen++;
        pdu_length += ntoh24(pdu->dlength);
    }
    else if (data_length) {
        printk("iSCSI: session %p pdu %p with dlength %d, but data NULL\n",
               session, pdu, data_length);
        return 0;
    }

    /* add any padding needed */
    if (pdu_length % PAD_WORD_LEN) {
        memset(padding, 0x0, sizeof(padding));
        pad = PAD_WORD_LEN - (pdu_length % PAD_WORD_LEN);
    }
    if (pad) {
        iov[msg.msg_iovlen].iov_base = padding;
        iov[msg.msg_iovlen].iov_len = pad;
        msg.msg_iovlen++;
        pdu_length += pad;
    }

    /* set a timer, though we shouldn't really need one */
    if (timeout) {
        session->login_phase_timer = jiffies + (timeout * HZ);
        wmb();
    }

    if (LOG_ENABLED(ISCSI_LOG_LOGIN)) {
        char *text = data;
        char *end = text + ntoh24(pdu->dlength);
        int show_text = 0;

        if ((pdu->opcode & ISCSI_OPCODE_MASK) == ISCSI_OP_LOGIN_CMD) {
            struct IscsiLoginHdr *login_pdu = (struct IscsiLoginHdr *)pdu;
            /* show the login phases and tbit */
            printk("iSCSI: session %p sending login pdu with current phase %d, next %d, tbit %d, dlength %d at %lu, timeout at %lu (%d seconds)\n",
                   session, login_pdu->curr, login_pdu->next, login_pdu->tbit, ntoh24(pdu->dlength), 
                   jiffies, session->login_phase_timer, session->login_timeout);
            show_text = 1;
        }
        else if ((pdu->opcode & ISCSI_OPCODE_MASK) == ISCSI_OP_TEXT_CMD) {
            printk("iSCSI: session %p sending text pdu, dlength %d at %lu, timeout at %lu (%d seconds)\n",
                   session, ntoh24(pdu->dlength), 
                   jiffies, session->login_phase_timer, session->login_timeout);
            show_text = 1;
        }
        else {
            printk("iSCSI: session %p sending pdu with opcode 0x%x, dlength %d at %lu, timeout at %lu (%d seconds)\n",
                   session, pdu->opcode, ntoh24(pdu->dlength), jiffies, session->login_phase_timer, session->login_timeout);
        }

        /* show all the text that we're sending */
        while (show_text && (text < end)) {
            printk("iSCSI: session %p login text: %s\n", session, text);
            text += strlen(text);
            while ((text < end) && (*text == '\0'))
                text++;
        }
    }

    rc = iscsi_sendmsg(session, &msg, pdu_length);

    /* clear the timer */
    session->login_phase_timer = 0;
    wmb();

    if (rc != pdu_length) {
        char *error;
        if ((rc < 0) && (error = iscsi_strerror(-rc)) && (error[0] != '\0'))
            printk("iSCSI: session %p failed to send login PDU, rc %d, %s\n", session, rc, iscsi_strerror(-rc));
        else
            printk("iSCSI: session %p failed to send login PDU, rc %d\n", session, rc);

        return 0;
    }

    DEBUG_FLOW5("iSCSI: session %p sent login pdu %p at %lu, length %d, dlength %d\n", 
                session, pdu, jiffies, pdu_length, ntoh24(pdu->dlength));

    return 1;
}

/* try to read an entire login PDU into the buffer, timing out after timeout seconds */
int iscsi_recv_pdu(iscsi_session_t *session, struct IscsiHdr *header, int max_header_length, char *data, int max_data_length, int timeout)
{
    struct msghdr msg;
    struct iovec iov[2];
    char padding[PAD_WORD_LEN];
    int rc = 0;
    int data_length;
    int ret = 0;

    if (header == NULL) {
        printk("iSCSI: session %p, can't receive PDU header into NULL\n", session);
        return 0;
    }

    if (max_header_length < sizeof(*header)) {
        printk("iSCSI: session %p, can't receive %d PDU header bytes into %d byte buffer\n", 
               session, sizeof(*header), max_header_length);
        return 0;
    }

    /* set the timer to implement the timeout requested */
    if (timeout)
        session->login_phase_timer = jiffies + (timeout * HZ);
    else
        session->login_phase_timer = 0;
    mb();
    if (LOG_ENABLED(ISCSI_LOG_LOGIN)) {
        printk("iSCSI: session %p trying to recv login pdu at %lu, timeout at %lu (%d seconds)\n", 
               session, jiffies, session->login_phase_timer, timeout);
    }

    /* read the PDU header */
    memset(iov, 0, sizeof(iov));
    iov[0].iov_base = (void *)header;
    iov[0].iov_len = sizeof(*header);
    memset(&msg, 0, sizeof(struct msghdr));
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;
    
    rc = iscsi_recvmsg(session, &msg, sizeof(*header));

    /* FIXME: check for additional header segments */

    if (signal_pending(current)) {
        printk("iSCSI: session %p recv_login_pdu timed out at %lu\n", session, jiffies);
        goto done;
    }
    
    if (rc != sizeof(*header)) {
        if (rc < 0) {
            char *error = iscsi_strerror(-rc);
            if (error && error[0] != '\0') {
                printk("iSCSI: session %p recv_login_pdu failed to recv %d login PDU bytes, rc %d, %s\n", 
                       session, iov[0].iov_len, rc, iscsi_strerror(-rc));
            }
            else {
                printk("iSCSI: session %p recv_login_pdu failed to recv %d login PDU bytes, rc %d\n", 
                       session, iov[0].iov_len, rc);
            }
        }
        else if (rc == 0) {
            printk("iSCSI: session %p recv_login_pdu: connection closed\n", session);
        }
        else {
            /* short reads should be impossible unless a signal occured,
             * which we already checked for.
             */
            printk("iSCSI: bug - session %p recv_login_pdu, short read %d of %d\n", session, rc, sizeof(*header));
        }
        goto done;
    }
    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    mb();

    /* possibly read PDU data */
    data_length = ntoh24(header->dlength);
    if (data_length) {
        /* check for buffer overflow */
        if (data_length > max_data_length) {
            printk("iSCSI: session %p recv_login_pdu can't read %d bytes of login PDU data, only %d bytes of buffer available\n",
                   session, data_length, max_data_length);
            goto done;
        }

        /* read the PDU's text data payload */
        memset(&msg, 0, sizeof(struct msghdr));
        msg.msg_iov = iov;
        msg.msg_iovlen = 1;

        memset(iov, 0, sizeof(iov));
        iov[0].iov_base = data;
        iov[0].iov_len = data_length;

        /* handle PDU padding */
        if (data_length % PAD_WORD_LEN) {
            int pad = PAD_WORD_LEN - (data_length % PAD_WORD_LEN);

            iov[1].iov_base = padding;
            iov[1].iov_len = pad;
            msg.msg_iovlen = 2;
            data_length += pad;
        }

        rc = iscsi_recvmsg(session, &msg, data_length);

        if (signal_pending(current)) {
            printk("iSCSI: session %p recv_login_pdu timed out at %lu\n", session, jiffies);
            goto done;
        }
        
        if (rc != data_length) {
            if (rc < 0) {
                char *error = iscsi_strerror(-rc);
                if (error && error[0] != '\0') {
                    printk("iSCSI: session %p recv_login_pdu failed to recv %d login data PDU bytes, rc %d, %s\n", 
                           session, data_length, rc, iscsi_strerror(-rc));
                }
                else {
                    printk("iSCSI: session %p recv_login_pdu failed to recv %d login data PDU bytes, rc %d\n", 
                           session, data_length, rc);
                }
                ret = rc;
            }
            else if (rc == 0) {
                printk("iSCSI: session %p recv_login_pdu: connection closed\n", session);
            }
            else {
                /* short reads should be impossible unless a signal occured,
                 * which we already checked for.
                 */
                printk("iSCSI: bug - session %p recv_login_pdu, short read %d of %d\n", session, rc, data_length);
            }
            goto done;
        }

        /* assume a PDU round-trip, connection is ok */
        session->last_rx = jiffies;
        mb();
    }

    if (LOG_ENABLED(ISCSI_LOG_LOGIN)) {
        char *text = data;
        char *end = text + ntoh24(header->dlength);
        int show_text = 0;

        if (header->opcode == ISCSI_OP_LOGIN_RSP) {
            struct IscsiLoginRspHdr *login_pdu = (struct IscsiLoginRspHdr *)header;
            /* show the login phases and tbit */
            printk("iSCSI: session %p received login pdu response at %lu with current phase %d, next %d, tbit %d, dlength %d\n",
                   session, jiffies, login_pdu->curr, login_pdu->next, login_pdu->tbit, ntoh24(header->dlength));
            show_text = 1;
        }
        else if (header->opcode == ISCSI_OP_TEXT_RSP) {
            printk("iSCSI: session %p received text pdu response with dlength %d at %lu\n",
                   session, ntoh24(header->dlength), jiffies);
            show_text = 1;
        }
        else {
            printk("iSCSI: session %p received pdu with opcode 0x%x, dlength %d at %lu\n",
                   session, header->opcode, ntoh24(header->dlength), jiffies);
        }

        /* show all the text that we're sending */
        while (show_text && (text < end)) {
            printk("iSCSI: session %p login resp text: %s\n", session, text);
            text += strlen(text);
            while ((text < end) && (*text == '\0'))
                text++;
        }
    }
    
    ret = 1;
    
 done:    
    /* clear the timer */
    session->login_phase_timer = 0;
    mb();
    iscsi_handle_signals(session);

    return ret;
}


#if DEBUG_TRACE
static void
iscsi_fill_trace(unsigned char type, Scsi_Cmnd *sc, iscsi_task_t *task, unsigned long data1, unsigned long data2)
{
    iscsi_trace_entry_t *te;
    DECLARE_NOQUEUE_FLAGS;

    SPIN_LOCK_NOQUEUE(&iscsi_trace_lock);

    te = &trace_table[trace_index];
    trace_index++;
    if ( trace_index >= ISCSI_TRACE_COUNT ) {
        trace_index = 0;
    }
    memset(te, 0x0, sizeof(*te));

    te->type = type;
    if (sc) {
        te->cmd = sc->cmnd[0];
        te->host = sc->host->host_no;
        te->channel = sc->channel;
        te->target = sc->target;
        te->lun = sc->lun;
    }
    if (task) {
        te->itt = task->itt;
    }
    te->data1 = data1;
    te->data2 = data2;
    te->jiffies = jiffies;

    SPIN_UNLOCK_NOQUEUE(&iscsi_trace_lock);
}
#endif

/* FIXME: update for 16 byte CDBs, such as:
   lock unlock cache 16
   pre-fetch 16
   read 16
   rebuild 16
   regenerate 16
   synchronize cache 16
   verify 16
   write 16
   write and verify 16
   write same 16
   xdwrite extended 16

   Then increase ISCSI_MAX_CMD_LEN to 16 in iscsi.h.
*/
/* FIXME: for that matter, check the existing list for correctness */
static int
iscsi_set_direction(iscsi_task_t *task)
{
    if (task && task->scsi_cmnd) {
        switch (task->scsi_cmnd->cmnd[0]) {
            case TEST_UNIT_READY:
            case START_STOP:
            case REZERO_UNIT:
            case WRITE_FILEMARKS:
            case SPACE:
            case ERASE:
            case ALLOW_MEDIUM_REMOVAL:
                /* just control commands */
                set_bit(ISCSI_TASK_CONTROL, &task->flags);
                return ISCSI_TASK_CONTROL;
            case WRITE_6:           case WRITE_10:          case WRITE_12: 
            case 0x8a: /* WRITE_16 */        case 0x8e: /* write and verify 16 */
            case 0x93: /* write same 16 */
            case WRITE_LONG:        case WRITE_SAME:        case WRITE_BUFFER:
            case WRITE_VERIFY:      case WRITE_VERIFY_12:
            case COMPARE:           case COPY:              case COPY_VERIFY:
            case SEARCH_EQUAL:      case SEARCH_HIGH:       case SEARCH_LOW:
            case SEARCH_EQUAL_12:   case SEARCH_HIGH_12:    case SEARCH_LOW_12:
            case FORMAT_UNIT:       case REASSIGN_BLOCKS:   case RESERVE:
            case MODE_SELECT:       case MODE_SELECT_10:    case LOG_SELECT:
            case SEND_DIAGNOSTIC:   case CHANGE_DEFINITION: case UPDATE_BLOCK:
            case SET_WINDOW:        case MEDIUM_SCAN:       case SEND_VOLUME_TAG:
            case WRITE_LONG_2:
                set_bit(ISCSI_TASK_WRITE, &task->flags);
                return ISCSI_TASK_WRITE;
            default:
                set_bit(ISCSI_TASK_READ, &task->flags);
                return ISCSI_TASK_READ;
        }
    }
    
    return -1;
}

/* tagged queueing */
static void iscsi_set_tag(Scsi_Cmnd *cmd, struct IscsiScsiCmdHdr *hdr)
{
    if (cmd->device && cmd->device->tagged_queue) {
        switch (cmd->tag) {
            case HEAD_OF_QUEUE_TAG:
                hdr->flags.attr = ISCSI_ATTR_HEAD_OF_QUEUE;
                break;
            case ORDERED_QUEUE_TAG:
                hdr->flags.attr = ISCSI_ATTR_ORDERED;
                break;
            default:
                hdr->flags.attr = ISCSI_ATTR_SIMPLE;
                break;
        }
    }
    else
        hdr->flags.attr = ISCSI_ATTR_UNTAGGED;
}

void print_cmnd(Scsi_Cmnd *sc)
{
    printk("iSCSI: Scsi_Cmnd %p to (%u %u %u %u), Cmd 0x%x\n"
           "       done %p, scsi_done %p, host_scribble %p\n"
           "       reqbuf %p, req_len %u\n"
           "       buffer %p, bufflen %u\n"
           "       use_sg %u, old_use_sg %u, sglist_len %u\n"
           "       owner 0x%x, state  0x%x, eh_state 0x%x\n"
           "       cmd_len %u, old_cmd_len %u, abort_reason 0x%x\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0],
           sc->done, sc->scsi_done, sc->host_scribble, 
           sc->request_buffer, sc->request_bufflen, sc->buffer, sc->bufflen,
           sc->use_sg, sc->old_use_sg, sc->sglist_len,
           sc->owner, sc->state, sc->eh_state, 
           sc->cmd_len, sc->old_cmd_len, sc->abort_reason);

    if (sc->cmd_len >= 12)
        printk("iSCSI: cdb %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x\n",
               sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
               sc->cmnd[4], sc->cmnd[5], sc->cmnd[6], sc->cmnd[7],
               sc->cmnd[8], sc->cmnd[9], sc->cmnd[10], sc->cmnd[11]);
    else if (sc->cmd_len >= 8)
        printk("iSCSI: cdb %02x%02x%02x%02x %02x%02x%02x%02x\n",
               sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
               sc->cmnd[4], sc->cmnd[5], sc->cmnd[6], sc->cmnd[7]);
    else if (sc->cmd_len >= 6)
        printk("iSCSI: cdb %02x%02x%02x%02x %02x%02x\n",
               sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
               sc->cmnd[4], sc->cmnd[5]);
    else if (sc->cmd_len >= 4)
        printk("iSCSI: cdb %02x%02x%02x%02x\n",
               sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3]);
    else if (sc->cmd_len >= 2)
        printk("iSCSI: cdb %02x%02x\n", sc->cmnd[0], sc->cmnd[1]);

    if (sc->use_sg && sc->request_buffer) {
        struct scatterlist *sglist = (struct scatterlist *)sc->request_buffer;
        int i;
        
        for (i = 0; i < sc->use_sg; i++) {
#if (HAS_SCATTERLIST_PAGE && HAS_SCATTERLIST_ADDRESS)
            printk("iSCSI: sglist %p index %02d = addr %p, page %p, offset %u, len %u\n", 
                   sglist, i, sglist->address, sglist->page, sglist->offset, sglist->length);
#elif HAS_SCATTERLIST_PAGE
            printk("iSCSI: sglist %p index %02d = page %p, offset %u, len %u\n", 
                   sglist, i, sglist->page, sglist->offset, sglist->length);
#else
            printk("iSCSI: sglist %p index %02d = addr %p, len %u\n", 
                   sglist, i, sglist->address, sglist->length);
#endif
            sglist++;
        }
    }
}

#if DEBUG
/* caller must hold the session's scsi_cmnd_lock */
static void print_session_cmnds(iscsi_session_t *session)
{
    Scsi_Cmnd *search = session->scsi_cmnd_head;
    printk("iSCSI: session %p to %s unsent cmnd queue: head %p, tail %p, num %u\n", 
           session, session->log_name, session->scsi_cmnd_head, session->scsi_cmnd_tail,
           atomic_read(&session->num_cmnds));
    while (search) {
        printk("iSCSI: session %p u cmnd %p: state %4x, eh_state %4x, scribble %p, Cmd 0x%x to (%u %u %u %u)\n",
               session, search, search->state, search->eh_state, search->host_scribble,
               search->cmnd[0], search->host->host_no, search->channel, search->target, search->lun);
        search = (Scsi_Cmnd *)search->host_scribble;
    }
    printk("iSCSI: session %p to %s ignored cmnd queue: head %p, tail %p, num %u\n", 
           session, session->log_name, session->ignored_cmnd_head, session->ignored_cmnd_tail,
           atomic_read(&session->num_ignored_cmnds));
    search = session->ignored_cmnd_head;
    while (search) {
        printk("iSCSI: session %p i cmnd %p: state %4x, eh_state %4x, scribble %p, Cmd 0x%x to (%u %u %u %u)\n",
               session, search, search->state, search->eh_state, search->host_scribble,
               search->cmnd[0], search->host->host_no, search->channel, search->target, search->lun);
        search = (Scsi_Cmnd *)search->host_scribble;
    }
}

/* caller must hold the session's task_lock */
static void print_session_tasks(iscsi_session_t *session)
{
    iscsi_task_t *task = NULL;
    Scsi_Cmnd *cmnd = NULL;

    if (session->rx_tasks.head)
        printk("iSCSI: session %p to %s rx_tasks: head %p, tail %p\n", 
               session, session->log_name, session->rx_tasks.head, session->rx_tasks.tail);
    if (session->tx_tasks.head)
        printk("iSCSI: session %p to %s tx_tasks: head %p, tail %p\n", 
               session, session->log_name, session->tx_tasks.head, session->tx_tasks.tail);
    if (session->completing_tasks.head)
        printk("iSCSI: session %p to %s completing_tasks: head %p, tail %p\n", 
               session, session->log_name, session->completing_tasks.head, session->completing_tasks.tail);
    if (session->rx_abort_tasks.head)
        printk("iSCSI: session %p to %s rx_abort_tasks: head %p, tail %p\n", 
               session, session->log_name, session->rx_abort_tasks.head, session->rx_abort_tasks.tail);
    if (session->tx_abort_tasks.head)
        printk("iSCSI: session %p to %s tx_abort_tasks: head %p, tail %p\n", 
               session, session->log_name, session->tx_abort_tasks.head, session->tx_abort_tasks.tail);
    if (session->aborted_tasks.head)
        printk("iSCSI: session %p to %s aborted_tasks: head %p, tail %p\n", 
               session, session->log_name, session->aborted_tasks.head, session->aborted_tasks.tail);
    if (session->rx_lun_reset_tasks.head)
        printk("iSCSI: session %p to %s rx_lun_reset_tasks: head %p, tail %p\n", 
               session, session->log_name, session->rx_lun_reset_tasks.head, session->rx_lun_reset_tasks.tail);
    if (session->tx_lun_reset_tasks.head)
        printk("iSCSI: session %p to %s tx_lun_reset_tasks: head %p, tail %p\n", 
               session, session->log_name, session->tx_lun_reset_tasks.head, session->tx_lun_reset_tasks.tail);
    if (session->lun_reset_tasks.head)
        printk("iSCSI: session %p to %s lun_reset_tasks: head %p, tail %p\n", 
               session, session->log_name, session->lun_reset_tasks.head, session->lun_reset_tasks.tail);

    printk("iSCSI: session %p to %s task queue: head %p, tail %p, num %u\n", 
           session, session->log_name, session->arrival_order.head, session->arrival_order.tail,  
           atomic_read(&session->num_active_tasks));

    task = session->arrival_order.head;
    while (task) {
        if ((cmnd = task->scsi_cmnd)) 
            printk("iSCSI: session %p task %p itt %u prev %p next %p, cmnd %p: state %4x, eh_state %4x, scribble %p, Cmd 0x%x to (%u %u %u %u)\n",
                   session, task, task->itt, task->prev, task->next,
                   cmnd, cmnd->state, cmnd->eh_state, cmnd->host_scribble,
                   cmnd->cmnd[0], cmnd->host->host_no, cmnd->channel, cmnd->target, cmnd->lun);
        else
            printk("iSCSI: session %p task %p itt %u for NULL cmnd, prev %p, next %p\n", 
                   session, task, task->itt, task->prev, task->next);
        
        task = task->order_next;
    }
}

#endif


static int iscsi_xmit_task(iscsi_task_t *task)
{
    struct msghdr msg;
    struct iovec iov;
    struct IscsiScsiCmdHdr stsch;
    int rc, wlen;
    iscsi_session_t *session = task->session;
    Scsi_Cmnd *sc = task->scsi_cmnd;

    if (!task) {
        printk("iSCSI: xmit_task NULL\n");
        return 0;
    }

    if (!sc) {
        printk("iSCSI: xmit_task %p, cmnd NULL\n", task);
        return 0;
    }
    
#if DEBUG_FLOW
    if (LOG_ENABLED(ISCSI_LOG_FLOW))
        printk("iSCSI: xmit_task %p, itt %u to (%u %u %u %u), Cmd 0x%x, cmd_len %u, bufflen %u\n",  
               task, task->itt, sc->host->host_no, sc->channel, sc->target,sc->lun, sc->cmnd[0], 
               sc->cmd_len, sc->request_bufflen);
#endif

    wlen = sizeof(stsch);
    memset( &stsch, 0, sizeof(stsch) );

    if (test_bit(ISCSI_TASK_READ, &task->flags)) {
        /* read */
        stsch.flags.read_data = 1;
        stsch.data_length = htonl(iscsi_expected_data_length(sc));
    }
    else if (test_bit(ISCSI_TASK_WRITE, &task->flags)) {
        /* write */
        stsch.flags.write_data = 1;                                        
        stsch.data_length = htonl(iscsi_expected_data_length(sc));
    }

    /* tagged command queueing */
    iscsi_set_tag( sc, &stsch );
    
    /* FIXME: if it's an untagged command, and we've already sent
     * an untagged command to the LUN, don't send a 2nd untagged command.
     * Leave it queued up and send it after the other command completes.
     * We also don't want to block commands for other LUNs.  Basically,
     * we need a per-LUN command queue.  For now, deal with it by
     * setting the Scsi_Device queue_depth to 1 without TCQ.  We can
     * reduce latency by keeping multiple commands per LUN queued to
     * the HBA, but only sending one.  That takes a more code though.
     */

    stsch.opcode = ISCSI_OP_SCSI_CMD;
    stsch.itt = htonl(task->itt);
    task->cmdsn = session->CmdSn;
    stsch.cmdsn = htonl(session->CmdSn);
    stsch.expstatsn = htonl(session->ExpStatSn);
    
    /* set the F-bit unless unsolicited Data-out PDUs will be sent */
    if (test_bit(ISCSI_TASK_INITIAL_R2T, &task->flags))
        stsch.flags.final = 0;
    else
        stsch.flags.final = 1;
    
    /* single level LUN format puts LUN in byte 1, 0 everywhere else */
    stsch.lun[1] = sc->lun;

    memcpy( stsch.scb, sc->cmnd, sc->cmd_len );

    iov.iov_base = &stsch;
    iov.iov_len = sizeof(stsch);
    memset( &msg, 0, sizeof(struct msghdr) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    ISCSI_TRACE(ISCSI_TRACE_TxCmd, sc, task, session->CmdSn, ntohl(stsch.data_length));

    /* FIXME: possibly send ImmediateData along with the cmd PDU */

    rc = iscsi_sendmsg( session, &msg, wlen );
    if ( rc != wlen ) {
        printk("iSCSI: session %p xmit_task sendmsg %d failed, error %d\n", session, wlen, rc);
        iscsi_drop_session(session);
        return 0;
    }

    session->CmdSn++;
    
    /* unsolicited data is handled elsewhere */

    return 1;
}


static void iscsi_xmit_queued_cmnds( iscsi_session_t *session)
{
    Scsi_Cmnd *sc;
    iscsi_task_t *task;
    DECLARE_NOQUEUE_FLAGS;

    if (!session) {
        DEBUG_ERR0("iSCSI: can't xmit queued commands, no session\n");
        return;
    }

    for (;;) {
        
        if (signal_pending(current)) {
            DEBUG_QUEUE1("iSCSI: session %p can't start tasks now, signal pending\n", session);
            return;
        }
    
        if (!sna_lte(session->CmdSn, session->MaxCmdSn)) {
            DEBUG_QUEUE5("iSCSI: session %p can't start %u tasks now, ExpCmdSN %u, CmdSn %u, MaxCmdSN %u\n",
                         session, atomic_read(&session->num_cmnds), 
                         session->ExpCmdSn, session->CmdSn, session->MaxCmdSn);
            return;
        }
        
        if (test_bit(SESSION_RESETTING, &session->control_bits)) {
            DEBUG_QUEUE1("iSCSI: session %p resetting, can't start tasks now\n", session);
            return;
        }

        if (atomic_read(&session->num_cmnds) == 0) {
            DEBUG_QUEUE2("iSCSI: no SCSI cmnds queued for session %p to %s\n", session, session->log_name);
            return;
        }
        
        DEBUG_QUEUE3("iSCSI: session %p xmit_queued_cmnds, CmdSN %u, MaxCmdSN %u\n",
                     session, session->CmdSn, session->MaxCmdSn);
        
        /* allocate a task, without holding any locks, so that the
         * allocation function may block if necessary.  
         */
        task = alloc_task(session);
        if (!task) {
            DEBUG_ERR2("iSCSI: couldn't allocate task for session %p to %s\n",
                       session, session->log_name);
            /* to prevent a stall of the driver, free_task must wakeup
             * the tx thread later.
             */
            return;
        }

        /* add the task to the session before removing the Scsi_Cmnd
         * from the queue, to ensure the error handlers can always
         * find either the command or the task.  Need to hold both
         * locks to atomically convert a cmnd to a task, and must
         * respect the lock order.  
         */
        spin_lock(&session->task_lock);

        /* Don't start any new tasks if a Logout has been requested.  */
        if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
            spin_unlock(&session->task_lock);
            DEBUG_QUEUE1("iSCSI: session %p logout requested, can't start tasks now\n", session);
            spin_lock(&session->hba->free_task_lock);
            free_task(task);
            spin_unlock(&session->hba->free_task_lock);
            return;
        }

        /* try to grab one SCSI command off the session's command queue */
        SPIN_LOCK_NOQUEUE(&session->scsi_cmnd_lock);
        if ((sc = session->scsi_cmnd_head) == NULL) {
            /* this should never happen if session->num_cmnds is accurate */
            printk("iSCSI: bug - no SCSI cmnds queued at %lu for session %p, num_cmnds %u, head %p, tail %p\n", 
                   jiffies, session, atomic_read(&session->num_cmnds), session->scsi_cmnd_head, session->scsi_cmnd_tail);
            atomic_set(&session->num_cmnds, 0);
            SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);
            spin_unlock(&session->task_lock);

            /* free the task */
            spin_lock(&session->hba->free_task_lock);
            free_task(task);
            spin_unlock(&session->hba->free_task_lock);
            return;
        }

        task->scsi_cmnd = sc;
        atomic_inc(&task->refcount);
        wmb();
        add_session_task(session, task);
        
        /* determine which direction the task's data will transfer */
        iscsi_set_direction(task);

        /* possibly queue up unsolicited data PDUs.
         * FIXME: the ImmediateData check needs to be fixed here.
         * With ImmediateData, we may or may not have to send
         * additional Data PDUs, depending on the amount of data, and
         * the Max PDU Length.
         */
        if (test_bit(ISCSI_TASK_WRITE, &task->flags) && !session->InitialR2T && !session->ImmediateData &&
            sc->request_buffer && sc->request_bufflen && iscsi_expected_data_length(sc)) {

            /* queue up unsolicited data PDUs.  the implied initial R2T doesn't count 
             * against the MaxOutstandingR2T, so we can't use the normal R2T fields of 
             * the task for the implied initial R2T.  Use a special flag for the implied 
             * initial R2T, and let the rx thread update tasks in the tx_tasks collection 
             * if an R2T comes in before the implied initial R2T has been processed.
             */
            set_bit(ISCSI_TASK_INITIAL_R2T, &task->flags);
            
            /* queue up an implied R2T data transfer for later */
            add_task(&session->tx_tasks, task);
            set_bit(TX_DATA, &session->control_bits);
        }
        else { 
            /* we'll receive something from the target after sending the command */
            add_task(&session->rx_tasks, task);
        }

        /* remove Cmnd from queue */
        session->scsi_cmnd_head = (Scsi_Cmnd *)sc->host_scribble;
        sc->host_scribble = NULL;
        if (session->scsi_cmnd_head == NULL)
            session->scsi_cmnd_tail = NULL;
        atomic_dec(&session->num_cmnds);
        
#if (DEBUG_SMP > 0) || (DEBUG_QUEUE > 0)
        if (LOG_ENABLED(ISCSI_LOG_SMP) || LOG_ENABLED(ISCSI_LOG_QUEUE))
            printk("iSCSI: cmnd %p became task %p at %lu for session %p, num_cmnds %u, head %p, tail %p\n", 
                   sc, task, jiffies, session, atomic_read(&session->num_cmnds), session->scsi_cmnd_head, session->scsi_cmnd_tail);
#endif
        
        SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);
        spin_unlock(&session->task_lock);

        DEBUG_FLOW4("iSCSI: sending itt %u on session %p as CmdSN %u, MaxCmdSn %u\n",
                     task->itt, session, session->CmdSn, session->MaxCmdSn);

        if (iscsi_xmit_task(task)) {
            DEBUG_FLOW4("iSCSI: sent itt %u, task %p, cmnd %p, refcount %d\n", task->itt, task, sc, atomic_read(&task->refcount));
            atomic_dec(&task->refcount);
            DEBUG_FLOW3("iSCSI: after sending itt %u, task %p now has refcount %d\n", task->itt, task, atomic_read(&task->refcount));
        }
        else {
            printk("iSCSI: failed to send itt %u, task %p, cmnd %p, returning to queue\n", 
                   task->itt, task, sc);
            
            spin_lock(&session->task_lock);

            /* put it back on the head of the Scsi_Cmnd queue, we couldn't send it */
            SPIN_LOCK_NOQUEUE(&session->scsi_cmnd_lock);
            if ( session->scsi_cmnd_head ) {
                sc->host_scribble = (unsigned char *)session->scsi_cmnd_head;
                session->scsi_cmnd_head = sc;
            }
            else {
                sc->host_scribble = NULL;
                session->scsi_cmnd_head = session->scsi_cmnd_tail = sc;
            }
            atomic_inc(&session->num_cmnds);
            SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);

            /* get rid of the task */
            if (test_and_clear_bit(ISCSI_TASK_INITIAL_R2T, &task->flags)) {
                remove_task(&session->tx_tasks, task->itt);
            }
            else {
                remove_task(&session->rx_tasks, task->itt);
            }

            if (!remove_session_task(session, task)) {
                printk("iSCSI: bug - couldn't xmit task %p, or remove it from session %p to %s\n",
                       task, session, session->log_name);
                task = NULL;
            }
            spin_unlock(&session->task_lock);

            /* if we removed it from the session, free it */
            if (task) {
                spin_lock(&session->hba->free_task_lock);
                free_task(task);
                spin_unlock(&session->hba->free_task_lock);
            }

            /* don't try to send any more right now */
            return;
        }
    }
}

static inline void *sg_virtual_address(struct scatterlist *sg)
{
#if (HAS_SCATTERLIST_PAGE && HAS_SCATTERLIST_ADDRESS)
    /* page may or may not be mapped */
    if (sg->address) {
        return sg->address;
    }
    else if (sg->page) {
        return page_address(sg->page) + sg->offset;
    }
    return NULL;

#elif HAS_SCATTERLIST_PAGE
    /* should have already mapped the page */
    if (sg->page) {
        return page_address(sg->page) + sg->offset;
    }
    return NULL;
#else
    return sg->address;
#endif    

}

static inline void *kmap_sg(struct scatterlist *sg)
{
#if (HAS_SCATTERLIST_PAGE && HAS_SCATTERLIST_ADDRESS)
    /* page may or may not be mapped if HIGHMEM is in use */
    if (sg->address) {
        DEBUG_FLOW2("iSCSI: kmap sg %p to address %p\n", sg, sg->address);
        return sg->address;
    }
    else if (sg->page) {
        void *addr = kmap(sg->page);
        DEBUG_FLOW3("iSCSI: kmap sg %p page %p to addr %p\n", sg, sg->page, addr);
        return addr;
    }
    return NULL;

#elif HAS_SCATTERLIST_PAGE
    /* there is no address, must kmap the page */
    if (sg->page) {
        return kmap(sg->page);
    }
    return NULL;

#else
    /* just use the address */
    DEBUG_FLOW2("iSCSI: kmap sg %p to address %p\n", sg, sg->address);
    return sg->address;
#endif 
}

static inline void kunmap_sg(struct scatterlist *sg) 
{
#if (HAS_SCATTERLIST_PAGE && HAS_SCATTERLIST_ADDRESS)
    if (!sg->address && sg->page)
        kunmap(sg->page);
#elif HAS_SCATTERLIST_PAGE
    if (sg->page)
        kunmap(sg->page);
#endif    
    return;
}


static void iscsi_xmit_data(iscsi_task_t *task, uint32_t ttt, uint32_t data_offset, uint32_t data_length)
{
    struct msghdr msg;
    struct IscsiDataHdr stdh;
    Scsi_Cmnd *sc = NULL;
    iscsi_session_t *session = task->session;
    struct scatterlist *sglist = NULL, *sg, *first_sg = NULL, *last_sg = NULL;
    int wlen, rc, iovn;
    unsigned int segment_offset = 0, index = 0;
    int remain, xfrlen;
    uint32_t data_sn = 0;
    int bytes_to_fill, bytes_from_segment;
    char padding[4];
    int pad_bytes;

    /* make sure we have data to send */
    sc = task->scsi_cmnd;
    if (!sc) {
        printk("iSCSI: xmit_data task %p, cmnd NULL, ttt %u, offset %u, length %u\n", 
               task, ttt, data_offset, data_length);
        return;
    }
    if ((sc->request_bufflen == 0) || (sc->request_buffer == NULL)) {
        printk("iSCSI: xmit_data for itt %u, task %p, sc %p, dlength %u, expected %u, no data in buffer\n"
               "       request_buffer %p len %u, buffer %p len %u\n",
               task->itt, task, sc, data_length, iscsi_expected_data_length(sc),
               sc->request_buffer, sc->request_bufflen, sc->buffer, sc->bufflen);
        print_cmnd(sc);
        return;
    }
    if ((data_length == 0) || (iscsi_expected_data_length(sc) == 0)) {
        printk("iSCSI: xmit_data for itt %u, task %p, data length %u, expected %u\n"
               "       request_buffer %p len %u, buffer %p len %u\n",
               task->itt, task, data_length, iscsi_expected_data_length(sc),
               sc->request_buffer, sc->request_bufflen, sc->buffer, sc->bufflen);
        print_cmnd(sc);
        return;
    }
    remain = data_length;
    
    memset( &stdh, 0, sizeof(stdh) );
    stdh.opcode = ISCSI_OP_SCSI_DATA;
    stdh.itt = htonl(task->itt);
    stdh.ttt = ttt;
    stdh.offset = htonl(data_offset);
    
    /* PDU header */
    session->TxIov[0].iov_base = &stdh;
    session->TxIov[0].iov_len  = sizeof(stdh);

#if DEBUG_FLOW
    if (LOG_ENABLED(ISCSI_LOG_FLOW))
        printk("iSCSI: xmit_data for itt %u, task %p, credit %d @ %u\n"
               "       request_buffer %p len %u, buffer %p len %u\n",
               task->itt, task, remain, data_offset,
               sc->request_buffer, sc->request_bufflen, sc->buffer, sc->bufflen);
#endif
    
    /* Find the segment and offset within the segment to start writing from.  */
    if ( sc->use_sg ) {
        sg = sglist = (struct scatterlist *)sc->request_buffer;

        segment_offset = data_offset;

        for (index = 0; index < sc->use_sg; index++) {
            if (segment_offset < sglist[index].length)
                break;
            else 
                segment_offset -= sglist[index].length;
        }

        if (index >= sc->use_sg) {
            /* didn't find the offset, command will eventually timeout */
            printk("iSCSI: session %p xmit_data for itt %u couldn't find offset %u in sglist %p, sc %p, bufflen %u, use_sg %u\n",
                   session, task->itt, data_offset, sglist, sc, sc->request_bufflen, sc->use_sg);
            print_cmnd(sc);
            ISCSI_TRACE(ISCSI_TRACE_OutOfData, sc, task, index, sc->use_sg);
            return;
        }
    }

    ISCSI_TRACE(ISCSI_TRACE_TxData, sc, task, data_offset, data_length);
    
    /* 
     * find all the PDU data
     */
    while (!signal_pending(current) && remain) {
        
        /* if command PDUs are small (no immediate data),
         * start commands as soon as possible, so that we can
         * overlap the R2T latency with the time it takes to
         * send data for commands already issued.  This increases 
         * throughput without significantly increasing the completion 
         * time of commands already issued.
         */
        if (!session->ImmediateData) {
            DEBUG_FLOW1("iSCSI: checking for new commands before sending data to %s\n",
                        session->log_name);
            iscsi_xmit_queued_cmnds(session);
        }

        stdh.datasn = htonl(data_sn++);
        stdh.offset = htonl(data_offset);
        stdh.expstatsn = htonl(session->ExpStatSn);
        
        if (session->MaxXmitDataSegmentLength && (remain > session->MaxXmitDataSegmentLength)) {
            /* enforce the target's data segment limit */
            bytes_to_fill = session->MaxXmitDataSegmentLength;
        } 
        else {
            /* final PDU of a data burst */
            bytes_to_fill = remain;
            stdh.final = 1;
        }

        /* check if we need to pad the PDU */
        if (bytes_to_fill % PAD_WORD_LEN) {
            pad_bytes = PAD_WORD_LEN - (bytes_to_fill % PAD_WORD_LEN);
            memset(padding, 0x0, sizeof(padding));
        }
        else {
            pad_bytes = 0;
        }
        
        DEBUG_FLOW4("iSCSI: remain %d, bytes_to_fill %d, sc->use_sg %u, MaxRecvDataSegmentLength %d\n",
                    remain, bytes_to_fill, sc->use_sg, session->MaxRecvDataSegmentLength);

        if (sc->use_sg) {
            iovn = 1;
            xfrlen = 0;

            /* while there is more data and we want to send more data */
            while (bytes_to_fill > 0) {
                
                if (index >= sc->use_sg) {
                    printk("iSCSI: session %p xmit_data index %d exceeds sc->use_sg %d, bytes_to_fill %d, out of buffers\n", 
                           session, index, sc->use_sg, bytes_to_fill);
                    /* the command will eventually timeout */
                    print_cmnd(sc);
                    ISCSI_TRACE(ISCSI_TRACE_OutOfData, sc, task, index, sc->use_sg);
                    goto done;
                }
                if (signal_pending(current)) {
                    DEBUG_FLOW1("iSCSI: session %p signal pending, returning from xmit_data\n", session);
                    goto done;
                }

                sg = &sglist[index];

                /* make sure the segment is mapped */
                if (!kmap_sg(sg)) {
                    printk("iSCSI: session %p xmit_data couldn't map segment %p\n", session, sg);
                    goto done;
                }
                else if (first_sg == NULL) {
                    first_sg = sg;
                }
                last_sg = sg;

                /* sanity check the sglist segment length */
                if (sg->length <= segment_offset) {
                    /* the sglist is corrupt */
                    printk("iSCSI: session %p xmit_data index %d, length %u too small for offset %u, bytes_to_fill %d, sglist has been corrupted\n", 
                           session, index, sg->length, segment_offset, bytes_to_fill);
                    /* the command will eventually timeout */
                    print_cmnd(sc);
                    ISCSI_TRACE(ISCSI_TRACE_BadTxSeg, sc, task, sg->length, segment_offset);
                    goto done;
                }
                
                bytes_from_segment = sg->length - segment_offset;
                if ( bytes_from_segment > bytes_to_fill ) {
                    /* only need part of this segment */
                    session->TxIov[iovn].iov_base = sg_virtual_address(sg) + segment_offset;
                    session->TxIov[iovn].iov_len  = bytes_to_fill;
                    xfrlen += bytes_to_fill;
                    DEBUG_FLOW6("iSCSI: session %p xmit_data xfrlen %d, to_fill %d, from_segment %d, iov[%2d] = partial sg[%2d]\n", 
                                session, xfrlen, bytes_to_fill, bytes_from_segment, iovn, index);
                    iovn++;
                    segment_offset += bytes_to_fill;
                    break;
                } 
                else {
                    /* need all of this segment, and possibly more from the next */
                    session->TxIov[iovn].iov_base = sg_virtual_address(sg) + segment_offset;
                    session->TxIov[iovn].iov_len  = bytes_from_segment;
                    xfrlen += bytes_from_segment;
                    DEBUG_FLOW6("iSCSI: session %p xmit_data xfrlen %d, to_fill %d, from_segment %d, iov[%2d] = sg[%2d]\n", 
                                session, xfrlen, bytes_to_fill, bytes_from_segment, iovn, index);
                    bytes_to_fill -= bytes_from_segment;
                    iovn++;
                    /* any remaining data starts at offset 0 of the next segment */
                    index++;
                    segment_offset = 0; 
                }
            }
            
            if (xfrlen <= 0) {
                printk("iSCSI: session %p xmit_data picked xfrlen of 0, sc->use_sg %d, bytes_to_fill %d\n",
                       session, sc->use_sg, bytes_to_fill);
                iscsi_drop_session(session);
                goto done;
            }
        } 
        else {
            /* no scatter-gather */
            if ((sc->request_buffer + data_offset + bytes_to_fill) <= (sc->request_buffer + sc->request_bufflen)) {
                /* send all the data */
                session->TxIov[1].iov_base = sc->request_buffer + data_offset;
                session->TxIov[1].iov_len = xfrlen = bytes_to_fill;
                iovn = 2;
            }
            else if ((sc->request_buffer + data_offset) < (sc->request_buffer + sc->request_bufflen)) {
                /* send some data, but can't send all requested */
                xfrlen = sc->request_bufflen - data_offset;
                printk("iSCSI: xmit_data ran out of data, buffer %p len %u but offset %d length %d, sending final %d bytes\n",
                       sc->request_buffer, sc->request_bufflen, data_offset, bytes_to_fill, xfrlen);
                session->TxIov[1].iov_base = sc->request_buffer + data_offset;
                session->TxIov[1].iov_len = xfrlen;
                iovn = 2;
                stdh.final = 1;
                remain = xfrlen;
            }
            else {
                /* can't send any data */
                printk("iSCSI: xmit_data ran out of data, buffer %p len %u but offset %d length %d, sending no more data\n",
                       sc->request_buffer, sc->request_bufflen, data_offset, bytes_to_fill);
                goto done;
            }
        }

        if (pad_bytes) {
            session->TxIov[iovn].iov_base = padding;
            session->TxIov[iovn].iov_len = pad_bytes;
            iovn++;
        }
        
        hton24(stdh.dlength, xfrlen);
        
        memset( &msg, 0, sizeof(msg) );
        msg.msg_iov = &session->TxIov[0];
        msg.msg_iovlen = iovn;
        /* msg.msg_flags = MSG_DONTWAIT; */

        ISCSI_TRACE(ISCSI_TRACE_TxDataPDU, sc, task, data_offset, xfrlen);

        /* FIXME: can we really hold the largest possible size we might need to send in an int? */
        wlen = sizeof(stdh) + xfrlen + pad_bytes;
        rc = iscsi_sendmsg( session, &msg, wlen );
        if ( rc != wlen ) {
            printk("iSCSI: session %p xmit_data failed to send %d bytes, rc %d\n", session, wlen, rc);
            iscsi_drop_session(session);
            goto done;
        }
        
        remain -= xfrlen;

        DEBUG_FLOW5("iSCSI: xmit_data sent %d @ %u for itt %u, remaining %d, final %d\n",
                    xfrlen, data_offset, task->itt, remain, stdh.final);

        data_offset += xfrlen;

        if (first_sg) {
            /* undo any temporary mappings */
            for (sg = first_sg; sg <= last_sg; sg++) {
                kunmap_sg(sg);
            }
            first_sg = last_sg = NULL;
        } 
    }

 done:
    if (first_sg) {
        /* undo any temporary mappings */
        for (sg = first_sg; sg <= last_sg; sg++) {
            kunmap_sg(sg);
        }
    } 
}

static void iscsi_xmit_r2t_data(iscsi_session_t *session)
{
    iscsi_task_t *task;
    uint32_t itt;
    uint32_t ttt;
    uint32_t offset;
    uint32_t length;
    
    spin_lock(&session->task_lock);
    while ((task = pop_task(&session->tx_tasks))) {
        itt = task->itt;

        /* save the values that get set when we receive an R2T from the target, since
         * they may get clobbered after we put the task back in rx_tasks.
         */
        ttt = task->ttt;
        offset = task->data_offset;
        length = task->data_length;

        atomic_inc(&task->refcount);
        add_task(&session->rx_tasks, task);
        spin_unlock(&session->task_lock);
        
        /* implied initial R2T */
        if (test_and_clear_bit(ISCSI_TASK_INITIAL_R2T, &task->flags)) {
            uint32_t implied_length;
            
            if (session->FirstBurstLength)
                implied_length = MIN(session->FirstBurstLength, iscsi_expected_data_length(task->scsi_cmnd));
            else
                implied_length = iscsi_expected_data_length(task->scsi_cmnd); /* 0 means no limit */

            /* FIXME: if we do ImmediateData, we'll have to subtract it off as well */
            
            DEBUG_FLOW5("iSCSI: session %p sending implied initial R2T data (%u @ 0) for itt %u, task %p to %s\n", 
                        session, implied_length, itt, task, session->log_name);
            
            if (implied_length)
                iscsi_xmit_data(task, RSVD_TASK_TAG, 0U, implied_length);
        }

        if (signal_pending(current)) {
            atomic_dec(&task->refcount);
            return;
        }

        /* normal R2T from the target */
        if (ttt != RSVD_TASK_TAG) {
            DEBUG_FLOW7("iSCSI: session %p sending R2T data (%u @ %u) for itt %u, ttt %u, task %p to %s\n", 
                        session, length, offset, itt, ntohl(ttt), task, session->log_name);
            iscsi_xmit_data(task, ttt, offset, length);
        }

        atomic_dec(&task->refcount);
        
        if (signal_pending(current))
            return;

        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}


/* send a reply to a nop that requested one */
static void
iscsi_xmit_nop_reply(iscsi_session_t *session, iscsi_nop_info_t *nop_info)
{
    struct IscsiNopOutHdr stnoh;
    struct msghdr msg;
    struct iovec iov[3];
    int rc;
    int pad[4];
    int length = sizeof(stnoh);

    memset( &stnoh, 0, sizeof(stnoh) );
    stnoh.opcode = ISCSI_OP_NOOP_OUT | ISCSI_OP_IMMEDIATE;
    stnoh.itt  = RSVD_TASK_TAG;
    stnoh.ttt  = nop_info->ttt;
    stnoh.poll = 1; /* always set in draft 8 */
    memcpy(stnoh.lun, nop_info->lun, sizeof(stnoh.lun));
    hton24(stnoh.dlength, nop_info->dlength);
    stnoh.cmdsn  = htonl(session->CmdSn); /* don't increment after immediate cmds */
    stnoh.expstatsn = htonl(session->ExpStatSn);

    iov[0].iov_base = &stnoh;
    iov[0].iov_len = sizeof(stnoh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = iov;
    if (nop_info->dlength) {
        /* PDU header */
        /* data */
        iov[1].iov_base = nop_info->data;
        iov[1].iov_len = nop_info->dlength;
        /* pad */
        if (nop_info->dlength % PAD_WORD_LEN) {
            memset(pad, 0, sizeof(pad));
            msg.msg_iovlen = 3;
            iov[2].iov_base = pad;
            iov[2].iov_len = PAD_WORD_LEN - (nop_info->dlength % PAD_WORD_LEN);
            length = sizeof(stnoh) + nop_info->dlength + iov[2].iov_len;
        }
        else {
            msg.msg_iovlen = 2;
            length = sizeof(stnoh) + nop_info->dlength;
        }
    }
    else {
        /* just a PDU header */
        msg.msg_iovlen = 1;
        length = sizeof(stnoh);
    }
    
    rc = iscsi_sendmsg( session, &msg, length);
    if ( rc != length ) {
        DEBUG_ERR2("iSCSI: xmit_nop %d failed, rc %d\n", length, rc);
        iscsi_drop_session(session);
    }
    
    ISCSI_TRACE( ISCSI_TRACE_TxNopReply, NULL, NULL, nop_info->ttt, nop_info->dlength);
}

/* send replies for NopIns that requested them */
static void
iscsi_xmit_nop_replys(iscsi_session_t *session)
{
    iscsi_nop_info_t *nop_info;

    /* these aren't really tasks, but it's not worth having a separate lock for them */
    spin_lock(&session->task_lock);

    /* space for one data-less reply is preallocated in the session itself */
    if (session->nop_reply.ttt != RSVD_TASK_TAG) {
        spin_unlock(&session->task_lock);

        iscsi_xmit_nop_reply(session, &session->nop_reply);
        session->nop_reply.ttt = RSVD_TASK_TAG;

        spin_lock(&session->task_lock);
    }

    /* if we get multiple reply requests, or they have data, they'll get queued up */
    while ((nop_info = session->nop_reply_head)) {
        session->nop_reply_head = nop_info->next;
        if (!session->nop_reply_head)
            session->nop_reply_tail = NULL;
        spin_unlock(&session->task_lock);

        iscsi_xmit_nop_reply(session, nop_info);
        kfree(nop_info);
        DEBUG_ALLOC1("iSCSI: kfree nop_info %p after sending nop reply\n", nop_info);

        if (signal_pending(current))
            return;
        
        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}

static int
iscsi_xmit_abort(iscsi_task_t *task, uint32_t mgmt_itt)
{
    struct msghdr msg;
    struct iovec iov;
    int rc;
    iscsi_session_t *session;
    struct IscsiScsiTaskMgtHdr ststmh;
    Scsi_Cmnd *sc = task->scsi_cmnd;

    session = task->session;
    if ( ! session ) {
        printk("iSCSI: no session for task %p, command %p, can't abort\n", task, sc);
        return 0;
    }

    memset( &ststmh, 0, sizeof(ststmh) );
    ststmh.opcode = ISCSI_OP_SCSI_TASK_MGT_MSG;
    /* Flag it as an Immediate CMD */
    ststmh.opcode |= ISCSI_OP_IMMEDIATE;

    /* allocate an itt.  The reply will only have the rtt
     * if the task was not found, so we need to be able to find
     * the task being aborted based on the abort's itt.
     */
    task->mgmt_itt = mgmt_itt;
    ststmh.itt = htonl(task->mgmt_itt);
    ststmh.rtt = htonl(task->itt);

    ststmh.lun[1] = sc->lun;
    ststmh.function = ISCSI_TM_FUNC_ABORT_TASK;
    ststmh.cmdsn = htonl(session->CmdSn);   /* CmdSN not incremented after imm cmd */
    ststmh.expstatsn = htonl(session->ExpStatSn);

    iov.iov_base = &ststmh;
    iov.iov_len = sizeof(ststmh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    ISCSI_TRACE(ISCSI_TRACE_TxAbort, sc, task, task->mgmt_itt, 0);

    DEBUG_EH5("iSCSI: session %p sending abort mgmt %u for task %p, rtt %u at %lu\n",
              session, ntohl(ststmh.itt), task, task->itt, jiffies);

    /* indicate that we're done with the task */
    atomic_dec(&task->refcount);
    task = NULL;
    sc = NULL;

    /* send the abort */
    rc = iscsi_sendmsg( session, &msg, sizeof(ststmh) );
    if ( rc != sizeof(ststmh) ) {
        printk("iSCSI: xmit_abort failed, itt %u, rtt %u, rc %d\n", 
               ntohl(ststmh.itt), ntohl(ststmh.rtt), rc);
        iscsi_drop_session(session);
        return 0;
    }

    return 1;
}

/* send aborts for every task that needs one */
static void iscsi_xmit_aborts(iscsi_session_t *session) 
{
    iscsi_task_t *task;
    uint32_t mgmt_itt;

    spin_lock(&session->task_lock);
    while ((task = pop_task(&session->tx_abort_tasks))) {
        atomic_inc(&task->refcount);
        add_task(&session->rx_abort_tasks, task);
        mgmt_itt = allocate_itt(session);
        spin_unlock(&session->task_lock);
    
        iscsi_xmit_abort(task, mgmt_itt);
        /* xmit_abort decrements the task refcount when it's done with it */

        if (signal_pending(current))
            return;

        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}

static int
iscsi_xmit_reset(iscsi_session_t *session, uint8_t reset_type, iscsi_task_t *task, uint32_t mgmt_itt)
{
    struct msghdr msg;
    struct iovec iov;
    int rc;
    struct IscsiScsiTaskMgtHdr ststmh;
    Scsi_Cmnd *sc = NULL;

    memset( &ststmh, 0, sizeof(ststmh) );
    ststmh.opcode = ISCSI_OP_SCSI_TASK_MGT_MSG;
    /* Flag it as an Immediate CMD */
    ststmh.opcode |= ISCSI_OP_IMMEDIATE;

    /* record the itt */
    if (task)
        task->mgmt_itt = mgmt_itt;
    if (reset_type == ISCSI_TM_FUNC_TARGET_WARM_RESET)
        session->warm_reset_itt = mgmt_itt;
    if (reset_type == ISCSI_TM_FUNC_TARGET_COLD_RESET)
        session->cold_reset_itt = mgmt_itt;
    wmb();

    ststmh.itt = htonl(mgmt_itt); 
    ststmh.function = reset_type;
    ststmh.cmdsn = htonl(session->CmdSn);   /* CmdSN not incremented after imm cmd */
    ststmh.expstatsn = htonl(session->ExpStatSn);

    switch (reset_type) {
        case ISCSI_TM_FUNC_LOGICAL_UNIT_RESET:
            /* need a LUN for this */
            if (task && task->scsi_cmnd) {
                sc = task->scsi_cmnd;
                ststmh.lun[1] = sc->lun;
                DEBUG_EH7("iSCSI: sending logical unit reset to (%u %u %u %u), mgmt %u, task %p, sc %p\n", 
                          sc->host->host_no, sc->channel, sc->target, sc->lun, mgmt_itt, task, sc);
            }
            else {
                printk("iSCSI: failed to send logical unit reset, no SCSI command\n");
                return 0;
            }
            break;
        case ISCSI_TM_FUNC_TARGET_WARM_RESET:
            DEBUG_EH3("iSCSI: sending target warm reset for session %p to %s, mgmt %u\n", 
                      session, session->log_name, mgmt_itt);
            break;
        case ISCSI_TM_FUNC_TARGET_COLD_RESET:
            DEBUG_EH3("iSCSI: sending target cold reset for session %p to %s, mgmt %u\n", 
                      session, session->log_name, mgmt_itt);
            break;
        default:
            printk("iSCSI: unknown reset type %u for session %p to %s\n", 
                   reset_type, session, session->log_name);
            return 0;
            break;
    }

    iov.iov_base = &ststmh;
    iov.iov_len = sizeof(ststmh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    ISCSI_TRACE(ISCSI_TRACE_TxReset, sc, task, ntohl(ststmh.itt), reset_type);

    /* send the reset */
    rc = iscsi_sendmsg( session, &msg, sizeof(ststmh) );
    if ( rc != sizeof(ststmh) ) {
        DEBUG_ERR1("iSCSI: xmit_reset failed, rc %d\n", rc);
        iscsi_drop_session(session);
        return 0;
    }
    
    return 1;
}

static void iscsi_xmit_lun_resets(iscsi_session_t *session)
{
    iscsi_task_t *task;
    uint32_t mgmt_itt;

    spin_lock(&session->task_lock);
    while ((task = pop_task(&session->tx_lun_reset_tasks))) {
        atomic_inc(&task->refcount);
        add_task(&session->rx_lun_reset_tasks, task);
        mgmt_itt = allocate_itt(session);
        spin_unlock(&session->task_lock);

        iscsi_xmit_reset(session, ISCSI_TM_FUNC_LOGICAL_UNIT_RESET, task, mgmt_itt);
        atomic_dec(&task->refcount);

        if (signal_pending(current))
            return;

        /* relock before checking loop condition */
        spin_lock(&session->task_lock);
    }
    spin_unlock(&session->task_lock);
}


static void iscsi_xmit_logout(iscsi_session_t *session, uint32_t itt, int reason)
{
    struct IscsiLogoutHdr stlh;
    struct msghdr msg;
    struct iovec iov[3];
    int rc, wlen;
    
    memset(&stlh, 0, sizeof(stlh));
    stlh.opcode = ISCSI_OP_LOGOUT_CMD | ISCSI_OP_IMMEDIATE;
    stlh.fbit = 1;
    stlh.reason = reason;
    stlh.itt  = htonl(itt);
    stlh.cmdsn = htonl(session->CmdSn);
    stlh.expstatsn = htonl(session->ExpStatSn);

    memset(iov, 0, sizeof(iov));
    iov[0].iov_base = &stlh;
    iov[0].iov_len = sizeof(stlh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;
    wlen = sizeof(stlh);

    rc = iscsi_sendmsg( session, &msg, wlen);
    if ( rc != wlen ) {
        DEBUG_ERR3("iSCSI: session %p xmit_logout error, rc %d, wlen %d\n", session, rc, wlen);
        iscsi_drop_session(session);
    }
}

static void
iscsi_xmit_ping(iscsi_session_t *session, uint32_t itt, unsigned char *data, int length)
{
    struct IscsiNopOutHdr stph;
    struct msghdr msg;
    struct iovec iov[3];
    unsigned char pad[4];
    int rc, wlen;
    
    memset(&stph, 0, sizeof(stph));
    stph.opcode = ISCSI_OP_NOOP_OUT | ISCSI_OP_IMMEDIATE;
    stph.poll = 1;  /* draft 8 always wants this bit set now */
    stph.itt  = htonl(itt); /* reply request */
    stph.ttt  = RSVD_TASK_TAG;
    stph.cmdsn = htonl(session->CmdSn);
    stph.expstatsn = htonl(session->ExpStatSn);

    memset(iov, 0, sizeof(iov));
    iov[0].iov_base = &stph;
    iov[0].iov_len = sizeof(stph);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;
    wlen = sizeof(stph);
    
    if (data && length) {
        hton24(stph.dlength, length);

        iov[1].iov_base = data;
        iov[1].iov_len = length;
        msg.msg_iovlen = 2;
        wlen += length;

        /* may need to pad as well */
        if (length % PAD_WORD_LEN) {
            memset(pad, 0, sizeof(pad));
            msg.msg_iovlen = 3;
            iov[2].iov_base = pad;
            iov[2].iov_len = PAD_WORD_LEN - (length % PAD_WORD_LEN);
            wlen += iov[2].iov_len;
        }
        else {
            iov[2].iov_len = 0;
        }

        DEBUG_FLOW5("iSCSI: session %p tx Nop/data itt %u, lengths %d, %d, %d\n", 
                    session, itt, iov[0].iov_len, iov[1].iov_len, iov[2].iov_len);
    }
    
    ISCSI_TRACE(ISCSI_TRACE_TxPing, NULL, NULL, itt, length);

    rc = iscsi_sendmsg( session, &msg, wlen);
    if ( rc != wlen ) {
        DEBUG_ERR3("iSCSI: session %p xmit_ping error, rc %d, wlen %d\n", session, rc, wlen);
        iscsi_drop_session(session);
    }
}

/* called by the /proc code, so we can block */
static void iscsi_ping_test_session(iscsi_session_t *session, int total_data_length)
{
    unsigned char *rx_buffer;
    unsigned char *tx_buffer;

    /* assume that we can run the test, and allocate the memory we'll need.
     * draft 8 only allows 4K of ping data per Nop.
     */
    rx_buffer = kmalloc(4096, GFP_ATOMIC);
    tx_buffer = kmalloc(4096, GFP_ATOMIC);

    if (rx_buffer && tx_buffer) {
        unsigned char *data;
        unsigned int value = 0;

        /* put a simple pattern in the data */
        for (data = tx_buffer; data < tx_buffer + 4096; data++) {
            *data = value & 0xFF;
            value++;
        }
        
        spin_lock(&session->task_lock);
        if (session->ping_test_start == 0) {
            /* start a ping test */
            session->ping_test_start = jiffies;
            session->ping_test_data_length = total_data_length;
            session->ping_test_rx_length = total_data_length;
            session->ping_test_rx_start = 0;
            session->ping_test_tx_buffer = tx_buffer;
            
            wmb();
            wake_tx_thread(TX_PING_DATA, session);
            printk("iSCSI: session %p starting Nop data test with total length %u at %lu\n", session, total_data_length, jiffies);
        }
        else {
            printk("iSCSI: session %p can't start Nop data test, test started at %lu still in progress at %lu\n", 
                   session, session->ping_test_start, jiffies);
        }
        spin_unlock(&session->task_lock);

        /* the tx and rx thread will free the buffers when they're done with them,
         * so that we can just return.
         */
    }
    else {
        printk("iSCSI: session %p can't start Nop data test, couldn't allocate buffers at %lu\n", session, jiffies);
        if (rx_buffer)
            kfree(rx_buffer);
        if (tx_buffer)
            kfree(tx_buffer);
    }
}

/* the writer thread */
static int iscsi_tx_thread( void *vtaskp )
{
    iscsi_session_t *session;

    if ( ! vtaskp ) {
        DEBUG_ERR0("iSCSI: tx thread task parameter NULL\n");
        return 0;
    }

    session = (iscsi_session_t *)vtaskp;
    /* whoever created the thread already incremented the session's refcount for us */

    /* become a daemon kernel thread, and abandon any user space resources */
    sprintf(current->comm,"iscsi-tx");
    iscsi_daemonize(); 
    session->tx_pid = current->pid;
    current->flags |= PF_MEMALLOC;
    mb();

    /* check to see if iscsi_terminate_session was called before we
     * started running, since we can't get a signal from it until
     * until we set session->tx_pid.  
     */
    if (test_bit(SESSION_TERMINATING, &session->control_bits))
        goto ThreadExit;

    /* Block all signals except SIGHUP and SIGKILL */
    spin_lock_irq(&current->sig->siglock);
    siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGHUP));
    RECALC_PENDING_SIGNALS;
    spin_unlock_irq(&current->sig->siglock);                                                            

    DEBUG_INIT3("iSCSI: tx thread %d for session %p starting on cpu%d\n", current->pid, session, smp_processor_id());

    while (!test_bit(SESSION_TERMINATING, &session->control_bits)) {

        DEBUG_INIT3("iSCSI: tx thread %d for session %p waiting for new session to be established at %lu\n",
                    current->pid, session, jiffies);

        /* wait for a session to be established */
        while (!test_bit(SESSION_ESTABLISHED, &session->control_bits) || (session->socket == NULL)) {
            /* tell the rx thread that we're blocked, and that it can
             * safely call iscsi_sendmsg now as part of the Login
             * phase, since we're guaranteed not to be doing any IO
             * until the session is up.  We don't use a semaphore
             * because the counts might get off if we receive signals,
             * and loop again.  This is really just an on/off setting.  
             */
            set_bit(TX_THREAD_BLOCKED, &session->control_bits);
            wake_up(&session->tx_blocked_wait_q);
            
            /* wait for the rx thread to tell us the session is up.
             * We could use a semaphore, but we want to be able to
             * wakeup both the tx thread an ioctl call when a session
             * first comes up, so we use a condition variable
             * equivalent.  
             */
            DEBUG_INIT3("iSCSI: tx thread %d blocking on session %p at %lu\n", current->pid, session, jiffies);
            wait_event_interruptible(session->login_wait_q, test_bit(SESSION_ESTABLISHED, &session->control_bits));

            if (iscsi_handle_signals(session)) {
                printk("iSCSI: tx thread %d signalled at %lu while waiting for session %p\n", current->pid, jiffies, session);
            }
            
            if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
                /* we're all done */
                goto ThreadExit;
            }
        }

        /* we're up and running with a new session */
        clear_bit(TX_THREAD_BLOCKED, &session->control_bits);
        DEBUG_INIT4("iSCSI: tx thread %d for session %p starting to process new session with socket %p at %lu\n",
                    current->pid, session, session->socket, jiffies);

        /* make sure we start sending commands again */
        set_bit(TX_SCSI_COMMAND, &session->control_bits);
        set_bit(TX_WAKE, &session->control_bits);
        
        /* process tx requests for this session, until the session drops */
        while (!signal_pending(current)) {
            
            DEBUG_FLOW3("iSCSI: tx thread %d for session %p waiting at %lu\n", session->tx_pid, session, jiffies);
            wait_event_interruptible(session->tx_wait_q, test_and_clear_bit(TX_WAKE, &session->control_bits));
            
            DEBUG_FLOW3("iSCSI: tx thread %d for session %p is awake at %lu\n", session->tx_pid, session, jiffies);

            if (signal_pending(current)) break;
            
            /* See if we should send a ping (Nop with reply requested) */
            if (test_and_clear_bit(TX_PING, &session->control_bits)) {
                uint32_t itt;

                DEBUG_FLOW1("iSCSI: sending Nop/poll on session %p\n", session);
                /* may need locking someday.  see allocate_itt comment */
                itt = allocate_itt(session);
                iscsi_xmit_ping(session, itt, NULL, 0);
            }

            if (signal_pending(current)) break;

            /* See if we should send a ping (Nop with reply requested) containing test data*/
            if (test_and_clear_bit(TX_PING_DATA, &session->control_bits)) {
                int data_length, length;
                unsigned char *buffer;
                unsigned long tx_start, tx_stop;
                
                /* grab the total data length and buffer to use */
                spin_lock(&session->task_lock);
                length = data_length = session->ping_test_data_length;
                buffer = session->ping_test_tx_buffer;
                session->ping_test_tx_buffer = NULL;
                spin_unlock(&session->task_lock);

                tx_start = jiffies;
                while ((length > 0) && buffer) {
                    /* may need locking someday.  see allocate_itt comment */
                    uint32_t itt = allocate_itt(session);
                    
                    DEBUG_FLOW1("iSCSI: sending Nop/poll with data on session %p\n", session);
                    iscsi_xmit_ping(session, itt, buffer, MIN(4096, length));
                    
                    if (signal_pending(current)) {
                        printk("iSCSI: session %p Nop data tx failed at %lu\n", session, jiffies);
                        break;
                    }

                    length -= MIN(4096, length);
                }
                tx_stop = jiffies;
                
                if (buffer)
                    kfree(buffer);

                printk("iSCSI: session %p tx Nop data test - tx %d of %d bytes, start %lu, stop %lu, jiffies %lu, HZ %u\n",
                       session, data_length - length, data_length, tx_start, tx_stop, tx_stop - tx_start, HZ);
            }

            if (signal_pending(current)) break;
            
            /* See if we should send one or more Nops (replies requested by the target) */
            if (test_and_clear_bit(TX_NOP_REPLY, &session->control_bits)) {
                DEBUG_FLOW1("iSCSI: sending Nop replies on session %p\n", session);
                iscsi_xmit_nop_replys(session);
            }

            if (signal_pending(current)) break;
            
            /* See if we should abort any tasks */
            if (test_and_clear_bit(TX_ABORT, &session->control_bits)) {
                DEBUG_FLOW1("iSCSI: sending aborts on session %p\n", session);
                iscsi_xmit_aborts(session);
            }

            if (signal_pending(current)) break;
            
            /* See if we should reset any LUs */
            if (test_and_clear_bit(TX_LUN_RESET, &session->control_bits)) {
                DEBUG_FLOW1("iSCSI: sending logical unit resets on session %p\n", session);
                iscsi_xmit_lun_resets(session);
            }

            if (signal_pending(current)) break;
            
            /* See if we should warm reset the target */
            if (test_and_clear_bit(TX_WARM_TARGET_RESET, &session->control_bits)) {
                uint32_t itt;
                
                DEBUG_FLOW1("iSCSI: sending target warm reset to %s\n", session->log_name);
                /* may need locking someday.  see allocate_itt comment */
                itt = allocate_itt(session);
                iscsi_xmit_reset(session, ISCSI_TM_FUNC_TARGET_WARM_RESET, NULL, itt);
            }

            if (signal_pending(current)) break;

            /* See if we should cold reset the target */
            if (test_and_clear_bit(TX_COLD_TARGET_RESET, &session->control_bits)) {
                uint32_t itt;
                
                DEBUG_FLOW1("iSCSI: sending target cold reset to %s\n", session->log_name);
                /* may need locking someday.  see allocate_itt comment */
                itt = allocate_itt(session);
                iscsi_xmit_reset(session, ISCSI_TM_FUNC_TARGET_COLD_RESET, NULL, itt);
            }
            
            if (signal_pending(current)) break;
            
            /* New SCSI command received, or MaxCmdSN incremented, or task freed */
            if (test_and_clear_bit(TX_SCSI_COMMAND, &session->control_bits)) {
                /* if possible, issue new commands */
                iscsi_xmit_queued_cmnds(session);
            }

            if (signal_pending(current)) break;
            
            /* See if we need to send more data */
            if (test_and_clear_bit(TX_DATA, &session->control_bits)) {
                /* NOTE: this may call iscsi_xmit_queued_cmnds under some conditions */
                iscsi_xmit_r2t_data(session);
            }

            if (signal_pending(current)) break;

            if (test_and_clear_bit(TX_LOGOUT, &session->control_bits)) {
                uint32_t itt;

                DEBUG_INIT2("iSCSI: session %p sending Logout at %lu\n", session, jiffies);
                /* may need locking someday.  see allocate_itt comment */
                itt = allocate_itt(session);
                session->logout_itt = itt;
                wmb();
                iscsi_xmit_logout(session, itt, ISCSI_LOGOUT_REASON_CLOSE_SESSION);
            }
        }

        /* handle any signals that may have occured */
        iscsi_handle_signals(session);
    }

 ThreadExit:        
    DEBUG_INIT2("iSCSI: tx thread %d for session %p exiting\n", session->tx_pid, session);
    
    /* the rx thread may be waiting for the tx thread to block.  make it happy */
    set_bit(TX_THREAD_BLOCKED, &session->control_bits);
    wake_up(&session->tx_blocked_wait_q);

    /* we're done */
    set_current_state(TASK_RUNNING);
    session->tx_pid = 0;
    mb();
    drop_reference(session);
    
    return 0;
}

/* 
 * complete a task in the session's completing queue, and return a pointer to it,
 * or NULL if the task could not be completed.
 */
static iscsi_task_t *complete_task(iscsi_session_t *session, uint32_t itt)
{
    iscsi_task_t *task;
    unsigned long last_log = 0;
    int refcount;
    DECLARE_MIDLAYER_FLAGS;


    while (!signal_pending(current)) {
        DEBUG_QUEUE1("iSCSI: attempting to complete itt %u\n", itt);

        spin_lock(&session->task_lock);

        if ((task = remove_task(&session->completing_tasks, itt))) {
            Scsi_Cmnd *sc = task->scsi_cmnd;

            if (sc == NULL) {
                add_task(&session->completing_tasks, task);
                spin_unlock(&session->task_lock);
                DEBUG_QUEUE3("iSCSI: session %p can't complete itt %u, task %p, no SCSI cmnd\n",
                             session, itt, task);
                return NULL;
            }

            if (test_bit(ISCSI_TASK_ABORTING, &task->flags)) {
                add_task(&session->completing_tasks, task);
                spin_unlock(&session->task_lock);
                DEBUG_EH5("iSCSI: session %p can't complete itt %u, task %p, cmnd %p, aborting with mgmt_itt %u\n",
                          session, itt, task, sc, task->mgmt_itt);
                return NULL;
            }

            if (test_bit(ISCSI_HBA_IN_EH_STRATEGY, &session->hba->flags)) {
                add_task(&session->completing_tasks, task);
                spin_unlock(&session->task_lock);
                DEBUG_EH6("iSCSI: session %p can't complete itt %u, task %p, cmnd %p, HBA %p in eh_strategy at %lu\n",
                          session, itt, task, sc, session->hba, jiffies);
                return NULL;
            }

            /* it's possible the tx thread is using the task right now.
             * the task's refcount can't increase while it's in the completing
             * collection, so wait for the refcount to hit zero, or the task
             * to leave the completing collection, whichever happens first.
             */
            if ((refcount = atomic_read(&task->refcount)) == 0) {
                /* this is the expected case */
#if DEBUG_EH
                if (LOG_ENABLED(ISCSI_LOG_EH) && (sc->cmnd[0] == TEST_UNIT_READY)) {
                    printk("iSCSI: completing TUR at %lu, itt %u, task %p, command %p, (%u %u %u %u), Cmd 0x%x, result 0x%x\n", 
                           jiffies, itt, task, sc, 
                           sc->host->host_no, sc->channel, sc->target, sc->lun, 
                           sc->cmnd[0], sc->result);
                }
                else
#endif
                {
#if DEBUG_QUEUE
                    if (LOG_ENABLED(ISCSI_LOG_QUEUE))
                        printk("iSCSI: completing itt %u, task %p, command %p, (%u %u %u %u), Cmd 0x%x, done %p, result 0x%x\n", 
                               itt, task, sc, 
                               sc->host->host_no, sc->channel, sc->target, sc->lun, 
                               sc->cmnd[0], sc->scsi_done, sc->result);
#endif
                }

                /* remove the task from the session, to ensure a
                 * session drop won't try to complete the task again.
                 */
                if (remove_session_task(session, task)) {
                    DEBUG_QUEUE4("iSCSI: removed itt %u, task %p from session %p to %s\n", 
                                 task->itt, task, session, session->log_name);
                }

                /* this task no longer has a Scsi_Cmnd associated with it */
                task->scsi_cmnd = NULL;

                ISCSI_TRACE(ISCSI_TRACE_CmdDone, sc, task, sc->result, 0);

                /* tell the SCSI midlayer that the command is done */
                LOCK_MIDLAYER_LOCK(session->hba->host);
                sc->host_scribble = (void *)0xdead0001;
                CMND_ERROR_STAGE(sc) = ISCSI_ERROR_STAGE_RECOVERED;
                if (sc->scsi_done)
                    sc->scsi_done(sc);
                UNLOCK_MIDLAYER_LOCK(session->hba->host);

                spin_unlock(&session->task_lock);
#if DEBUG_SMP || DEBUG_QUEUE
                if (LOG_ENABLED(ISCSI_LOG_SMP) || LOG_ENABLED(ISCSI_LOG_QUEUE))
                    printk("iSCSI: completed itt %u, task %p, command %p, (%u %u %u %u), Cmd 0x%x, result 0x%x\n", 
                           itt, task, sc, 
                           sc->host->host_no, sc->channel, sc->target, sc->lun, 
                           sc->cmnd[0], sc->result);
#endif

                
                return task;
            }
            else {
                /* task is still in use, can't complete it yet.  Since
                 * this only happens when a command is aborted by the
                 * target unexpectedly, this error case can be slow.
                 * Just keep polling for the refcount to hit zero.  If
                 * the tx thread is blocked while using a task, the
                 * timer thread will eventually send a signal to both
                 * the rx thread and tx thread, so this loop will
                 * terminate one way or another.  
                 */
                if ((last_log == 0) || time_before_eq(last_log + HZ, jiffies)) {
                    DEBUG_QUEUE4("iSCSI: waiting to complete itt %u, task %p, cmnd %p, refcount %d\n", itt, task, sc, refcount);
                }

                push_task(&session->completing_tasks, task);
                spin_unlock(&session->task_lock);

                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(MSECS_TO_JIFFIES(10));
            }
        }
        else {
            /* an abort removed it from the completing collection. */
            DEBUG_QUEUE1("iSCSI: can't complete itt %u, task no longer completing\n", itt);
            spin_unlock(&session->task_lock);
            return NULL;
        }
    }

    return NULL;
}

/* update LUN info for /proc/scsi/iscsi 
 * Watch the inquiry data rather than TEST_UNIT_READY response,
 * since the TURs come from the target driver (sd, st, etc), and
 * that may not happen until after a SCSI device node is opened
 * and the target driver is loaded.  We get INQUIRY commands when
 * the HBA registers, and when our driver's LUN probing forces
 * them.  This is more reliable than waiting for TURs.
 */
static void process_inquiry_data(iscsi_session_t *session, Scsi_Cmnd *sc, uint8_t *data)
{
    if (data && sc) {
        /* look at the peripheral qualifier (bits 5,6, and 7 of the first byte), SPC-3 7.4.2 */
        if (LOG_ENABLED(ISCSI_LOG_INIT))
            printk("iSCSI: session %p cmd %p INQUIRY peripheral 0x%02x from (%u %u %u %u), %s\n",
                   session, sc, *data, 
                   sc->host->host_no, sc->channel, sc->target, sc->lun, session->log_name);
        
        switch ((*data & 0xE0) >> 5) {
            case 0: /* possibly a LUN */
                if (sc->lun < ISCSI_MAX_LUN) {
                    /* detected a LUN */
                    set_bit(sc->lun, session->luns_detected);
                    
                    /* unless it's one of our driver's LUN probes, the SCSI layer will activate this LUN */
                    if ((sc->scsi_done != iscsi_done) && 
                        !test_and_set_bit(sc->lun, session->luns_activated)) 
                    {
                        /* assume we found a useable LUN */
                        session->num_luns++;
                        wmb();
                    }
                }
                break;
            case 1: /* capable of supporting a physical device at this LUN, but not currently connected */
                /* as of 2002may09, the available Linux kernels treat qualifier 1 the same as qualifier 0,
                 * even though that's not really appropriate, and fills the log with a bunch of messages
                 * about unknown device types.  Map it to qualifier 3, which gets silently ignored.
                 */
                *data = 0x7F;  /* Linux only ignores a 0x7F */
                /* fall-through */
            case 2: /* reserved */
            case 3: /* not capable of supporting a physical device at this LUN */
            default:
                if (sc->lun < ISCSI_MAX_LUN) { 
                    clear_bit(sc->lun, session->luns_detected);

                    if ((sc->scsi_done != iscsi_done) && 
                        test_and_clear_bit(sc->lun, session->luns_activated)) 
                    {
                        /* there's not really a useable LUN */
                        session->num_luns--;
                        wmb();
                    }
                }
                break;
        }
    }
    else {
        printk("iSCSI: failed to process inquiry data for session %p, sc %p, data %p\n",
               session, sc, data);
    }
}


static void iscsi_recv_logout(iscsi_session_t *session, struct IscsiLogoutRspHdr *stlh)
{
    updateSN(session, ntohl(stlh->expcmdsn), ntohl(stlh->maxcmdsn));

    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    session->logout_itt = RSVD_TASK_TAG;
    wmb();

    if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
        switch (stlh->response) {
            case ISCSI_LOGOUT_SUCCESS:
#if 0
                /* Time2Wait field is only valid if ErrorRecoveryLevel == 2.  We're always level 0. */
                session->time2wait = (long)ntohs(stlh->t2wait) & 0x0000FFFFL;
                printk("iSCSI: session %p logged out at %lu, time2wait %ld\n", session, jiffies, session->time2wait);
#else
                /* set session's time2wait to zero?  use DefaultTime2Wait? */
                session->time2wait = 0;
                printk("iSCSI: session %p logged out at %lu\n", session, jiffies);
#endif
                wmb();
                iscsi_drop_session(session);
                break;
            case ISCSI_LOGOUT_CID_NOT_FOUND:
                printk("iSCSI: session %p logout failed, cid not found\n", session);
                iscsi_drop_session(session);
                break;
            case ISCSI_LOGOUT_RECOVERY_UNSUPPORTED:
                printk("iSCSI: session %p logout failed, connection recovery not supported\n", session);
                iscsi_drop_session(session);
                break;
            case ISCSI_LOGOUT_CLEANUP_FAILED:
                printk("iSCSI: session %p logout failed, cleanup failed\n", session);
                iscsi_drop_session(session);
                break;
            default:
                printk("iSCSI: session %p logout failed, response 0x%x\n", session, stlh->response);
                iscsi_drop_session(session);
                break;
        }
    }
    else {
        printk("iSCSI: session %p received logout response at %lu, but never sent a login request\n", session, jiffies);
        iscsi_drop_session(session);
    }
}


static int iscsi_recv_nop_data(iscsi_session_t *session, unsigned char *buffer, int data_length)
{
    /* read the nop data into the nop_info struct, and throw any pad bytes away */
    struct msghdr msg;
    int bytes_read = 0, rc = 0;
    int pad = (data_length % PAD_WORD_LEN) ? (PAD_WORD_LEN - (data_length % PAD_WORD_LEN)) : 0;
    
    while (bytes_read < data_length + pad) {
        int num_bytes = data_length + pad - bytes_read;
        
        /* data */
        session->RxIov[0].iov_base = buffer + bytes_read;
        session->RxIov[0].iov_len = data_length - bytes_read;

        /* pad */
        session->RxIov[1].iov_base = session->RxBuf;
        session->RxIov[1].iov_len = sizeof(session->RxBuf);

        memset(&msg, 0, sizeof(struct msghdr));
        msg.msg_iov = session->RxIov;
        msg.msg_iovlen = 2;
        
        rc = iscsi_recvmsg(session, &msg, num_bytes);
        if ( rc <= 0) {
            printk("iSCSI: session %p recv_nop_data failed to recv %d data bytes, rc %d\n", session, num_bytes, rc);
            iscsi_drop_session(session);
            return bytes_read;
        }
        if (signal_pending(current)) {
            return bytes_read;
        }
        
        bytes_read += rc;
    }

    return data_length;
}

static void iscsi_recv_nop(iscsi_session_t *session, struct IscsiNopInHdr *stnih)
{
    int dlength = ntoh24(stnih->dlength);

    DEBUG_FLOW2("iSCSI: recv_nop for session %p from %s\n", session, session->log_name);

    if (stnih->itt != RSVD_TASK_TAG) {
        /* FIXME: check StatSN */
        session->ExpStatSn = ntohl(stnih->statsn) + 1;
        updateSN(session, ntohl(stnih->expcmdsn), ntohl(stnih->maxcmdsn));
        /* it's a reply to one of our Nop-outs, so there was a PDU round-trip, and the connection is ok */
        session->last_rx = jiffies;
        wmb();

        ISCSI_TRACE( ISCSI_TRACE_RxPingReply, NULL, NULL, ntohl(stnih->itt), dlength);

        /* if there is ping data in the reply, check to see if it matches what we expect */
        if (dlength) {
            unsigned long rx_start, rx_stop;

            /* FIXME: make sure the dlength won't overflow the buffer */
            rx_start = jiffies;
            if (iscsi_recv_nop_data(session, session->RxBuf, dlength) != dlength) {
                return;
            }
            rx_stop = jiffies;
            
            spin_lock(&session->task_lock);
            if (session->ping_test_rx_start == 0)
                session->ping_test_rx_start = rx_start;

            session->ping_test_rx_length -= dlength;

            if (session->ping_test_rx_length <= 0) {
                /* all done */
                printk("iSCSI: session %p rx Nop data test - rx %d of %d bytes, start %lu, stop %lu, jiffies %lu, HZ %u\n",
                       session, session->ping_test_data_length - session->ping_test_rx_length, session->ping_test_data_length, 
                       session->ping_test_rx_start, rx_stop, rx_stop - session->ping_test_rx_start, HZ);
                printk("iSCSI: session %p Nop data test %d bytes, start %lu, stop %lu, jiffies %lu, HZ %u\n",
                       session, session->ping_test_data_length, 
                       session->ping_test_start, rx_stop, rx_stop - session->ping_test_start, HZ);
                session->ping_test_start = 0;
                session->ping_test_rx_start = 0;
            }
            spin_unlock(&session->task_lock);
        }
    }
    else {
        /* FIXME: check StatSN, but don't advance it */
        updateSN(session, ntohl(stnih->expcmdsn), ntohl(stnih->maxcmdsn));
    }

    /* check the ttt to decide whether to reply with a Nop-out */
    if (stnih->ttt != RSVD_TASK_TAG) {
        iscsi_nop_info_t *nop_info;
        
        ISCSI_TRACE( ISCSI_TRACE_RxNop, NULL, NULL, ntohl(stnih->itt), stnih->ttt);
        
        if (dlength == 0) {
            /* we preallocate space for one data-less nop reply in the
             * session structure, to avoid having to invoke the kernel
             * memory allocator in the common case where the target
             * has at most one outstanding data-less nop reply
             * requested at any given time.
             */
            spin_lock(&session->task_lock);
            if ((session->nop_reply.ttt == RSVD_TASK_TAG) && (session->nop_reply_head == NULL)) {
                session->nop_reply.ttt = stnih->ttt;
                memcpy(session->nop_reply.lun, stnih->lun, sizeof(session->nop_reply.lun));
                spin_unlock(&session->task_lock);
                DEBUG_FLOW2("iSCSI: preallocated nop reply for ttt %u, dlength %d\n", ntohl(stnih->ttt), dlength);
                wake_tx_thread(TX_NOP_REPLY, session);
                return;
            } 
            spin_unlock(&session->task_lock);
        }

        /* otherwise, try to allocate a nop_info struct and queue it up */
        nop_info = kmalloc(sizeof(iscsi_nop_info_t) + dlength, GFP_ATOMIC);
        if (nop_info) {
            DEBUG_ALLOC2("iSCSI: allocated nop_info %p, %u bytes\n", nop_info, sizeof(iscsi_nop_info_t) + dlength);
            nop_info->next = NULL;
            nop_info->ttt = stnih->ttt;
            memcpy(nop_info->lun, stnih->lun, sizeof(nop_info->lun));
            nop_info->dlength = dlength;

            /* try to save any data from the nop for the reply */
            if (dlength) {
                if (iscsi_recv_nop_data(session, nop_info->data, dlength) != dlength) {
                    kfree(nop_info);
                    return;
                }
            }
            
            /* queue it up */
            spin_lock(&session->task_lock);
            if (session->nop_reply_head) {
                session->nop_reply_tail->next = nop_info;
                session->nop_reply_tail = nop_info;
            }
            else {
                session->nop_reply_head = session->nop_reply_tail = nop_info;
            }
            spin_unlock(&session->task_lock);

            DEBUG_FLOW2("iSCSI: queued nop reply for ttt %u, dlength %d\n", ntohl(stnih->ttt), dlength);
            wake_tx_thread(TX_NOP_REPLY, session);
        }
        else {
            printk("iSCSI: session %p couldn't queue nop reply for ttt %u\n", session, ntohl(stnih->ttt));
        }
    }
}

static void
iscsi_recv_cmd(iscsi_session_t *session, struct IscsiScsiRspHdr *stsrh, unsigned char *xbuf )
{
    iscsi_task_t *task;
    Scsi_Cmnd *sc;
    unsigned int senselen = 0;
    unsigned int expected;
    uint32_t itt = ntohl(stsrh->itt);

    /* FIXME: check StatSN */
    session->ExpStatSn = ntohl(stsrh->statsn)+1;
    updateSN(session, ntohl(stsrh->expcmdsn), ntohl(stsrh->maxcmdsn));
    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    wmb();

    /* find the task for the itt we received */
    spin_lock(&session->task_lock);
    if ((task = remove_task(&session->rx_tasks, itt))) {
        /* task was waiting for this command response */
        DEBUG_QUEUE3("iSCSI: recv_cmd - rx_tasks has itt %u, task %p, refcount %d\n", itt, task, atomic_read(&task->refcount));
        atomic_inc(&task->refcount);
        add_task(&session->completing_tasks, task);
    }
    else if ((task = remove_task(&session->tx_tasks, itt))) {
        /* target aborted the command for some reason, even
         * though we're trying to send it more data.  
         */
        DEBUG_QUEUE3("iSCSI: recv_cmd - tx_tasks has itt %u, task %p, refcount %d\n", itt, task, atomic_read(&task->refcount));
        atomic_inc(&task->refcount);
        add_task(&session->completing_tasks, task);
    }
    else if ((task = find_task(&session->rx_abort_tasks, itt))) {
        /* mark it as aborted */
        set_bit(ISCSI_TASK_ABORTED, &task->flags);
        spin_unlock(&session->task_lock);
        DEBUG_EH4("iSCSI: session %p recv_cmd ignoring cmd response for itt %u, task %p, command %p, waiting for abort response\n",
                  session, itt, task, task->scsi_cmnd);
        return;
    }
    else if ((task = find_task(&session->tx_abort_tasks, itt))) {
        set_bit(ISCSI_TASK_ABORTED, &task->flags);
        spin_unlock(&session->task_lock);
        DEBUG_EH4("iSCSI: session %p recv_cmd - ignoring cmd response for itt %u, task %p, command %p, abort queued\n",
                  session, itt, task, task->scsi_cmnd);
        return;
    }
    spin_unlock(&session->task_lock);

    if (!task) {
        DEBUG_INIT1("iSCSI: recv_cmd - response for itt %u, but no such task\n", itt);
        return;
    }

    sc = task->scsi_cmnd;

    /* check for sense data */
    if (ntoh24(stsrh->dlength) > 1) {
        /* Sense data format per draft-08, 3.4.6.  2-byte sense length, then sense data, then iSCSI response data */
        senselen = (xbuf[0] << 8) | xbuf[1];
        if (senselen > (ntoh24(stsrh->dlength) - 2))
            senselen = (ntoh24(stsrh->dlength) - 2);
        xbuf += 2;
    }

    DEBUG_FLOW8("iSCSI: session %p recv_cmd - itt %u, task %p, cmnd %p, Cmd 0x%x, cmd_len %d, rsp dlength %d, senselen %d\n",
                session, itt, task, sc, sc->cmnd[0], sc->cmd_len, ntoh24(stsrh->dlength), senselen);

    /* decide what command status to report to the SCSI midlayer */
    if (stsrh->response) {
        /* iSCSI transport failed.  Fail the command, and fake some
         * sense data, since it's not clear whether the Linux SCSI
         * layer and target drivers are robust enough to realize a
         * failure occured unless there is sense data or some kind of
         * SCSI error indication.  DID_ERROR will cause retries, but
         * once those are exhausted, the scsi command will get
         * finished, and the code that finished the command tends to
         * just look at the sense data (if any).
         */
        memset(sc->sense_buffer, 0, sizeof(sc->sense_buffer));
        set_lun_comm_failure(sc);
        
        if (LOG_ENABLED(ISCSI_LOG_INIT) || LOG_ENABLED(ISCSI_LOG_FLOW))
            printk("iSCSI: recv_cmd %p, iSCSI response 0x%x, SCSI status 0x%x, senselen %d\n",
                   sc, stsrh->response, stsrh->cmd_status, senselen);
        
        sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stsrh->cmd_status);
    } 
    else {
        /* iSCSI transport worked, just pass along the SCSI status we received */
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(stsrh->cmd_status);
    }

    /* grab any sense data that came with the command.  It could be
     * argued that we should only do this if the SCSI status is check
     * condition.  It could also be argued that the target should only
     * send sense if the SCSI status is check condition.  If the
     * target bothered to send sense, we pass it along, since it
     * may indicate a problem, and it's safer to report a possible
     * problem than it is to assume everything is fine.
     */
    if (senselen) {
        /* fill in the Scsi_Cmnd's sense data */
        memset(sc->sense_buffer, 0, sizeof(sc->sense_buffer));
        memcpy(sc->sense_buffer, xbuf, MIN(senselen, sizeof(sc->sense_buffer)));
        
        /* if sense data logging is enabled, or it's deferred
         * sense that we're going to do something special with, 
         * or if it's an unexpected unit attention, which Linux doesn't
         * handle well, log the sense data.
         */
        if ((LOG_ENABLED(ISCSI_LOG_SENSE)) || 
            (((xbuf[0] == 0x71) || (xbuf[0] == 0xF1)) && translate_deferred_sense) ||
            ((SENSE_KEY(xbuf) == UNIT_ATTENTION) && (test_bit(SESSION_RESETTING, &session->control_bits) == 0)))
        {
            if (senselen >= 26) {
                printk("iSCSI: session %p recv_cmd %p, Cmd 0x%x, status 0x%x, senselen %d, "
                       "key %02x, ASC/ASCQ %02X/%02X, itt %u task %p to (%u %u %u %u), %s\n"
                       "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x " 
                       "%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                       session, sc, sc->cmnd[0], stsrh->cmd_status, senselen, 
                       SENSE_KEY(xbuf), ASC(xbuf), ASCQ(xbuf),
                       task->itt, task, 
                       sc->host->host_no, sc->channel, sc->target, sc->lun, session->log_name,
                       xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                       xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                       xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                       xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                       xbuf[16],xbuf[17],xbuf[18],xbuf[19],
                       xbuf[20],xbuf[21],xbuf[22],xbuf[23],
                       xbuf[24], xbuf[25]);
            }
            else if ( senselen >= 18) {
                printk("iSCSI: session %p recv_cmd %p, Cmd 0x%x, status 0x%x, senselen %d, "
                       "key %02x, ASC/ASCQ %02X/%02X, itt %u task %p to (%u %u %u %u), %s\n"
                       "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                       session, sc, sc->cmnd[0], stsrh->cmd_status, senselen, 
                       SENSE_KEY(xbuf), ASC(xbuf), ASCQ(xbuf),
                       task->itt, task, 
                       sc->host->host_no, sc->channel, sc->target, sc->lun, session->log_name,
                       xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                       xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                       xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                       xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                       xbuf[16],xbuf[17]);
            } 
            else {
                printk("iSCSI: session %p recv_cmd %p, Cmd 0x%x, status 0x%x, senselen %d, key %02x, "
                       "itt %u task %p to (%u %u %u %u), %s\n"
                       "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x\n",
                       session, sc, sc->cmnd[0], stsrh->cmd_status, senselen, SENSE_KEY(xbuf),
                       task->itt, task, 
                       sc->host->host_no, sc->channel, sc->target, sc->lun, session->log_name,
                       xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                       xbuf[4],xbuf[5],xbuf[6],xbuf[7]);
            }
        }
    }
#if FAKE_DEFERRED_ERRORS
    else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc && (senselen == 0) &&
             (sc->cmnd[0] != TEST_UNIT_READY) &&
             (stsrh->cmd_status == 0) && (stsrh->response == 0) &&
             (task->cmdsn >= FAKE_DEFERRED_ERROR_FREQUENCY) &&
             ((task->cmdsn % FAKE_DEFERRED_ERROR_FREQUENCY) == 0))
    {
        printk("iSCSI: session %p faking deferred error sense on itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n",
               session, itt, task->cmdsn, task, sc, sc->cmnd[0], sc->host->host_no, sc->channel, sc->target, sc->lun);

        /* fake a deferred error check condition for this command, indicating a target reset
         * Sense: 71000600 00000012 00000000 2903
         */
        sc->sense_buffer[0] = 0x71;
        sc->sense_buffer[2] = UNIT_ATTENTION;
        sc->sense_buffer[7] = 0x6;
        sc->sense_buffer[12] = 0x29;
        sc->sense_buffer[13] = 0x03;
        stsrh->cmd_status = 0x2;
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(0x02);
    }
#endif
#if FAKE_NO_REPORT_LUNS
    else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc && (senselen == 0) &&
             (sc->cmnd[0] == REPORT_LUNS) &&
             (stsrh->cmd_status == 0) && (stsrh->response == 0))
    {
        printk("iSCSI: session %p faking failed REPORT_LUNS itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n",
               session, itt, task->cmdsn, task, sc, sc->cmnd[0], 
               sc->host->host_no, sc->channel, sc->target, sc->lun);

        /* fake an illegal request check condition for this command.
         */
        sc->sense_buffer[0] = 0x70;
        sc->sense_buffer[2] = ILLEGAL_REQUEST;
        sc->sense_buffer[7] = 0x6;
        sc->sense_buffer[12] = 0x20; /* INVALID COMMAND OPERATION CODE */
        sc->sense_buffer[13] = 0x00;
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(0x02);
        stsrh->cmd_status = 0x2;
        stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
        stsrh->flags.underflow = 1;
    }
#endif
#if FAKE_PROBE_CHECK_CONDITIONS
    else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc && (senselen == 0) &&
             (sc->scsi_done == iscsi_done) && (sc->retries <= 1) &&
             (stsrh->cmd_status == 0) && (stsrh->response == 0))
    {
        printk("iSCSI: session %p faking failed probe itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n",
               session, itt, task->cmdsn, task, sc, sc->cmnd[0], 
               sc->host->host_no, sc->channel, sc->target, sc->lun);

        /* fake an command aborted check condition to test the recovery of probe commands */
        sc->sense_buffer[0] = 0x70;
        sc->sense_buffer[2] = NOT_READY;
        sc->sense_buffer[7] = 0x6;
        sc->sense_buffer[12] = 0x08;
        sc->sense_buffer[13] = 0x00;
        stsrh->cmd_status = 0x2;
        stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
        stsrh->flags.underflow = 1;
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(0x02);
    }
#endif
#if FAKE_PROBE_UNDERFLOW
    else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc && (senselen == 0) &&
             (sc->scsi_done == iscsi_done) && (sc->retries <= 3) &&
             (stsrh->cmd_status == 0) && (stsrh->response == 0))
    {
        printk("iSCSI: session %p faking probe underflow for itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n",
               session, itt, task->cmdsn, task, sc, sc->cmnd[0], 
               sc->host->host_no, sc->channel, sc->target, sc->lun);

        stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
        stsrh->flags.underflow = 1;
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
        sc->resid = iscsi_expected_data_length(sc);
# else
        task->rxdata = 0;
# endif
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(0x0);
    }
#endif

    /* record the (possibly fake) status in the trace */
    ISCSI_TRACE( ISCSI_TRACE_RxCmdStatus, sc, task, stsrh->cmd_status, stsrh->response);

    /* check for underflow and overflow */
    /* FIXME: should we only do this for iSCSI response == command completed at target?? */
    expected = iscsi_expected_data_length(sc);
    if (stsrh->flags.overflow || stsrh->flags.underflow || 
        ((test_bit(ISCSI_TASK_READ, &task->flags)) && (task->rxdata < expected))) 
    {
        if (LOG_ENABLED(ISCSI_LOG_QUEUE) || LOG_ENABLED(ISCSI_LOG_FLOW) || 
            (senselen && (SENSE_KEY(xbuf) == UNIT_ATTENTION))) 
        {
            /* for debugging, always log this for UNIT ATTENTION */
            printk("iSCSI: session %p task %p itt %u to (%u %u %u %u), Cmd 0x%x, %c%c %s, received %u, residual %u, expected %u\n",
                   session, task, task->itt, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], 
                   stsrh->flags.overflow ? 'O' : ' ', stsrh->flags.underflow ? 'U' : ' ',
                   stsrh->flags.overflow ? "overflow" : "underflow",
                   task->rxdata, ntohl(stsrh->residual_count), expected);
        }
            
        if (stsrh->flags.underflow) {
            ISCSI_TRACE(ISCSI_TRACE_RxUnderflow, sc, task, ntohl(stsrh->residual_count), expected);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
            sc->resid = ntohl(stsrh->residual_count);
#else
            if ( task->rxdata < sc->underflow ) {
                sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stsrh->cmd_status);
            }
#endif
        } 
        else if (stsrh->flags.overflow) {
            /* FIXME: not sure how to tell the SCSI layer of an overflow, so just give it an error */
            ISCSI_TRACE(ISCSI_TRACE_RxOverflow, sc, task, ntohl(stsrh->residual_count), expected);
            sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stsrh->cmd_status);
        }
        else if (task->rxdata < expected) {
            /* All the read data did not arrive.  This can happen without an underflow indication
             * from the target if the data is discarded by the driver, due to failed sanity checks
             * on the PDU or digest errors.
             */
            ISCSI_TRACE(ISCSI_TRACE_HostUnderflow, sc, task, task->rxdata, expected);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
            sc->resid = expected - task->rxdata;
#else
            if (task->rxdata < sc->underflow) {
                sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stsrh->cmd_status);
            }
#endif
        }
    }

    ISCSI_TRACE( ISCSI_TRACE_RxCmd, sc, task, task->rxdata, expected);

    if (senselen && ((xbuf[0] == 0x71) || (xbuf[0] == 0xF1)) && translate_deferred_sense) {
        printk("iSCSI: session %p recv_cmd %p translating deferred sense to current sense for itt %u\n",
               session, sc, task->itt);
        sc->sense_buffer[0] &= 0xFE;
    }

#if TEST_ABORTS
    if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc &&
        (sc->cmnd[0] != TEST_UNIT_READY) &&
        (stsrh->cmd_status == 0) &&
        (task->cmdsn >= ABORT_FREQUENCY) &&
        ((task->cmdsn % ABORT_FREQUENCY) >= 0) && ((task->cmdsn % ABORT_FREQUENCY) < ABORT_COUNT)) 
    {
        /* don't complete this command, so that we can test the error handling code. */
        spin_lock(&session->task_lock);
        if ((task = remove_task(&session->completing_tasks, itt))) {
            /* record the fact that we already got a command response */
            set_bit(ISCSI_TASK_ABORTED, &task->flags);
            /* make sure we can find it later */
            add_task(&session->rx_tasks, task);
            printk("iSCSI: ignoring completion of itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n",
                   itt, task->cmdsn, task, sc, sc->cmnd[0], sc->host->host_no, sc->channel, sc->target, sc->lun);
        }
        spin_unlock(&session->task_lock);
        atomic_dec(&task->refcount);
        return;
    }
#endif

#if TEST_PROBE_RECOVERY
    if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc &&
        ((sc->cmnd[0] == REPORT_LUNS) || (sc->cmnd[0] == INQUIRY)) &&
        (stsrh->cmd_status == 0) &&
        (task->cmdsn >= ABORT_FREQUENCY) &&
        ((task->cmdsn % ABORT_FREQUENCY) >= 0) && ((task->cmdsn % ABORT_FREQUENCY) < ABORT_COUNT)) 
    {
        /* don't complete this command, so that we can test the probe error handling code. */
        spin_lock(&session->task_lock);
        if ((task = remove_task(&session->completing_tasks, itt))) {
            add_task(&session->rx_tasks, task);
            printk("iSCSI: ignoring completion of itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n",
                   itt, task->cmdsn, task, sc, sc->cmnd[0], sc->host->host_no, sc->channel, sc->target, sc->lun);
        }
        spin_unlock(&session->task_lock);
        atomic_dec(&task->refcount);
        return;
    }
#endif

    /* we're done using the command and task */
    atomic_dec(&task->refcount);
    task = NULL;
    sc = NULL;

    /* now that we're done with it, try to complete it.  This may fail
     * if an abort for this command comes in while we're not holding
     * certain locks.  That's ok.  
     */
    DEBUG_FLOW2("iSCSI: session %p recv_cmd attempting to complete itt %u\n", session, itt);
    if ((task = complete_task(session, itt))) {
        spin_lock(&session->hba->free_task_lock);
        free_task(task);
        spin_unlock(&session->hba->free_task_lock);
    }
}

static void iscsi_recv_r2t(iscsi_session_t *session, struct IscsiRttHdr *strh )
{
    iscsi_task_t *task = NULL;
    uint32_t itt = ntohl(strh->itt);

    updateSN(session, ntohl(strh->expcmdsn), ntohl(strh->maxcmdsn));
    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    mb();

    spin_lock(&session->task_lock);
    if ((task = remove_task(&session->rx_tasks, itt))) {
        task->ttt = strh->ttt;
        task->data_length = ntohl(strh->data_length);
        task->data_offset = ntohl(strh->data_offset);
        wmb();
        add_task(&session->tx_tasks, task);
        ISCSI_TRACE(ISCSI_TRACE_R2T, task->scsi_cmnd, task, task->data_offset, task->data_length);
        DEBUG_FLOW5("iSCSI: session %p R2T for task %p itt %u, %u bytes @ offset %u\n",
                    session, task, ntohl(strh->itt), ntohl(strh->data_length), ntohl(strh->data_offset));
        wake_tx_thread(TX_DATA, session);
    }
    else if ((task = find_task(&session->tx_tasks, itt))) {
        if (task->ttt == RSVD_TASK_TAG) {
            /* no R2T from the target yet, it's probably queued for an implied initial R2T */
            task->ttt = strh->ttt;
            task->data_length = ntohl(strh->data_length);
            task->data_offset = ntohl(strh->data_offset);
            wmb();
            ISCSI_TRACE(ISCSI_TRACE_R2T, task->scsi_cmnd, task, task->data_offset, task->data_length);
            DEBUG_FLOW5("iSCSI: session %p R2T for task %p itt %u, %u bytes @ offset %u (as well as initial)\n",
                        session, task, ntohl(strh->itt), ntohl(strh->data_length), ntohl(strh->data_offset));
            wake_tx_thread(TX_DATA, session);
        }
        else {
            /* bug in the target.  MaxOutsandingR2T == 1 should have prevented this from occuring */
            printk("iSCSI: session %p ignoring R2T for task %p, itt %u, %u bytes @ offset %u, ttt %u, "
                   "already have R2T for %u @ %u, ttt %u\n",
                   session, task, ntohl(strh->itt), ntohl(strh->data_length), ntohl(strh->data_offset), ntohl(strh->ttt),
                   task->data_length, task->data_offset, ntohl(task->ttt));
        }
    }
    else {
        /* the task wasn't waiting for an R2T, aborting or already received an R2T */
        printk("iSCSI: session %p ignoring R2T for itt %u, %u bytes @ offset %u\n",
               session, ntohl(strh->itt), ntohl(strh->data_length), ntohl(strh->data_offset));
    }
    spin_unlock(&session->task_lock);
}


static void
iscsi_recv_data( iscsi_session_t *session, struct IscsiDataRspHdr *stdrh )
{
    iscsi_task_t *task = NULL;
    Scsi_Cmnd *sc = NULL;
    struct scatterlist *sglist = NULL, *sg, *first_sg = NULL, *last_sg = NULL;
    int dlength, remaining, rc;
    int bytes_read = 0;
    uint32_t offset;
    unsigned int iovn = 0, pad = 0;
    unsigned int segment_offset = 0;
    struct msghdr msg;
    struct iovec iov;
    uint32_t itt = ntohl(stdrh->itt);
    uint8_t *peripheral = NULL;
    
    if (stdrh->status_present) {
        /* FIXME: check StatSN */
        session->ExpStatSn = ntohl(stdrh->statsn)+1;
    }
    updateSN(session, ntohl(stdrh->expcmdsn), ntohl(stdrh->maxcmdsn));
    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    wmb();

    dlength = ntoh24( stdrh->dlength );
    offset = ntohl( stdrh->offset );
    /* Compute padding bytes that follow the data */
    pad = dlength % PAD_WORD_LEN;
    if (pad) {
        pad = PAD_WORD_LEN - pad;
    }

    spin_lock(&session->task_lock);
    if ((task = find_task(&session->rx_tasks, itt))) {
        /* receive the data, and leave it in rx_tasks for more data or a cmd response */
        atomic_inc(&task->refcount);
        sc = task->scsi_cmnd;
    }
#if TEST_DELAYED_DATA
    if (task && dlength && ((task->cmdsn % 500) == 0)) {
        printk("iSCSI: testing delayed data for task %p, itt %u, cmdsn %u, dlength %u at %lu\n", 
               task, task->itt, task->cmdsn, dlength, jiffies);
        atomic_dec(&task->refcount);
        task = NULL;
        sc = NULL;
        spin_unlock(&session->task_lock);
        session->last_rx = jiffies + (45 * HZ);
        mb();
        set_current_state(TASK_UNINTERRUPTIBLE);
        schedule_timeout(37 * HZ);
        spin_lock(&session->task_lock);
        printk("iSCSI: test of delayed data continuing at %lu\n", jiffies);
    }
    if (task) {
        /* do nothing */
    }
#endif
    else if ((task = find_task(&session->rx_abort_tasks, itt))) {
        /* discard the data */
        printk("iSCSI: session %p recv_data, itt %u, task %p aborting, discarding received data, offset %u len %u\n", 
               session, ntohl(stdrh->itt), task, offset, dlength);
        task = NULL;
    }
    else if ((task = find_task(&session->tx_abort_tasks, itt))) {
        /* discard the data */
        printk("iSCSI: session %p recv_data, itt %u, task %p will abort, discarding received data, offset %u len %u\n", 
               session, ntohl(stdrh->itt), task, offset, dlength);
        task = NULL;
    }
    else {
        printk("iSCSI: session %p recv_data, no task for itt %u (next itt %u), discarding received data, offset %u len %u\n", 
               session, ntohl(stdrh->itt), session->itt, offset, dlength);
        task = NULL;
    }
    spin_unlock(&session->task_lock);

    
    if (task) {
        if (test_bit(ISCSI_TASK_BAD_DATA, &task->flags)) {
            /* an earlier data PDU was bogus, toss this one too */
            printk("iSCSI: session %p recv_data for itt %u, task %p, cmnd %p, bufflen %u, discarding more data for offset %u len %u\n", 
                   session, itt, task, sc, sc->request_bufflen, offset, dlength);
            atomic_dec(&task->refcount);
            task = NULL;
            sc = NULL; 
        }
        else if (sc && ((offset + dlength) > sc->request_bufflen)) {
            /* all the data won't fit in the buffer */
            printk("iSCSI: session %p recv_data for itt %u, task %p, cmnd %p, bufflen %u, discarding data for offset %u len %u\n", 
                   session, itt, task, sc, sc->request_bufflen, offset, dlength);
            
            /* If Data PDUs arrive out of order, computing how much valid
             * data contiguous data there is at the beginning of the
             * buffer can get complicated.  Instead, we just say there's
             * no valid data at all if we get a bogus data PDU.
             */
            task->rxdata = 0;
            set_bit(ISCSI_TASK_BAD_DATA, &task->flags);
            atomic_dec(&task->refcount);
            task = NULL;
            sc = NULL;
        }
        else {
            /* ok, everything checked out, save the data */
            DEBUG_FLOW8("iSCSI: session %p recv_data itt %u, task %p, sc %p, use_sg %d, datasn %u, offset %u, dlength %u\n", 
                        session, itt, task, sc, sc->use_sg, ntohl(stdrh->datasn), offset, dlength);
        }
    }

    /* configure for receiving the data */
    if (sc && sc->use_sg) {
        int index;

        /* scatter-gather */
        sg = sglist = (struct scatterlist *)sc->request_buffer;
        segment_offset = offset;

        for (index = 0; index < sc->use_sg; index++) {
            if (segment_offset < sglist[index].length)
                break;
            else 
                segment_offset -= sglist[index].length;
        }

        if (index >= sc->use_sg) {
            /* didn't find the offset, toss the data and let the command underflow */
            printk("iSCSI: session %p recv_data for itt %u couldn't find offset %u in sglist %p, sc %p, bufflen %u, use_sg %u, discarding data\n",
                   session, task->itt, offset, sglist, sc, sc->request_bufflen, sc->use_sg);
            print_cmnd(sc);
            ISCSI_TRACE(ISCSI_TRACE_BadOffset, sc, task, offset, sc->request_bufflen);
            task->rxdata = 0;
            atomic_dec(&task->refcount);
            task = NULL;
            sc = NULL;
        }
        else {
            remaining = dlength;
            iovn = 0;
            
            /* setup all the data buffers */
            while (sc && (remaining > 0) && (index < sc->use_sg)) {
                sg = &sglist[index];

                if (!kmap_sg(sg)) {
                    printk("iSCSI: session %p recv_data itt %u task %p failed to map sg %p\n", 
                           session, task->itt, task, sg);
                    break;
                }
                else if (first_sg == NULL) {
                    first_sg = sg;
                }

                last_sg = sg;

                /* sanity check the sglist segment length */
                if (sg->length <= segment_offset) {
                    /* the sglist is corrupt */
                    printk("iSCSI: session %p recv_data index %d, length %u too small for offset %u, remaining %d, sglist has been corrupted\n", 
                           session, index, sg->length, segment_offset, remaining);
                    print_cmnd(sc);
                    ISCSI_TRACE(ISCSI_TRACE_BadRxSeg, sc, task, sg->length, segment_offset);
                    task->rxdata = 0; /* toss the data, and let the command complete with a total underflow */
                    atomic_dec(&task->refcount);
                    task = NULL;
                    sc = NULL;
                    goto toss_data;
                }

                session->RxIov[iovn].iov_base = sg_virtual_address(sg) + segment_offset;
                session->RxIov[iovn].iov_len  = MIN(remaining, sg->length - segment_offset);
                remaining -= session->RxIov[iovn].iov_len;
                
                DEBUG_FLOW7("iSCSI: recv_data itt %u, iov[%2d] = sg[%2d] = %p, %u of %u bytes, remaining %u\n", 
                            itt, iovn, sg - sglist, session->RxIov[iovn].iov_base, session->RxIov[iovn].iov_len, 
                            sg->length, remaining);
                index++;
                iovn++;
                segment_offset = 0;
            }
            
            if (remaining != 0) {
                /* we ran out of buffer space with more data remaining.
                 * this should never happen if the Scsi_Cmnd's bufflen
                 * matches the combined length of the sglist segments.
                 */
                printk("iSCSI: session %p recv_data for cmnd %p, bufflen %u, offset %u len %u, remaining data %u, discarding all data\n",
                       session, sc, sc->request_bufflen, offset, dlength, remaining);
                print_cmnd(sc);
                atomic_dec(&task->refcount);
                task = NULL;
                sc = NULL;
            }
            else if (pad) {
                session->RxIov[iovn].iov_base = session->RxBuf;
                session->RxIov[iovn].iov_len  = pad;
                DEBUG_FLOW6("iSCSI: recv_data itt %u, iov[%2d] = sg[%2d] = %p, %u of %u, pad data\n", 
                            itt, iovn, sg - sglist, session->RxIov[iovn].iov_base,
                            session->RxIov[iovn].iov_len, sg->length);
                iovn++;
            }
        }
    }
    else if (sc) {
        /* no scatter-gather, just read it into the buffer */
        session->RxIov[0].iov_base = sc->request_buffer + offset;
        session->RxIov[0].iov_len  = dlength;
        if (pad) {
            session->RxIov[1].iov_base = session->RxBuf;
            session->RxIov[1].iov_len  = pad;
            iovn = 2;
        }
        else
            iovn = 1;
    }

    /* just throw away the PDU */
    if (!sc) {        
        bytes_read = 0;

    toss_data:
        if (first_sg) {
            /* undo any temporary mappings */
            for (sg = first_sg; sg <= last_sg; sg++) {
                kunmap_sg(sg);
            }
        } 

        while (!signal_pending(current) && (bytes_read < dlength + pad)) {
            int num_bytes = MIN(dlength + pad - bytes_read, sizeof(session->RxBuf));

            iov.iov_base = session->RxBuf;
            iov.iov_len = sizeof(session->RxBuf);
            memset( &msg, 0, sizeof(struct msghdr) );
            msg.msg_iov = &iov;
            msg.msg_iovlen = 1;
            rc = iscsi_recvmsg(session, &msg, num_bytes);
            if ( rc <= 0) {
                DEBUG_ERR3("iSCSI: session %p recv_data failed to recv %d data PDU bytes, rc %d\n", session, num_bytes, rc);
                iscsi_drop_session(session);
            }
            else {
                /* assume a PDU round-trip, connection is ok */
                session->last_rx = jiffies;
                mb();
            }
            bytes_read += rc;
        }

        if (task) {
            atomic_dec(&task->refcount); 
            task = NULL; 
        }
        return;
    }

    /* save the address of the first byte of INQUIRY data */
    if ((sc->cmnd[0] == INQUIRY) && (offset == 0) && (dlength > 0))
        peripheral = session->RxIov[0].iov_base;  

    /* accept the data */
    memset( &msg, 0, sizeof(struct msghdr) );
    msg.msg_iov = session->RxIov;
    msg.msg_iovlen = iovn;

    DEBUG_FLOW4("iSCSI: recv_data itt %u calling recvmsg %d bytes, iovn %u, RxIov[0].base = %p\n", 
                itt, dlength + pad, iovn, session->RxIov[0].iov_base);
    rc = iscsi_recvmsg( session, &msg, dlength + pad);
    if (rc == dlength + pad) {
        /* if the data arrives out-of-order, it becomes much harder
         * for us to correctly calculate the residual if we don't get
         * enough data and also don't get an underflow from the
         * target.  This can happen if we discard Data PDUs due to
         * bogus offsets/lengths, or digests.  Since we always negotiate 
         * for Data PDUs in-order, this should never happen, but check 
         * for it anyway.
         */
        if (task->rxdata == offset)
            task->rxdata += dlength;
        else
            printk("iSCSI: session %p received out-of-order data, expected offset %u, received offset %u\n",
                   session, task->rxdata, offset);

        /* assume a PDU round-trip, connection is ok */
        session->last_rx = jiffies;
        wmb();
    }
    else if ( rc == dlength) {
        task->rxdata += dlength;
        /* FIXME: until iscsi_recvmsg handles the retries, allow the pad
         * to come in separately.
         */
        iov.iov_base = session->RxBuf;
        iov.iov_len = pad;
        memset( &msg, 0, sizeof(struct msghdr) );
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        rc = iscsi_recvmsg( session, &msg, pad );
        if ( rc != pad ) {
            DEBUG_ERR3("iSCSI: session %p recv_data failed to recv %d pad bytes, rc %d\n", session, pad, rc);
            atomic_dec(&task->refcount);
            iscsi_drop_session(session);
            return;
        }
        else {
            /* assume a PDU round-trip, connection is ok */
            session->last_rx = jiffies;
            wmb();
        }
    }
    else {
        DEBUG_ERR3("iSCSI: session %p recv_data failed to recv %d data PDU bytes, rc %d\n", session, dlength, rc);
        atomic_dec(&task->refcount);
        iscsi_drop_session(session);
        return;
    }

    /* update LUN info based on the INQUIRY data, since we've got it mapped now */
    if (peripheral)
        process_inquiry_data(session, task->scsi_cmnd, peripheral);

    /* done with the data buffers */
    if (first_sg) {
        /* undo any temporary mappings */
        for (sg = first_sg; sg <= last_sg; sg++) {
            kunmap_sg(sg);
        }
    } 
    
    ISCSI_TRACE(ISCSI_TRACE_RxData, sc, task, offset, dlength);

    if (stdrh->status_present) {
        unsigned int expected = iscsi_expected_data_length(sc);

        /* we got status, meaning the command completed in a way that
         * doesn't give us any sense data, and the command must be
         * completed now, since we won't get a command response PDU.
         */
        DEBUG_FLOW4("iSCSI: Data-in with status 0x%x for itt %u, task %p, sc %p\n", 
                    stdrh->cmd_status, ntohl(stdrh->itt), task, task->scsi_cmnd);
        ISCSI_TRACE( ISCSI_TRACE_RxDataCmdStatus, sc, task, stdrh->cmd_status, 0);
        sc->result = HOST_BYTE(DID_OK) | STATUS_BYTE(stdrh->cmd_status);

        
        if (stdrh->overflow || stdrh->underflow || 
            ((test_bit(ISCSI_TASK_READ, &task->flags)) && (task->rxdata != expected))) 
        {
            if (LOG_ENABLED(ISCSI_LOG_QUEUE) || LOG_ENABLED(ISCSI_LOG_FLOW)) {
                printk("iSCSI: session %p task %p itt %u to (%u %u %u %u), Cmd 0x%x, %c%c %s, received %u, residual %u, expected %u\n",
                       session, task, task->itt, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], 
                       stdrh->overflow ? 'O' : ' ', stdrh->underflow ? 'U' : ' ',
                       stdrh->overflow ? "overflow" : "underflow",
                       task->rxdata, ntohl(stdrh->residual_count), expected);
            }

            if ( stdrh->underflow ) {
                ISCSI_TRACE(ISCSI_TRACE_RxUnderflow, sc, task, ntohl(stdrh->residual_count), expected);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
                sc->resid = ntohl(stdrh->residual_count);
#else
                if ( task->rxdata < sc->underflow ) {
                    sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stdrh->cmd_status);
                }
#endif
            } 
            else if (stdrh->overflow) {
                /* FIXME: not sure how to tell the SCSI layer of an overflow, so just give it an error */
                ISCSI_TRACE(ISCSI_TRACE_RxOverflow, sc, task, ntohl(stdrh->residual_count), expected);
                sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stdrh->cmd_status);
            }
            else {
                /* All the read data did not arrive */
                ISCSI_TRACE(ISCSI_TRACE_HostUnderflow, sc, task, task->rxdata, expected);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,19)
                /* we don't know which parts of the buffer didn't get data, so report the whole buffer missing */
                sc->resid = expected;
#else
                sc->result = HOST_BYTE(DID_ERROR) | STATUS_BYTE(stdrh->cmd_status);
#endif
            }
        }
        
#if TEST_ABORTS
        if (test_bit(SESSION_ESTABLISHED, &session->control_bits) &&
            (sc->cmnd[0] != TEST_UNIT_READY) &&
            (stdrh->cmd_status == 0) &&
            (task->cmdsn >= ABORT_FREQUENCY) &&
            ((task->cmdsn % ABORT_FREQUENCY) >= 0) && ((task->cmdsn % ABORT_FREQUENCY) <= ABORT_COUNT))
        {
            /* don't complete this command, so that we can test the error handling
             * code.
             */
            printk("iSCSI: ignoring completion of itt %u, CmdSN %u, task %p, sc %p, Cmd 0x%x to (%u %u %u %u)\n", 
                   itt, task->cmdsn, task, sc, sc->cmnd[0], sc->host->host_no, sc->channel, sc->target, sc->lun);
            atomic_dec(&task->refcount);
            return;
        }
#endif

        /* done using the command and task */
        atomic_dec(&task->refcount);

        /* try to complete the task */
        spin_lock(&session->task_lock);
        if ((task = remove_task(&session->rx_tasks, itt))) {
            add_task(&session->completing_tasks, task);
            spin_unlock(&session->task_lock);

            if ((task = complete_task(session, itt))) {
                spin_lock(&session->hba->free_task_lock);
                free_task(task);
                spin_unlock(&session->hba->free_task_lock);
            }
            else {
                DEBUG_FLOW1("iSCSI: Data-in with status for itt %u, but task couldn't be completed\n", 
                            ntohl(stdrh->itt));
            }
        }
        else {
            /* task was probably aborted */
            spin_unlock(&session->task_lock);
            DEBUG_FLOW1("iSCSI: Data-in with status for itt %u, but task aborted\n", ntohl(stdrh->itt));
        }
    }
    else {
        /* done modifying the command and task */
        atomic_dec(&task->refcount);
    }
}

static void iscsi_recv_mgmt_rsp(iscsi_session_t *session, struct IscsiScsiTaskMgtRspHdr *ststmrh )
{
    iscsi_task_t *task = NULL;
    uint32_t mgmt_itt = ntohl(ststmrh->itt);

    /* FIXME: check StatSN */
    session->ExpStatSn = ntohl(ststmrh->statsn)+1;
    updateSN(session, ntohl(ststmrh->expcmdsn), ntohl(ststmrh->maxcmdsn));
    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    wmb();

    spin_lock(&session->task_lock);
    if (mgmt_itt == session->warm_reset_itt) {
        if ( ststmrh->response ) {
            DEBUG_EH6("iSCSI: session %p warm target reset rejected (0x%x) for mgmt %u, rtt %u, session %p to %s\n", 
                      session, ststmrh->response, ntohl(ststmrh->itt), ntohl(ststmrh->rtt), session, session->log_name);
        }
#if FAKE_PDU_REJECTS        
        else if (mgmt_itt % 3) {
            printk("iSCSI: session %p testing PDU reject for warm target reset mgmt %u\n",
                   session, mgmt_itt);
        }
#endif
        else {
            set_bit(SESSION_RESET, &session->control_bits);
            printk("iSCSI: session %p warm target reset success for mgmt %u, session %p to %s\n",
                   session, session->warm_reset_itt, session, session->log_name);
        }

        session->warm_reset_itt = RSVD_TASK_TAG;
        ISCSI_TRACE(ISCSI_TRACE_RxReset, NULL, task, mgmt_itt, ststmrh->response);
        wake_up(&session->reset_wait_q);
    }
    else if ((task = remove_mgmt_task(&session->rx_abort_tasks, mgmt_itt))) {
        /* We don't really care if the target actually aborted the
         * command, we just want it to be done so that we can free the
         * task before returning from eh_abort.
         */
#if FAKE_PDU_REJECTS        
        if (mgmt_itt % 3) {
            set_bit(ISCSI_TASK_REJECTED, &task->flags);
            add_task(&session->rx_abort_tasks, task);
            printk("iSCSI: session %p testing PDU reject for task %p, mgmt %u\n",
                   session, task, mgmt_itt);
        }
        else 
#endif
        {
            if (ststmrh->response == 0) {
                printk("iSCSI: session %p abort success for itt %u, mgmt %u, task %p, cmnd %p, Cmd 0x%x\n",
                       session, task->itt, ntohl(ststmrh->itt), task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
                add_task(&session->aborted_tasks, task);
            }
            else if (test_bit(ISCSI_TASK_ABORTED, &task->flags)) {
                /* completed before the abort could be delivered, just treat it as aborted */
                printk("iSCSI: session %p abort success due to completion for itt %u, mgmt %u, task %p, cmnd %p, Cmd 0x%x\n",
                       session, task->itt,  ntohl(ststmrh->itt), task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
                add_task(&session->aborted_tasks, task);
            }
            else {
                DEBUG_EH7("iSCSI: session %p abort rejected (0x%x) for itt %u, mgmt %u, task %p, cmnd %p, Cmd 0x%x\n", 
                          session, ststmrh->response, task->itt, ntohl(ststmrh->itt),
                          task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
                /* put it back in rx_abort_tasks so that it can be found by later resets */
                add_task(&session->rx_abort_tasks, task);
            }
        }
        ISCSI_TRACE(ISCSI_TRACE_RxAbort, task->scsi_cmnd, task, mgmt_itt, ststmrh->response);
        wake_up(&task->wait_q);
    }        
    else if ((task = remove_mgmt_task(&session->rx_lun_reset_tasks, mgmt_itt))) {
        if ( ststmrh->response ) {
            DEBUG_EH7("iSCSI: session %p LUN reset rejected (0x%x) for itt %u, rtt %u, task %p, cmnd %p, Cmd 0x%x\n", 
                      session, ststmrh->response, ntohl(ststmrh->itt), ntohl(ststmrh->rtt), 
                      task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
            /* put it back in rx tasks.  The error handler will find it there any retry it */
            add_task(&session->rx_lun_reset_tasks, task);
        }
#if FAKE_PDU_REJECTS        
        else if (mgmt_itt % 3) {
            set_bit(ISCSI_TASK_REJECTED, &task->flags);
            add_task(&session->rx_lun_reset_tasks, task);
            printk("iSCSI: session %p testing PDU reject for LUN reset task %p, itt %u\n",
                   session, task, mgmt_itt);
        }
#endif
        else {
            add_task(&session->lun_reset_tasks, task);
            printk("iSCSI: session %p LUN reset success for itt %u, task %p, cmnd %p, Cmd 0x%x\n",
                   session, ntohl(ststmrh->itt), task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
        }
        ISCSI_TRACE(ISCSI_TRACE_RxReset, NULL, task, mgmt_itt, ststmrh->response);
        wake_up(&task->wait_q);
    }
    else {
        printk("iSCSI: session %p mgmt response 0x%x for unknown itt %u, rtt %u\n", 
               session, ststmrh->response, ntohl(ststmrh->itt), ntohl(ststmrh->rtt));
    }
    spin_unlock(&session->task_lock);
}


static void iscsi_recv_reject(iscsi_session_t *session, struct IscsiRejectRspHdr *reject, unsigned char *xbuf)
{
    int dlength = ntoh24(reject->dlength);
    uint32_t itt = 0;
    iscsi_task_t *task = NULL;
    struct IscsiHdr pdu;

    /* FIXME: check StatSN */
    session->ExpStatSn = ntohl(reject->statsn)+1;
    updateSN(session, ntohl(reject->expcmdsn), ntohl(reject->maxcmdsn));
    /* assume a PDU round-trip, connection is ok */
    session->last_rx = jiffies;
    mb();

    if (reject->reason == IMM_CMD_REJECT) {
        if (dlength >= sizeof(pdu)) {
            /* look at the rejected PDU */
            memcpy(&pdu, xbuf, sizeof(pdu));
            itt = ntohl(pdu.itt);

            /* try to find the task corresponding to this itt, and wake up any process waiting on it */
            spin_lock(&session->task_lock);
            if (itt == session->warm_reset_itt) {
                DEBUG_EH2("iSCSI: session %p warm reset PDU rejected, itt %u\n", session, session->warm_reset_itt);
                session->warm_reset_itt = RSVD_TASK_TAG;
                wmb();
                wake_up(&session->reset_wait_q);
            }
            else if (itt == session->cold_reset_itt) {
                DEBUG_EH2("iSCSI: session %p cold reset PDU rejected, itt %u\n", session, session->cold_reset_itt);
                session->cold_reset_itt = RSVD_TASK_TAG;
                wmb();
                wake_up(&session->reset_wait_q);
            }
            else if ((task = find_mgmt_task(&session->rx_abort_tasks, itt))) {
                DEBUG_EH5("iSCSI: session %p abort PDU rejected, itt %u, task %p, cmnd %p, Cmd 0x%x\n",
                          session, itt, task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
                set_bit(ISCSI_TASK_REJECTED, &task->flags);
                wake_up(&task->wait_q);
            }
            else if ((task = find_mgmt_task(&session->rx_lun_reset_tasks, itt))) {
                DEBUG_EH5("iSCSI: session %p LUN reset PDU rejected, itt %u, task %p, cmnd %p, Cmd 0x%x\n",
                          session, itt, task, task->scsi_cmnd, task->scsi_cmnd->cmnd[0]);
                set_bit(ISCSI_TASK_REJECTED, &task->flags);
                wake_up(&task->wait_q);
            }
            else if ((pdu.opcode & ISCSI_OPCODE_MASK) == ISCSI_OP_LOGOUT_CMD) {
                /* our Logout was rejected.  just let the logout response timer drop the session */
                printk("iSCSI: session %p login PDU rejected, itt %u\n", session, itt);
                session->logout_itt = RSVD_TASK_TAG;
                wmb();
            }
            else {
                printk("iSCSI: session %p, itt %u immediate command rejected at %lu\n", 
                       session, itt, jiffies);
            }
            spin_unlock(&session->task_lock);
        }
        else {
            printk("iSCSI: session %p, immediate command rejected at %lu, dlength %u\n", 
                   session, jiffies, dlength);
        }
    }
    else {
        if (dlength >= sizeof(pdu)) {
            /* look at the rejected PDU */
            memcpy(&pdu, xbuf, sizeof(pdu));
            itt = ntohl(pdu.itt);
            printk("iSCSI: dropping session %p because target rejected a PDU, reason 0x%x, dlength %d, rejected itt %u, opcode 0x%x\n",
                   session, reject->reason, dlength, itt, pdu.opcode);
        }
        else {
            printk("iSCSI: dropping session %p because target rejected a PDU, reason 0x%x, dlength %u\n",
                   session, reject->reason, dlength);
        }
        iscsi_drop_session(session);
    }
}


static void iscsi_recv_async_event(iscsi_session_t *session, struct IscsiAsyncEvtHdr *staeh, unsigned char *xbuf)
{
    unsigned int senselen = ntoh24(staeh->dlength);

    /* FIXME: check StatSN */
    session->ExpStatSn = ntohl(staeh->statsn)+1;
    updateSN(session, ntohl(staeh->expcmdsn), ntohl(staeh->maxcmdsn));

    ISCSI_TRACE(ISCSI_TRACE_RxAsyncEvent, NULL, NULL, staeh->async_event, staeh->async_vcode);

    switch (staeh->async_event) {
        case ASYNC_EVENT_SCSI_EVENT:
            /* no way to pass this up to the SCSI layer, since there is no command associated with it */
            if (LOG_ENABLED(ISCSI_LOG_SENSE)) {
                if (senselen >= 26) {
                    printk("iSCSI: SCSI Async event, senselen %d, key %02x, ASC/ASCQ %02X/%02X, session %p to %s\n"
                           "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x " 
                           "%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                           senselen, SENSE_KEY(xbuf), ASC(xbuf), ASCQ(xbuf), session, session->log_name,
                           xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                           xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                           xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                           xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                           xbuf[16],xbuf[17],xbuf[18],xbuf[19],
                           xbuf[20],xbuf[21],xbuf[22],xbuf[23],
                           xbuf[24], xbuf[25]);
                }
                else if ( senselen >= 18) {
                    printk("iSCSI: SCSI Async event, senselen %d, key %02x, ASC/ASCQ %02X/%02X, session %p to %s\n"
                           "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                           senselen, SENSE_KEY(xbuf), ASC(xbuf), ASCQ(xbuf), session, session->log_name,
                           xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                           xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                           xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                           xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                           xbuf[16],xbuf[17]);
                } 
                else if ( senselen >= 14) {
                    printk("iSCSI: SCSI Async event, senselen %d, key %02x, ASC/ASCQ %02X/%02X, session %p to %s\n"
                           "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                           senselen, SENSE_KEY(xbuf), ASC(xbuf), ASCQ(xbuf), session, session->log_name,
                           xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                           xbuf[4],xbuf[5],xbuf[6],xbuf[7],
                           xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                           xbuf[12],xbuf[13]);
                } 
                else {
                    printk("iSCSI: SCSI Async event, senselen %d, key %02x, session %p to %s\n"
                           "iSCSI: Sense %02x%02x%02x%02x %02x%02x%02x%02x\n",
                           senselen, SENSE_KEY(xbuf), session, session->log_name,
                           xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                           xbuf[4],xbuf[5],xbuf[6],xbuf[7]);
                }
            }
            break;
        case ASYNC_EVENT_REQUEST_LOGOUT:
            DEBUG_ERR2("iSCSI: target requests logout within %u seconds for session to %s\n", 
                       ntohs(staeh->param3), session->log_name);
            /* FIXME: this is really a request to drop a connection, not the whole session,
             * but we currently only have one connection per session, so there's no difference
             * at the moment.
             */

            /* we need to get the task lock to make sure the TX thread isn't in the middle
             * of adding another task to the session.
             */
            spin_lock(&session->task_lock);
            iscsi_request_logout(session, ntohs(staeh->param3) - (HZ / 10), session->active_timeout);
            spin_unlock(&session->task_lock);
            break;
        case ASYNC_EVENT_DROPPING_CONNECTION:
            printk("iSCSI: session %p target dropping connection %u, reconnect min %u max %u\n", 
                   session, ntohs(staeh->param1), ntohs(staeh->param2), ntohs(staeh->param3));
            session->time2wait = (long)ntohs(staeh->param2) & 0x0000FFFFFL;
            wmb();
            break;
        case ASYNC_EVENT_DROPPING_ALL_CONNECTIONS:
            DEBUG_ERR3("iSCSI: session %p target dropping all connections, reconnect min %u max %u\n",
                       session->log_name, ntohs(staeh->param2), ntohs(staeh->param3));
            session->time2wait = (long)ntohs(staeh->param2) & 0x0000FFFFFL;
            wmb();
            break;
        case ASYNC_EVENT_VENDOR_SPECIFIC:
            printk("iSCSI: session %p ignoring vendor-specific async event, vcode 0x%x\n",
                   session, staeh->async_vcode);
            break;
        case ASYNC_EVENT_PARAM_NEGOTIATION:
            printk("iSCSI: session %p received async event param negotiation, dropping session\n", session);
            iscsi_drop_session(session);
            break;
        default:
            printk("iSCSI: session %p received unknown async event 0x%x at %lu\n",
                   session, staeh->async_event, jiffies);
            break;
    }
}


/* wait for the tx thread to block or exit, ignoring signals.
 * the rx thread needs to know that the tx thread is not running before
 * it can safely close the socket and start a new login phase on a new socket,
 * Also, tasks still in use by the tx thread can't safely be completed on
 * a session drop.
 */
static int wait_for_tx_blocked(iscsi_session_t *session)
{
    while (session->tx_pid) {
        DEBUG_INIT2("iSCSI: thread %d waiting for tx thread %d to block\n",
                    current->pid, session->tx_pid);

        wait_event_interruptible(session->tx_blocked_wait_q, 
                                 test_bit(TX_THREAD_BLOCKED, &session->control_bits));
        
        if (iscsi_handle_signals(session)) {
            DEBUG_INIT3("iSCSI: wait_for_tx_blocked signalled at %lu while waiting for session %p tx %d\n", 
                        jiffies, session, session->tx_pid);
        }
        /* if the session is terminating, the tx thread will exit, waking us up in the process 
         * we don't want to return until the tx thread is blocked, since there's not much
         * the rx thread can do until the tx thread is guaranteed not to be doing anything.
         */
        if (test_bit(TX_THREAD_BLOCKED, &session->control_bits)) {
            DEBUG_INIT2("iSCSI: rx thread %d found tx thread %d blocked\n",
                        current->pid, session->tx_pid);
            return 1;
        }
    }

    /* dead and blocked are fairly similar, really */
    DEBUG_INIT2("iSCSI: rx thread %d found tx thread %d exited\n",
                current->pid, session->tx_pid);
    return 1;
}


/* Wait for a session to be established.  
 * Returns 1 if the session is established, zero if the timeout expires
 * or the session is terminating/has already terminated.
 */
static int wait_for_session(iscsi_session_t *session, int use_timeout)
{
    int ret = 0;
    wait_queue_t waitq;

    if (test_bit(SESSION_ESTABLISHED, &session->control_bits))
        return 1;

    if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
        printk("iSCSI: wait_for_session %p failed, session terminating\n", session);
        return 0;
    }

    init_waitqueue_entry(&waitq, current);
    add_wait_queue(&session->login_wait_q, &waitq);

    DEBUG_INIT3("iSCSI: pid %d waiting for session %p at %lu\n", current->pid, session, jiffies);

    for (;;) {
        set_current_state(TASK_INTERRUPTIBLE);

        if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            ret = 1;
            goto done;
        }

        if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
            ret = 0;
            goto done;
        }

        if (signal_pending(current)) {
            ret = 0;
            goto done;						
        }

        if (use_timeout && session->replacement_timeout) {
            unsigned long timeout;
            long sleep_jiffies = 0;

            if (test_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits)) {
                ret = 0;
                goto done;
            }
            
            if (session->session_drop_time)
                timeout = session->session_drop_time + (HZ * session->replacement_timeout);
            else
                timeout = jiffies + (HZ * session->replacement_timeout);
            
            if (time_before_eq(timeout, jiffies)) {
                printk("iSCSI: pid %d timed out in wait_for_session %p\n", current->pid, session);
                ret = 0;
                goto done;
            }
            
            /* handle wrap-around */
            if (jiffies < timeout)
                sleep_jiffies = timeout - jiffies;
            else
                sleep_jiffies = ULONG_MAX - jiffies + timeout;
            
            schedule_timeout(sleep_jiffies);
        }
        else
            schedule();
    }

 done:
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&session->login_wait_q, &waitq);

    if (ret == 0)
        printk("iSCSI: wait_for_session %p failed\n", session);
    return ret;
}

/* caller must hold the session's portal_lock */
static int set_portal(iscsi_session_t *session, unsigned int p)
{
    iscsi_portal_info_t *portals = session->portals;

    if (portals == NULL) {
        printk("iSCSI: session %p has no portal info, can't set portal %d\n", session, p);
        return 0;
    }
    
    if (p >= session->num_portals) {
        printk("iSCSI: session %p has only %d portals, can't set portal %d\n",
               session, session->num_portals, p);
        return 0;
    }

    session->current_portal = p;

    /* address */
    session->ip_length = portals[p].ip_length;
    memcpy(session->ip_address, portals[p].ip_address, portals[p].ip_length);
    session->port = portals[p].port;
    
    /* TCP options */
    session->tcp_window_size = portals[p].tcp_window_size;
    /* FIXME: type of service */
    
    /* iSCSI operational params */
    session->ImmediateData = 0; /* FIXME: use portal[p].ImmediateData once the code is updated to actually handle it */
    session->InitialR2T = portals[p].InitialR2T;
    session->MaxRecvDataSegmentLength = portals[p].MaxRecvDataSegmentLength;
    session->FirstBurstLength= portals[p].FirstBurstLength;
    session->MaxBurstLength= portals[p].MaxBurstLength; 
    session->DefaultTime2Wait = session->portals[session->current_portal].DefaultTime2Wait;
    session->DefaultTime2Retain = session->portals[session->current_portal].DefaultTime2Retain;

    /* portal group tag */
    session->portal_group_tag = portals[p].tag;
    
    /* timeouts */
    session->login_timeout = portals[p].login_timeout;
    session->auth_timeout = portals[p].auth_timeout;
    session->active_timeout = portals[p].active_timeout;
    session->idle_timeout = portals[p].idle_timeout;
    session->ping_timeout = portals[p].ping_timeout; 
    session->abort_timeout = MAX(1, portals[p].abort_timeout);
    session->reset_timeout = MAX(1, portals[p].reset_timeout);
    session->replacement_timeout = MAX(60, portals[p].replacement_timeout);
    
    wmb();
    DEBUG_INIT2("iSCSI: session %p set to portal %d\n", session, session->current_portal);

    return 1;
}

/* caller must hold the session's portal_lock */
static void next_portal(iscsi_session_t *session)
{
    /* FIXME: this assumes portals don't change once we've started the session.
     * If they do, we probably ought to set the current portal to UINT_MAX when
     * the change occurs, and check for it here, so that we can make sure that
     * whatever portal we pick now has the right tag, even if there's only
     * one portal.  This can get complicated if we don't actually know the portal
     * tags.  Hopefully all targets are passing them to us in the discovery info.
     */
    if (session->num_portals > 1) {
        /* If we've logged out or never established a session,
         * then we can switch to any portal.
         * If a session was dropped with outstanding commands,
         * then we need to login to a portal in the same target portal
         * group, in order to ensure an implicit session logout occurs.
         * The logout prevents command reordering from occuring when 
         * commands that got delayed in the network get delivered to the
         * old session on the old portal after we've started a new session
         * to a different portal.  
         */
        if (!session->ever_established || test_bit(SESSION_LOGGED_OUT, &session->control_bits)) {
            /* can safely use any portal */
            if (session->requested_portal == UINT_MAX) {
                /* no portal was specifically requested, so pick one with
                 * whatever policy we want to.  For now, round-robin.  May
                 * need to add more policies, and make them configurable.  
                 */
                session->current_portal = (session->current_portal + 1) % session->num_portals;
            }
            else {
                /* try the requested portal */
                session->current_portal = session->requested_portal;
                session->requested_portal = UINT_MAX;
            }
        }
        else {
            /* must stay in the current target portal group */
            int current_tag = session->portal_group_tag;
            unsigned int start, p;
            int failed = 1;

            if (session->portal_group_tag < 0) {
                printk("iSCSI: session %p portal group tag unknown, can't switch portals\n", session);
                return;
            }
            
            if (session->requested_portal == UINT_MAX) {
                /* no portal was specifically requested, so pick one with
                 * whatever policy we want to.  For now, round-robin.  May
                 * need to add more policies, and make them configurable.  
                 */
                
                start = p = (session->current_portal + 1) % session->num_portals; 
                do {
                    if (session->portals[p].tag < 0) {
                        /* FIXME: if we can count on the target sending TargetPortalGroupTag,
                         * and a login failure due to a mismatch switches to the next portal,
                         * then we can just try it, and let the login fail if there's a mismatch.
                         */
                        DEBUG_INIT3("iSCSI: session %p skipping portal %u group unknown, must login to group %u\n",
                                    session, p, current_tag);
                        p = (p + 1) % session->num_portals;
                    }
                    else if (session->portals[p].tag == current_tag) {
                        session->current_portal = p;
                        failed = 0;
                        break;
                    }
                    else {
                        DEBUG_INIT4("iSCSI: session %p skipping portal %u group %u, must login to group %u\n",
                                    session, p, session->portals[p].tag, current_tag);
                        p = (p + 1) % session->num_portals;
                    }
                } while (start != p);
                
                if (failed) {
                    /* no longer have a portal we can login to safely */
                    /* FIXME: should we try to trigger a rediscovery in the daemon somehow, 
                     * so that we might get portals we can use?
                     */
                    logmsg(AS_ERROR, "iSCSI: session %p can't find any portals in group %u, staying with portal %u",
                           session, current_tag, session->current_portal);
                    /* we still set the current portal so that a failed temporary redirect
                     * will revert to the original address.
                     */
                }
            }
            else {
                /* try the requested portal if possible, else stay where we are */
                if (session->requested_portal >= session->num_portals) {
                    printk("iSCSI: session %p can't switch to portal %u, only have %u portals\n",
                           session, session->requested_portal, session->num_portals);
                    session->requested_portal = UINT_MAX;
                    return;
                }
                else if (session->portals[session->requested_portal].tag != current_tag) {
                    printk("iSCSI: session %p can't switch to portal %u group %u, must stay in portal group %u\n",
                           session, session->requested_portal, session->portals[session->requested_portal].tag,
                           current_tag);
                    session->requested_portal = UINT_MAX;
                    return;
                }
                else {
                    /* switch to it */
                    session->current_portal = session->requested_portal;
                    session->requested_portal = UINT_MAX;
                }
            }
        }
    }

    /* always set the portal, so that we replace the session IP address
     * and undo any temporary redirects.
     */
    set_portal(session, session->current_portal);
}

static int iscsi_establish_session(iscsi_session_t *session)
{
    int ret = -1;
    uint8_t status_class;
    uint8_t status_detail;

    spin_lock(&session->portal_lock);
    if (session->requested_portal != UINT_MAX) {
        /* request to change to a specific portal */
        next_portal(session);
    }
    else {
        iscsi_portal_info_t *portals = session->portals;
        /* Set the session timeouts and iSCSI op params based on the current portal's settings.
         * Don't change the address, since a termporary redirect may have already changed the address,
         * and we want to use the redirected address rather than the portal's address.
         */
        session->login_timeout = portals[session->current_portal].login_timeout;
        session->auth_timeout = portals[session->current_portal].auth_timeout;
        session->active_timeout = portals[session->current_portal].active_timeout;
        session->idle_timeout = portals[session->current_portal].idle_timeout;
        session->ping_timeout = portals[session->current_portal].ping_timeout; 
        session->abort_timeout = MAX(1, portals[session->current_portal].abort_timeout);
        session->reset_timeout = MAX(1, portals[session->current_portal].reset_timeout);
        session->replacement_timeout = MAX(60, portals[session->current_portal].replacement_timeout);

        session->InitialR2T = session->portals[session->current_portal].InitialR2T;
        session->ImmediateData = session->portals[session->current_portal].ImmediateData;
        session->MaxRecvDataSegmentLength = session->portals[session->current_portal].MaxRecvDataSegmentLength;
        session->FirstBurstLength = session->portals[session->current_portal].FirstBurstLength;
        session->MaxBurstLength = session->portals[session->current_portal].MaxBurstLength;
        session->DefaultTime2Wait = session->portals[session->current_portal].DefaultTime2Wait;
        session->DefaultTime2Retain = session->portals[session->current_portal].DefaultTime2Retain;

        session->portal_group_tag = session->portals[session->current_portal].tag;
    }
    spin_unlock(&session->portal_lock);

    if (LOG_ENABLED(ISCSI_LOG_LOGIN) || LOG_ENABLED(ISCSI_LOG_INIT))
        printk("iSCSI: trying to establish session %p to %s, portal %u, rx %d, tx %d at %lu\n",
               session, session->TargetName, session->current_portal, session->rx_pid, session->tx_pid, jiffies);
    else
        printk("iSCSI: trying to establish session %p to %s, portal %u\n", session, session->TargetName, session->current_portal);


    /* set a timer on the connect */
    if (session->login_timeout) {
        session->login_phase_timer = jiffies + (session->login_timeout * HZ);
        wmb();
    }
    if (LOG_ENABLED(ISCSI_LOG_LOGIN))
        printk("iSCSI: session %p attempting to connect at %lu, timeout at %lu (%d seconds)\n",
               session, jiffies, session->login_phase_timer, session->login_timeout);

    if (!iscsi_connect(session)) {
        if (signal_pending(current))
            printk("iSCSI: session %p connect timed out at %lu\n", session, jiffies);
        else
            printk("iSCSI: session %p connect failed at %lu\n", session, jiffies);
        
        /* switch to the next portal */
        spin_lock(&session->portal_lock);
        next_portal(session);
        spin_unlock(&session->portal_lock);
        goto done;
    }

    /* clear the connect timer */
    session->login_phase_timer = 0;
    wmb();
    iscsi_handle_signals(session);
    
    /* try to make sure other timeouts don't go off as soon as the session is established */
    session->last_rx = jiffies;
    session->last_ping = jiffies - 1;

    /* initialize session fields for the iscsi-login code */
    session->type = ISCSI_SESSION_TYPE_NORMAL;
    session->TargetAlias[0] = '\0';  /* none unless declared by the target */
    session->MaxXmitDataSegmentLength = 8192; /* iSCSI default, unless declared otherwise by the target during login */
    wmb();

    /* use the session's RxBuf for a login PDU buffer, since it is
     * currently unused.  We can't afford to dynamically allocate
     * memory right now, since it's possible we're reconnecting, and
     * the VM system is already blocked trying to write dirty pages to
     * the iSCSI device we're trying to reconnect.  The session's
     * RxBuf was sized to have enough space for us to handle the login
     * phase.
     */
    if (!iscsi_login(session, session->RxBuf, sizeof(session->RxBuf), &status_class, &status_detail)) {
        printk("iSCSI: session %p login failed at %lu, rx %d, tx %d\n", 
               session, jiffies, session->rx_pid, session->tx_pid);
        iscsi_disconnect(session);
        /* possibly failure reasons include a bug in the login code, a
         * bug in the target, a protocol version mismatch, a target
         * portal group tag mismatch, and an unsupported feature
         * request by the target.  Try the next portal.
         */
        spin_lock(&session->portal_lock);
        next_portal(session);
        spin_unlock(&session->portal_lock);
        goto done;
    }

    /* check the login status */
    switch (status_class) {
        case STATUS_CLASS_SUCCESS:
            session->auth_failures = 0;
            ret = 1;
            break;
        case STATUS_CLASS_REDIRECT:
            switch (status_detail) {
                case ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP:
                    /* the session IP address was changed by the login
                     * library, sp just try again with this portal
                     * config but the new address.
                     */
                    session->auth_failures = 0;
                    wmb();
                    ret = 1;  /* not really success, but we want to retry immediately, with no delay */
                    /* FIXME: IPv6 */
                    printk("iSCSI: session %p login to portal %u temporarily redirected to %u.%u.%u.%u port %d\n",
                           session, session->current_portal, 
                           session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3],
                           session->port);
                    goto done;
                case ISCSI_LOGIN_STATUS_TGT_MOVED_PERM:
                    /* for a permanent redirect, we need to update the portal address, and then try again. */
                    ret = 1;  /* not really success, but we want to retry immediately, with no delay */
                    /* FIXME: IPv6 */
                    printk("iSCSI: session %p login to portal %u permanently redirected to %u.%u.%u.%u port %d\n",
                           session, session->current_portal, 
                           session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3],
                           session->port);
                    spin_lock(&session->portal_lock);
                    session->auth_failures = 0;
                    /* reset the address in the portal info */
                    session->portals[session->current_portal].ip_length = session->ip_length;
                    memcpy(session->portals[session->current_portal].ip_address, session->ip_address, session->ip_length);
                    session->portals[session->current_portal].port = session->port;
                    spin_unlock(&session->portal_lock);
                    goto done;
                default:
                    ret = -1;
                    session->auth_failures = 0;
                    wmb();
                    printk("iSCSI: session %p login rejected: redirection type 0x%x not supported\n", session, status_detail);
                    break;
            }
            iscsi_disconnect(session);
            goto done;
        case STATUS_CLASS_INITIATOR_ERR:
            switch (status_detail) {
                case ISCSI_LOGIN_STATUS_AUTH_FAILED:
                    printk("iSCSI: session %p login rejected: initiator failed authentication with target %s\n",
                           session, session->TargetName);
                    iscsi_disconnect(session);
                    /* Try a different portal for the retry */
                    spin_lock(&session->portal_lock);
                    session->auth_failures++;
                    if ((!session->ever_established) && (session->auth_failures >= session->num_portals)) {
                        /* give up, since we've tried every portal, and have never established a session */
                        printk("iSCSI: session %p terminating login attempts, %d of %d portals failed authentication or authorization\n",
                               session, session->auth_failures, session->num_portals);
                        ret = 0;
                    }
                    else {
                        /* try a different portal */
                        next_portal(session);
                        ret = -1;
                    }
                    spin_unlock(&session->portal_lock);
                    goto done;
                case ISCSI_LOGIN_STATUS_TGT_FORBIDDEN:
                    printk("iSCSI: session %p login rejected: initiator failed authorization with target %s\n",
                           session, session->TargetName);
                    iscsi_disconnect(session);
                    spin_lock(&session->portal_lock);
                    session->auth_failures++;
                    if ((!session->ever_established) && (session->auth_failures >= session->num_portals)) {
                        /* give up, since we've tried every portal, and have never established a session */
                        printk("iSCSI: session %p terminating login attempts, %d of %d portals failed authentication or authorization\n",
                               session, session->auth_failures, session->num_portals);
                        ret = 0;
                    }
                    else {
                        /* try a different portal */
                        next_portal(session);
                        ret = -1;
                    }
                    spin_unlock(&session->portal_lock);
                    ret = -1;
                    goto done;
                case ISCSI_LOGIN_STATUS_TGT_NOT_FOUND:
                    printk("iSCSI: session %p login rejected: initiator error - target not found (%02x/%02x)\n",
                           session, status_class, status_detail);
                    session->auth_failures = 0;
                    iscsi_disconnect(session);
                    ret = 0;
                    goto done;
                case ISCSI_LOGIN_STATUS_NO_VERSION:
                    /* FIXME: if we handle multiple protocol versions, before we log an error, try the other supported versions. */
                    printk("iSCSI: session %p login rejected: incompatible version (%02x/%02x), non-retryable, giving up\n",
                           session, status_class, status_detail);
                    session->auth_failures = 0;
                    iscsi_disconnect(session);
                    ret = 0;
                    goto done;
                default:
                    printk("iSCSI: session %p login rejected: initiator error (%02x/%02x), non-retryable, giving up\n",
                           session, status_class, status_detail);
                    session->auth_failures = 0;
                    iscsi_disconnect(session);
                    ret = 0;
                    goto done;
            }
        case STATUS_CLASS_TARGET_ERR:
            printk("iSCSI: session %p login rejected: target error (%02x/%02x)\n",
                   session, status_class, status_detail);
            session->auth_failures = 0;
            iscsi_disconnect(session);
            /* Try a different portal for the retry.  We have no idea
             * what the problem is, but maybe a different portal will
             * work better.
             */
            spin_lock(&session->portal_lock);
            next_portal(session);
            spin_unlock(&session->portal_lock);
            ret = -1;
            goto done;
        default:
            printk("iSCSI: session %p login response with unknown status class 0x%x, detail 0x%x\n",
                   session, status_class, status_detail);
            session->auth_failures = 0;
            iscsi_disconnect(session);
            ret = 0;
            goto done;
    }

    /* logged in, get the new session ready */
    clear_bit(SESSION_LOGGED_OUT, &session->control_bits);
    session->ever_established = 1;
    session->generation++;
    session->auth_failures = 0;
    session->last_rx = jiffies;
    session->last_ping = jiffies - 1;
    session->last_window_check = jiffies;
    session->last_peak_window_size = 0;
    session->last_kill = 0;
    session->current_peak_window_size = max_tasks_for_session(session);
    session->window_peak_check = jiffies;
    session->warm_reset_itt = RSVD_TASK_TAG;
    session->cold_reset_itt = RSVD_TASK_TAG;
    session->nop_reply.ttt = RSVD_TASK_TAG;
    session->nop_reply_head = session->nop_reply_tail = NULL;
    session->session_drop_time = 0; /* used to detect sessions that aren't coming back up */
    session->login_phase_timer = 0;
    if (session->TargetAlias[0])
        session->log_name = session->TargetAlias;
    mb();
    
    /* announce it */
    /* FIXME: log the portal being used? */
    if (session->TargetAlias[0] != '\0')
        printk("iSCSI: session %p #%lu to %s, alias %s, entering full-feature phase\n",
               session, session->generation, session->TargetName, session->TargetAlias);
    else
        printk("iSCSI: session %p #%lu to %s entering full-feature phase\n",
               session, session->generation, session->TargetName);

    if (LOG_ENABLED(ISCSI_LOG_INIT) || LOG_ENABLED(ISCSI_LOG_EH)) {
        printk("iSCSI: session %p #%lu established at %lu, isid 0x%02x%02x%02x%02x%02x%02x, tsih %u, %u unsent cmnds, %u ignored cmnds, %u tasks, bits 0x%08lx\n",
               session, session->generation, jiffies, 
               session->isid[0], session->isid[1], session->isid[2], session->isid[3], 
               session->isid[4], session->isid[5], session->tsid,
               atomic_read(&session->num_cmnds), atomic_read(&session->num_ignored_cmnds), 
               atomic_read(&session->num_active_tasks), session->control_bits);
    }

    /* mark the session as up and accepting commands again */
    set_bit(SESSION_ESTABLISHED, &session->control_bits);
    clear_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits);

    /* wake up everyone waiting for the session to be established */
    wake_up(&session->login_wait_q);
    
    /* make sure we start sending commands again */
    wake_tx_thread(TX_SCSI_COMMAND, session);

 done:
    /* clear any timer that may have been left running */
    session->login_phase_timer = 0;
    mb();
    /* cleanup after a possible timeout expiration */
    if (iscsi_handle_signals(session)) {
        if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
            DEBUG_INIT1("iSCSI: session %p terminating, giving up on login attempts\n", session);
            return 0;
        }
        else {
            DEBUG_INIT1("iSCSI: session %p received signal during login, retrying\n", session);
            return -1;
        }
    }
    
    return ret;
}

/* what kind of completion to do for complete_all_cmnds */
#define COMMAND_RESET 1
#define COMMAND_RETRY 2
#define COMMAND_FAIL  3

/* complete all of the cmnds for this session with the specified result code */
static void complete_all_cmnds(iscsi_session_t *session, int completion_type)
{
    Scsi_Cmnd *sc = NULL, *unsent_cmnds = NULL;
    Scsi_Cmnd *task_cmnd_head = NULL, *task_cmnd_tail = NULL;
    iscsi_task_t *task = NULL, *head = NULL;
    unsigned int num_cmnds;
    unsigned int num_tasks;
    DECLARE_MIDLAYER_FLAGS;
    DECLARE_NOQUEUE_FLAGS;

    /* Grab all tasks and unsent cmnds, and then complete them all in
     * the proper order.  We grab the commands before we abort
     * anything, in case the done() function tries to queue up retries
     * while we're aborting.  We don't want to abort anything that
     * arrives because of done() being called, we just want to abort
     * everything currently in the driver.  We do need to grab and
     * abort even the unsent cmnds, to ensure that dependent SCSI
     * commands won't get re-ordered in cases where the upcoming
     * completion re-queues a command behind the unsent cmnds currently
     * queued.
     *
     * We keep holding the task lock so that any eh_*_handlers will
     * block until we're finished completing commands.
     */

    spin_lock(&session->task_lock);
    
    if (test_bit(ISCSI_HBA_IN_EH_STRATEGY, &session->hba->flags)) {
        /* We can have task refcount problems if we try to do anything,
         * and eh_strategy will do it's own completion anyway.  All
         * commands should be either in STATE_TIMEOUT or STATE_FAILED
         * before eh_strategy starts, so it should recover and restart
         * everything for us.
         */
        DEBUG_EH3("iSCSI: session %p ignored complete_all_cmnds %d at %lu, eh_strategy running\n", 
                  session, completion_type, jiffies);
        goto done;
    }
    DEBUG_EH3("iSCSI: session %p complete_all_cmnds %d at %lu\n", session, completion_type, jiffies);
    
    /* grab all the sent commands (tasks) for this connection */
    num_tasks = atomic_read(&session->num_active_tasks);
    head = session->arrival_order.head;
    session->arrival_order.head = session->arrival_order.tail = NULL;
    atomic_set(&session->num_active_tasks, 0);
    /* clear out the task collections */
    session->rx_tasks.head = session->rx_tasks.tail = NULL;
    session->tx_tasks.head = session->tx_tasks.tail = NULL;
    session->completing_tasks.head = session->completing_tasks.tail = NULL;

    /* send a wake up to each task's wait queue for the aborting and reset task collections.
     * We'll never get a response now, so they should wake up and either retry or fail.
     */
    for (task = first_iscsi_task(&session->tx_abort_tasks); task; task = next_iscsi_task(&session->tx_abort_tasks, task))
        wake_up(&task->wait_q);
    for (task = first_iscsi_task(&session->rx_abort_tasks); task; task = next_iscsi_task(&session->rx_abort_tasks, task))
        wake_up(&task->wait_q);
    for (task = first_iscsi_task(&session->aborted_tasks); task; task = next_iscsi_task(&session->aborted_tasks, task))
        wake_up(&task->wait_q);
    for (task = first_iscsi_task(&session->tx_lun_reset_tasks); task; task = next_iscsi_task(&session->tx_lun_reset_tasks, task))
        wake_up(&task->wait_q);
    for (task = first_iscsi_task(&session->rx_lun_reset_tasks); task; task = next_iscsi_task(&session->rx_lun_reset_tasks, task))
        wake_up(&task->wait_q);
    for (task = first_iscsi_task(&session->lun_reset_tasks); task; task = next_iscsi_task(&session->lun_reset_tasks, task))
        wake_up(&task->wait_q);
    if ((session->warm_reset_itt != RSVD_TASK_TAG) || (session->cold_reset_itt != RSVD_TASK_TAG))
        wake_up(&session->reset_wait_q);
        
    session->rx_abort_tasks.head = session->rx_abort_tasks.tail = NULL;
    session->tx_abort_tasks.head = session->tx_abort_tasks.tail = NULL;
    session->aborted_tasks.head = session->aborted_tasks.tail = NULL;
    session->tx_lun_reset_tasks.head = session->tx_lun_reset_tasks.tail = NULL;
    session->rx_lun_reset_tasks.head = session->rx_lun_reset_tasks.tail = NULL;
    session->lun_reset_tasks.head = session->lun_reset_tasks.tail = NULL;
    session->warm_reset_itt = RSVD_TASK_TAG;
    session->cold_reset_itt = RSVD_TASK_TAG;

    /* grab the ignored and unsent cmnds */
    SPIN_LOCK_NOQUEUE(&session->scsi_cmnd_lock);
    if ((unsent_cmnds = session->ignored_cmnd_head))
        session->ignored_cmnd_tail->host_scribble = (unsigned char *)session->scsi_cmnd_head;
    else
        unsent_cmnds = session->scsi_cmnd_head;
    session->scsi_cmnd_head = session->scsi_cmnd_tail = NULL;
    session->ignored_cmnd_head = session->ignored_cmnd_tail = NULL;
    num_cmnds = atomic_read(&session->num_cmnds) + atomic_read(&session->num_ignored_cmnds);
    atomic_set(&session->num_cmnds, 0);
    atomic_set(&session->num_ignored_cmnds, 0);
    SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);
    
    /* free the tasks, while grabbing the cmnds so that we can complete them */
    spin_lock(&session->hba->free_task_lock);
    while ((task = head)) {
        head = task->order_next;

        if (atomic_read(&task->refcount) == 0) {
            ISCSI_TRACE(ISCSI_TRACE_TaskAborted, sc, task, completion_type, 0);

            task->next = task->prev = task->order_next = task->order_prev = NULL;
            sc = task->scsi_cmnd;
            task->scsi_cmnd = NULL;

            /* save it for completion */
            if (task_cmnd_head) {
                task_cmnd_tail->host_scribble = (void *)sc;
                task_cmnd_tail = sc;
            }
            else {
                task_cmnd_head = task_cmnd_tail = sc;
            }
            sc->host_scribble = NULL;
            
            /* free the task memory back to the kernel */
            free_task(task);
        }
        else {
            /* This should never happen, which is good, since we don't really 
             * have any good options here.  Leak the task memory, and fail to 
             * complete the cmnd, which may leave apps blocked forever in the kernel.
             */
            printk("iSCSI: session %p can't complete sent task %p, refcount %u, command %p\n", 
                   session, task, atomic_read(&task->refcount), task->scsi_cmnd);
        }
    }
    spin_unlock(&session->hba->free_task_lock);

    LOCK_MIDLAYER_LOCK(session->hba->host);

    /* complete the tasks */
    while ((sc = task_cmnd_head)) {
        task_cmnd_head = (Scsi_Cmnd *)sc->host_scribble;
        
        sc->host_scribble = NULL;
        if (sc->scsi_done) {
            switch (completion_type) {
                case COMMAND_RETRY:
                    /* since the semantics of DID_SOFT_ERROR vary, arrange for an unconditional retry */
                    sc->result = HOST_BYTE(DID_BUS_BUSY);
                    sc->retries--;
                    break;
                case COMMAND_RESET:
                    /* docs say to inform midlayer of a reset, though it looks like the midlayer will
                     * ignore the completion if the command has timed out already, and we'd get a
                     * double-completion if the command was in SCSI_STATE_FAILED.
                     */
                    sc->result = HOST_BYTE(DID_RESET);
                    break;
                default:
                case COMMAND_FAIL:
                    /* fail it */
                    sc->result = HOST_BYTE(DID_NO_CONNECT);
                    set_lun_comm_failure(sc);
                    break;
            }
            sc->scsi_done(sc);
        }
    }
    
    /* complete the unsent cmnds */
    while ((sc = unsent_cmnds)) {
        unsent_cmnds = (Scsi_Cmnd *)sc->host_scribble;

        sc->host_scribble = (void *)0x0a11Dead;
        if ( sc->scsi_done) {
            switch (completion_type) {
                case COMMAND_RETRY:
                    /* since the semantics of DID_SOFT_ERROR vary, arrange for an unconditional retry */
                    sc->result = HOST_BYTE(DID_BUS_BUSY);
                    sc->retries--;
                    break;
                case COMMAND_RESET:
                    /* docs say to inform midlayer of a reset, though it looks like the midlayer will
                     * ignore the completion if the command has timed out already, and we'd get a
                     * double-completion if the command was in SCSI_STATE_FAILED.
                     */
                    sc->result = HOST_BYTE(DID_RESET);
                    break;
                default:
                case COMMAND_FAIL:
                    /* fail it */
                    sc->result = HOST_BYTE(DID_NO_CONNECT);
                    set_lun_comm_failure(sc);
                    break;
            }
            ISCSI_TRACE(ISCSI_TRACE_CmndAborted, sc, NULL, completion_type, 0);
            sc->scsi_done(sc);
        }
        else {
            printk("iSCSI: unsent command %p already aborting\n", sc);
        }
    }

    DEBUG_EH3("iSCSI: session %p completed %u tasks and %u cmnds\n", 
              session, num_tasks, num_cmnds);

    UNLOCK_MIDLAYER_LOCK(session->hba->host);

 done:
    spin_unlock(&session->task_lock);
}

static int iscsi_rx_thread(void *vtaskp)
{
    iscsi_session_t *session;
    iscsi_hba_t *hba;
    int rc = -EPIPE, xlen;
    struct msghdr msg;
    struct iovec iov;
    struct IscsiHdr sth;
    unsigned char *rxbuf;
    long login_delay = 0;
    int pad;

    if ( ! vtaskp ) {
        DEBUG_ERR0("iSCSI: rx thread task parameter NULL\n");
        return 0;
    }

    session = (iscsi_session_t *)vtaskp;
    /* whoever created the thread already incremented the session's refcount for us */

    hba = session->hba;

    /* become a daemon kernel thread, and abandon any user space resources */
    sprintf(current->comm,"iscsi-rx");
    iscsi_daemonize(); 
    session->rx_pid = current->pid;
    current->flags |= PF_MEMALLOC;
    wmb();

    /* check to see if iscsi_terminate_session was called before we
     * started running, since we can't get a signal from it until
     * until we set session->rx_pid.
     */
    if (test_bit(SESSION_TERMINATING, &session->control_bits))
        goto ThreadExit;

    /* Block all signals except SIGHUP and SIGKILL */
    spin_lock_irq(&current->sig->siglock);
    siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGHUP));
    RECALC_PENDING_SIGNALS;
    spin_unlock_irq(&current->sig->siglock);                                                            

    DEBUG_INIT3("iSCSI: rx thread %d for session %p starting on cpu%d\n", current->pid, session, smp_processor_id());

    while (!test_bit(SESSION_TERMINATING, &session->control_bits)) {
        unsigned long login_failures = 0;
        
        /* we need a session for the rx and tx threads to use */
        while (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            if (login_delay) {
                printk("iSCSI: session %p to %s waiting %ld seconds before next login attempt\n",
                       session, session->log_name, login_delay);
                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(login_delay * HZ);
            }

            /* ensure we can write to the socket without interference */
            DEBUG_INIT3("iSCSI: rx thread %d waiting for tx blocked for session %p at %lu\n", 
                        current->pid, session, jiffies);
            wait_for_tx_blocked(session);
            if (test_bit(SESSION_TERMINATING, &session->control_bits))
                goto ThreadExit;
            
            /* now that the tx thread is idle, it's safe to clean up the old session, if there was one */
            iscsi_disconnect(session);
            clear_bit(SESSION_DROPPED, &session->control_bits);
            clear_bit(SESSION_TASK_ALLOC_FAILED, &session->control_bits);
            clear_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
            clear_bit(SESSION_WINDOW_CLOSED, &session->control_bits);
            
            /* try to get a new session */
            rc = iscsi_establish_session(session);
            if (rc > 0) {
                /* established or redirected */
                login_failures = 0;
            }
            else if (rc < 0) {
                /* failed, retry */
                login_failures++;
            }
            else {
                /* failed, give up */
                iscsi_terminate_session(session);
                goto ThreadExit;
            }

            /* slowly back off the frequency of login attempts */
            if (login_failures == 0)
                login_delay = 0;
            else if (login_failures < 30)
                login_delay = 1;  /* 30 seconds at 1 sec each */
            else if (login_failures < 48)
                login_delay = 5;  /* another 90 seconds at 5 sec each */
            else if (session->replacement_timeout && 
                     time_before_eq(session->session_drop_time + (HZ * session->replacement_timeout), jiffies))
            {
                login_delay = 10; /* every 10 seconds */
            }
            else {
                /* after the replacement timeout has expired, the
                 * device will probably be offline, so we probably
                 * don't need a session anymore, but it's possible the
                 * device isn't offline yet because of all the
                 * hard-coded sleeps in the SCSI midlayer after resets
                 * occur, and in any case it might be useful to know
                 * if we ever get a session back for debugging
                 * purposes, so we'll keep trying occasionally.
                 */
                login_delay = 60; 
            }
        }

        DEBUG_INIT3("iSCSI: rx thread %d established session %p at %lu\n", current->pid, session, jiffies);
        
        /* handle rx for this session */
        while (!signal_pending(current)) {
            /* check for anything to read on socket */
            iov.iov_base = &sth;
            iov.iov_len = sizeof(sth);
            memset(&msg, 0, sizeof(msg));
            msg.msg_iov = &iov;
            msg.msg_iovlen = 1;
            
            DEBUG_FLOW2("iSCSI: rx thread %d for session %p waiting to receive\n", session->rx_pid, session);
            
            rc = iscsi_recvmsg( session, &msg, sizeof(sth) );
            if (signal_pending(current)) {
                DEBUG_FLOW2("iSCSI: rx thread %d for session %p received signal\n", session->rx_pid, session);
                goto EndSession;
            }
            if ( rc == sizeof(sth) ) {
                /* received something */
                xlen = ntoh24( sth.dlength );

                if (sth.hlength) {
                    /* FIXME: read any additional header segments.
                     * For now, drop the session if one is received, since we can't handle them.
                     */
                    printk("iSCSI: session %p received opcode %x, ahs length %d, dlength %d, itt %u at %lu\n",
                           session, sth.opcode, sth.hlength, xlen, ntohl(sth.itt), jiffies);
                    printk("iSCSI: dropping session %p, additional header segments not supported by this driver version.\n",
                           session);
                    iscsi_drop_session(session);
                    goto EndSession;
                }
                
                /* If there are padding bytes, read them as well */
                pad = xlen % PAD_WORD_LEN;
                if (pad) {
                    pad = PAD_WORD_LEN - pad;
                    xlen += pad;
                }

                DEBUG_FLOW5("iSCSI: rx PDU, opcode %x, len %d on session %p by pid %u at %lu\n", 
                            sth.opcode, xlen, session, current->pid, jiffies);

                if ( xlen && (sth.opcode != ISCSI_OP_SCSI_DATA_RSP) && (sth.opcode != ISCSI_OP_NOOP_IN)) {
                    /* unless it's got a (possibly large) data payload, read the whole PDU into memory beforehand */
                    if ( xlen > ISCSI_RXCTRL_SIZE ) {
                        DEBUG_ERR2("iSCSI: PDU data length too large, opcode %x, dlen %d\n", sth.opcode, xlen);
                        iscsi_drop_session(session);
                        goto EndSession;
                    }
                    rxbuf = session->RxBuf;
                    iov.iov_base = rxbuf;
                    iov.iov_len = xlen;
                    memset( &msg, 0, sizeof(struct msghdr) );
                    msg.msg_iov = &iov;
                    msg.msg_iovlen = 1;
                    rc = iscsi_recvmsg( session, &msg, xlen );
                    if ( rc != xlen ) {
                        DEBUG_ERR3("iSCSI: PDU opcode %x, recvmsg %d failed, %d\n", sth.opcode, xlen, rc);
                        iscsi_drop_session(session);
                        goto EndSession;
                    }
                } 
                else {
                    rxbuf = NULL;
                }
                
                switch (sth.opcode) {
                    case ISCSI_OP_NOOP_IN:
                        iscsi_recv_nop( session, (struct IscsiNopInHdr *)&sth);
                        break;
                    case ISCSI_OP_SCSI_RSP:
                        iscsi_recv_cmd(session, (struct IscsiScsiRspHdr *)&sth, rxbuf );
                        break;
                    case ISCSI_OP_SCSI_TASK_MGT_RSP:
                        iscsi_recv_mgmt_rsp(session, (struct IscsiScsiTaskMgtRspHdr *)&sth );
                        break;
                    case ISCSI_OP_RTT_RSP:
                        iscsi_recv_r2t(session, (struct IscsiRttHdr *)&sth );
                        break;
                    case ISCSI_OP_SCSI_DATA_RSP:
                        iscsi_recv_data( session, (struct IscsiDataRspHdr *)&sth );
                        break;
                    case ISCSI_OP_ASYNC_EVENT:
                        iscsi_recv_async_event(session, (struct IscsiAsyncEvtHdr *)&sth, rxbuf);
                        break;
                    case ISCSI_OP_REJECT_MSG:
                        iscsi_recv_reject(session, (struct IscsiRejectRspHdr *)&sth, rxbuf);
                        break;
                    case ISCSI_OP_LOGOUT_RSP:
                        iscsi_recv_logout(session, (struct IscsiLogoutRspHdr *)&sth);
                        break;
                    default:
                        DEBUG_ERR2("iSCSI: dropping session %p after receiving unexpected opcode 0x%x\n", session, sth.opcode);
                        iscsi_drop_session(session);
                        goto EndSession;
                }
            } 
            else {
                if ( rc != -EAGAIN ) {
                    if (rc == 0) {
                        printk( "iSCSI: session %p closed by target %s\n", session, session->log_name);
                    }
                    else if (rc == -ECONNRESET) {
                        printk( "iSCSI: session %p to %s received connection reset\n", session, session->log_name);
                    } 
                    else if ( rc == -ERESTARTSYS ) {
                        printk( "iSCSI: session %p to %s received signal\n", session, session->log_name);
                    } 
                    else {
                        printk("iSCSI: unexpected read status %d on session %p to %s\n", rc, session, session->log_name);
                    }
                    iscsi_drop_session(session);
                    goto EndSession;
                }
            }
        }

    EndSession:
        DEBUG_INIT2("iSCSI: session %p going down at %lu\n", session, jiffies);

        /* calculate how long to wait before logging in again */
        if (session->time2wait >= 0) {
            /* the target gave us a specific Time2Wait */
            login_delay = session->time2wait;
            session->time2wait = -1;
            DEBUG_INIT2("iSCSI: session %p Time2Wait %ld\n", session, login_delay);
        }
        else {
            /* use the default */
            login_delay = session->DefaultTime2Wait;
            DEBUG_INIT2("iSCSI: session %p DefaultTime2Wait %ld\n", session, login_delay);
        }

        /* handle any signals that may have occured, which may kill the tx thread */
        iscsi_handle_signals(session);

        /* we need to wait for the tx thread to block before trying to complete commands,
         * since it may be using a task at the moment, which means we can't complete it yet.
         * even if the session is terminating, we must wait for the tx thread.
         */
        wait_for_tx_blocked(session);

        if (session->cold_reset_itt != RSVD_TASK_TAG) {
            printk("iSCSI: session %p dropped after cold target reset, SCSI commands completed by reset\n", session);
            session->cold_reset_itt = RSVD_TASK_TAG;
            wmb();

            complete_all_cmnds(session, COMMAND_RESET);

            /* inform the reset functions that the reset is completed */
            set_bit(SESSION_RESET, &session->control_bits);
            wake_up(&session->reset_wait_q);
        }
        else if (session->warm_reset_itt != RSVD_TASK_TAG) {
            printk("iSCSI: session %p dropped after warm target reset, SCSI commands completed by reset\n", session);
            session->warm_reset_itt = RSVD_TASK_TAG;
            wmb();

            complete_all_cmnds(session, COMMAND_RESET);

            /* inform the reset functions that the reset is completed */
            set_bit(SESSION_RESET, &session->control_bits);
            wake_up(&session->reset_wait_q);
        }
        else if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
            /* no point in retrying anything */
            if (test_bit(ISCSI_HBA_SHUTTING_DOWN, &session->hba->flags))
                DEBUG_INIT1("iSCSI: session %p terminating, failing SCSI commands\n", session);
            else
                printk("iSCSI: session %p terminating, failing SCSI commands\n", session);
            complete_all_cmnds(session, COMMAND_FAIL);
        }
        else {
            /* session dropped unexpectedly, often due to network problems */

            /* FIXME: the advantage to completing all commands now is
             * that it resets the command timers, and avoids halting
             * all I/O when the HBA goes into error recovery.  The
             * disadvantage is that we now have an infinite loop if
             * the connection repeatedly drops due to network
             * problems.
             *
             * We could just requeue everything internally and leave
             * the original timeout alone.  This has 2 possible
             * problems.  One is that the sum of the time to notice a
             * bad connection and get a new one may exceed the SCSI
             * command timeout, and send us into error recovery.
             * Resetting the timeout avoids this, but we don't want to
             * reset it forever, since all commands should eventually
             * either timeout or succeed to avoid blocking the reset
             * of the system indefinately.
             *
             * The other possible problem with internal requeueing of
             * commands already sent is that they may have already had
             * their effect, and we just didn't get the completion
             * response from the target.  Will a 2nd reservation
             * conflict with the first? etc.  If all commands are
             * idempotent, we can requeue internally with no problem.
             * If we complete everything back to the SCSI layer, it
             * can decide what's appropriate to do with each command,
             * so we don't have to worry about that.  Since the Linux
             * SCSI midlayer as of lk 2.4.18 doesn't have any logic to
             * do anything other than requeue the command to us,
             * commands that aren't idempotent are probably going to
             * have problems either way.
             *
             * Alternatively, we could add a forward progress timer.
             * If we haven't completed any (or enough) commands in the
             * past N seconds, make future session drops complete with
             * an error rather than a retry.  That would eventually
             * notify upper layers that there is a problem in the I/O
             * path, but still let us extend the timeout some so that
             * we can handle transient connection problems in a mostly
             * transparent way.  This is similar to the effect we used
             * to get by forcing error recovery and having a session
             * replacement timeout, but would allow I/O to other
             * targets to continue uninterrupted, since the HBA would
             * not get blocked by the error recovery process.
             */
            printk("iSCSI: session %p to %s dropped, retrying SCSI commands\n", session, session->log_name);
            complete_all_cmnds(session, COMMAND_RETRY);
        }
        
        /* can't send any nop replies now */
        spin_lock(&session->task_lock);
        session->nop_reply.ttt = RSVD_TASK_TAG;
        while (session->nop_reply_head) {
            iscsi_nop_info_t *nop_info = session->nop_reply_head;
            session->nop_reply_head = nop_info->next;
            DEBUG_ALLOC1("iSCSI: kfree nop_info %p\n", nop_info);
            kfree(nop_info);
        }
        session->nop_reply_tail = NULL;
        /* a ping test also fails on a connection drop */
        session->ping_test_start = 0;
        session->ping_test_rx_length = 0;
        session->ping_test_data_length = 0;
        if (session->ping_test_tx_buffer) {
            kfree(session->ping_test_tx_buffer);
            session->ping_test_tx_buffer = NULL;
        }
        spin_unlock(&session->task_lock);

        /* record the time the session went down */
        session->session_drop_time = jiffies ? jiffies : 1;
        mb();
    }

 ThreadExit:
    DEBUG_INIT2("iSCSI: rx thread %d for session %p exiting\n", session->rx_pid, session);
    /* indicate that we're already going down, so that we don't get killed */
    session->rx_pid = 0;
    mb();

    /* no point in retrying commands, complete them all back trying to provoke
     * an error rather than a retry. 
     */
    if (test_bit(ISCSI_HBA_SHUTTING_DOWN, &session->hba->flags))
        DEBUG_INIT1("iSCSI: session %p terminating, failing SCSI commands\n", session);
    else
        printk("iSCSI: session %p terminating, failing SCSI commands\n", session);
    complete_all_cmnds(session, COMMAND_FAIL);

    /* cleanup the socket */
    if (session->socket) {
        /* wait for the tx thread to exit */
        DEBUG_INIT2("iSCSI: rx thread %d waiting for tx thread %d to exit\n",
                    current->pid, session->tx_pid);
        while (session->tx_pid) {
            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(MSECS_TO_JIFFIES(10));
        }

        /* drop the connection */
        iscsi_disconnect(session);
    }

    set_bit(SESSION_TERMINATED, &session->control_bits);

    /* wake up any ioctls sleeping on the session */
    wake_up(&session->login_wait_q);
    up(&session->probe_sem);

    iscsi_remove_luns(session);

    drop_reference(session);

    return 0;
}


static int iscsi_inet_aton(char *asciiz, unsigned char *ip_address, int *ip_length)
{
    char *c = asciiz;
    unsigned char *ip = ip_address;
    uint32_t base = 10, value = 0;
    int empty;

    if ((asciiz == NULL) || (*asciiz == '\0') || (ip_address == NULL) || (ip_length == NULL))
        return 0;

    /* FIXME: IPv6 */

    /* look for an IPv4 dotted-decimal IP address */
    while (*c) {
        value = 0; 
        base = 10;
        empty = 1;

        /* each part must start with a digit */
        if (!is_digit(*c)) 
            return 0;

        /* figure out the base for this part */
        if (*c == '0') {
            empty = 0;
            base = 8;
            c++;
            if (*c == 'x') {
                base = 16;
                c++;
            }
        }

        /* get the value of this part */
        while (*c && (*c != '.')) {
            if ((base == 16) && (is_hex_lower(*c))) {
                value = (value * base) + (*c - 'a') + 10;
                c++;
                empty = 0;
            }
            else if ((base == 16) && (is_hex_upper(*c))) {
                value = (value * base) + (*c - 'A') + 10;
                c++;
                empty = 0;
            }
            else if (is_digit(*c)) {
                value = (value * base) + (*c - '0');
                c++;
                empty = 0;
            }
            else
                return 0;
        }

        /* reached the end of the part? */
        if (empty) {
            return 0;
        }
        else if (*c == '.') {
            c++;
            if (value <= 0xFF) {
                *ip++ = value & 0xFF;
                value = 0;
            }
            else
                return 0;
        }
    }
     
    if (*c == '\0')  {
        /* end of the ascii address */
        switch (ip - ip_address) {
            default:
                return 0;
            case 0:
                /* a = 32 */
                *ip++ = (value >> 24) & 0xFF;
                *ip++ = (value >> 16) & 0xFF;
                *ip++ = (value >> 8) &  0xFF;
                *ip++ = (value >> 0) &  0xFF;
                if (ip_length)
                    *ip_length = 4;
                return 1;
            case 1:
                /* a.b = 8,24*/
                if (value <= 0x00FFFFFF) {
                    *ip++ = (value >> 16) & 0xFF;
                    *ip++ = (value >> 8) &  0xFF;
                    *ip++ = (value >> 0) &  0xFF;
                    if (ip_length)
                        *ip_length = 4;
                    return 1;
                }
                else
                    return 0;
            case 2:
                /* a.b.c = 8,8,16 */
                if (value <= 0x0000FFFF) {
                    *ip++ = (value >> 8) &  0xFF;
                    *ip++ = (value >> 0) &  0xFF;
                    if (ip_length)
                        *ip_length = 4;
                    return 1;
                }
                else
                    return 0;
            case 3:
                /* a.b.c.d = 8,8,8,8 */
                if (value <= 0x000000FF) {
                    *ip++ = (value >> 0) &  0xFF;
                    if (ip_length)
                        *ip_length = 4;
                    return 1;
                }
                else
                    return 0;
        } 
    }
    
    return 0;
}

static int update_address(iscsi_session_t *session, char *address)
{
    char *tag;
    char *port;
    int ret = 0;
    unsigned char ip[16];
    int ip_length = 4;

    memset(ip, 0, sizeof(ip));
    
    if ((tag = iscsi_strrchr(address, ','))) {
        *tag = '\0';
        tag++;
    }
    if ((port = iscsi_strrchr(address, ':'))) {
        *port = '\0';
        port++;
    }

    /* update the session's IP address and port, based on the
     * TargetAddress passed to us.  For now, we can't resolve DNS
     * names, since that would require us to pass the request up to a
     * user-mode process, and use the NSS system.  We have our own
     * equivalent of inet_aton, and fail to change the address if we
     * get a DNS name instead of an IP address.
     */
    if (iscsi_inet_aton(address, ip, &ip_length)) {
        memcpy(session->ip_address, ip, sizeof(session->ip_address));
        session->ip_length = ip_length;
        if (port)
            session->port = iscsi_strtoul(port, NULL, 0);
        else
            session->port = ISCSI_LISTEN_PORT;

        if (tag)
            session->portal_group_tag = iscsi_strtoul(tag, NULL, 0);

        ret = 1;
    }

    /* restore the original strings */
    if (tag) {
        --tag;
        *tag = ',';
    }
    if (port) {
        --port;
        *port = ':';
    }

    wmb();
    return ret;
}

static int iscsi_session( iscsi_session_t *session, iscsi_session_ioctl_t *ioctld, iscsi_portal_info_t *portals)
{
    iscsi_hba_t *hba;
    int hba_number;
    int channel_number;
    pid_t rx_pid, tx_pid;
    int ret = 0;
    int p;
    DECLARE_NOQUEUE_FLAGS;
    
    DEBUG_INIT2("iSCSI: ioctl at %lu for session to %s\n", jiffies, ioctld->TargetName);

    /* find the HBA that has the requested iSCSI bus */
    hba_number = ioctld->iscsi_bus / ISCSI_MAX_CHANNELS_PER_HBA;
    channel_number = ioctld->iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;

    spin_lock(&iscsi_hba_list_lock);
    hba = iscsi_hba_list;
    while (hba && (hba_number-- > 0)) {
        hba = hba->next;
    }
    spin_unlock(&iscsi_hba_list_lock);

    if (!hba) {
        printk("iSCSI: couldn't find HBA with iSCSI bus %d\n", ioctld->iscsi_bus);
        drop_reference(session);
        return -EINVAL;
    }
    if (!test_bit(ISCSI_HBA_ACTIVE, &hba->flags)) {
        printk("iSCSI: HBA %p is not active, can't add session %p\n", hba, session);
        drop_reference(session);
        return -EINVAL;
    }
    if (!hba->host) {
        printk("iSCSI: HBA %p has no host, can't add session %p\n", hba, session);
        drop_reference(session);
        return -EINVAL;
    }
    if (test_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags)) {
        printk("iSCSI: HBA %p is shutting down, can't add session %p\n", hba, session);
        drop_reference(session);
        return 0;
    }

    /* initialize the session structure */
    session->socket = NULL;

    spin_lock_init(&session->portal_lock);
    session->num_portals = 0;
    session->portals = NULL;
    session->auth_failures = 0;
    session->requested_portal = UINT_MAX;
    session->portal_group_tag = -1;
    
    spin_lock_init( &session->scsi_cmnd_lock);
    session->scsi_cmnd_head = session->scsi_cmnd_tail = NULL;
    atomic_set(&session->num_cmnds, 0);
    session->ignored_cmnd_head = session->ignored_cmnd_tail = NULL;
    atomic_set(&session->num_ignored_cmnds, 0);

    sema_init(&session->probe_sem, 0); /* the first down should block */
    session->probe_order = ioctld->probe_order;
    
    spin_lock_init( &session->task_lock);
    session->arrival_order.head = session->arrival_order.tail = NULL;
    session->rx_tasks.head = session->rx_tasks.tail = NULL;
    session->tx_tasks.head = session->tx_tasks.tail = NULL;
    session->completing_tasks.head = session->completing_tasks.tail = NULL;
    session->rx_abort_tasks.head = session->rx_abort_tasks.tail = NULL;
    session->tx_abort_tasks.head = session->tx_abort_tasks.tail = NULL;
    session->aborted_tasks.head = session->aborted_tasks.tail = NULL;
    session->tx_lun_reset_tasks.head = session->tx_lun_reset_tasks.tail = NULL;
    session->rx_lun_reset_tasks.head = session->rx_lun_reset_tasks.tail = NULL;
    session->lun_reset_tasks.head = session->lun_reset_tasks.tail = NULL;
    atomic_set(&session->num_active_tasks, 0);
    
    init_waitqueue_head(&session->tx_wait_q);
    init_waitqueue_head(&session->tx_blocked_wait_q);
    init_waitqueue_head(&session->login_wait_q);
    init_waitqueue_head(&session->reset_wait_q);
    
    /* save the portal info, and pick which portal to start with */
    session->update_address = &update_address;
    session->portals = portals;
    session->num_portals = ioctld->num_portals;
    set_portal(session, 0U);
    
    session->iscsi_bus = ioctld->iscsi_bus;
    session->host_no = hba->host->host_no;
    session->channel = ioctld->iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;
    session->target_id = ioctld->target_id;
    session->itt = 1;
    session->generation = 0;
    session->ever_established = 0;
    session->time2wait = -1;
    session->logout_itt = RSVD_TASK_TAG;

    /* copy the iSCSI params */
    iscsi_strncpy(session->InitiatorName, ioctld->InitiatorName, sizeof(session->InitiatorName));
    session->InitiatorName[sizeof(session->InitiatorName)-1] = '\0';
    iscsi_strncpy(session->InitiatorAlias, ioctld->InitiatorAlias, sizeof(session->InitiatorAlias));
    session->InitiatorAlias[sizeof(session->InitiatorAlias)-1] = '\0';
    memcpy(session->isid, ioctld->isid, sizeof(session->isid));
    iscsi_strncpy(session->TargetName, ioctld->TargetName, sizeof(session->TargetName));
    session->TargetName[sizeof(session->TargetName)-1] = '\0';
    session->log_name = session->TargetName;

    /* in case the session never comes up */
    session->session_drop_time = jiffies;

    if (ioctld->password_length > 0) {
        if (ioctld->password_length <= sizeof(session->password)) {
            session->auth_client = kmalloc(sizeof(*session->auth_client), GFP_KERNEL);
            if (!session->auth_client) {
                printk("iSCSI: couldn't allocate authentication structure for session %p\n", session);
                drop_reference(session);
                return -ENOMEM;
            }        

            /* save the username and password */
            strncpy(session->username, ioctld->username, sizeof(session->username));
            session->password_length = MIN(sizeof(session->password), ioctld->password_length);
            memcpy(session->password, ioctld->password, session->password_length);
        }
        else {
            printk("iSCSI: password length %d too long, disabling authentication for session %p", 
                   ioctld->password_length, session);
        }
    }

    wmb();

    /* attach the session to the HBA */
    SPIN_LOCK_NOQUEUE(&hba->session_lock);
    if (!add_session(hba, session)) {
        /* couldn't add the session, tell the caller it failed in way that won't cause a retry */
        printk("iSCSI: couldn't add session %p to %s\n", session, session->TargetName);
        SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
        drop_reference(session);
        return 0;
    }
    /* log the session's bus, target id, TargetName, and all of the
     * portals, so that the user has a record of what targets the
     * kernel module was given.
     */
    printk("iSCSI: bus %d target %d = %s\n", session->iscsi_bus, session->target_id, session->TargetName);
    for (p = 0; p < session->num_portals; p++) {
        /* FIXME: IPv6 */
        printk("iSCSI: bus %d target %d portal %u = address %u.%u.%u.%u port %d group %d\n", 
               session->iscsi_bus, session->target_id, p, 
               session->portals[p].ip_address[0], session->portals[p].ip_address[1], 
               session->portals[p].ip_address[2], session->portals[p].ip_address[3],
               session->portals[p].port, session->portals[p].tag);
    }
    SPIN_UNLOCK_NOQUEUE(&hba->session_lock);


    /* we now own the target id.  if we're building device symlinks, clear away any old ones */
    session->dir_mode = ioctld->link_dir_mode;
    if (ioctld->link_base_dir[0] == '/') {
        char *check = ioctld->link_base_dir + strlen(ioctld->link_base_dir) - 1;

        /* remove any trailing slashes, since we add one below */
        while ((check > ioctld->link_base_dir) && (*check == '/'))
            *check-- = '\0';

        /* don't let people use the root directory */
        if (check == ioctld->link_base_dir) {
            printk("iSCSI: cannot use root directory as link base dir\n");
        }
        else {
            sprintf(session->target_link_dir, "%s/bus%d/target%d/", 
                    ioctld->link_base_dir, session->iscsi_bus, session->target_id);
            
            /* this will remove any existing LUN dirs, but MUST be called while
             * session->RxBuf is unused, since it gets used as temp space to hold dirents.
             */
            DEBUG_INIT2("iSCSI: session %p clearing LUN dirs under %s\n", session, session->target_link_dir);
            iscsi_remove_luns(session);
        }
    }
    else if (ioctld->link_base_dir[0]) {
        printk("iSCSI: link base directory must be an absolute directory name, ignoring %s\n", ioctld->link_base_dir);
    }

    /* start a tx thread */
    DEBUG_INIT2("iSCSI: session %p about to start tx and rx threads at %lu\n", 
                session, jiffies);
    atomic_inc(&session->refcount);
    tx_pid = kernel_thread(iscsi_tx_thread, (void *)session, CLONE_VM| CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
    if (tx_pid > 0) {
        DEBUG_INIT3("iSCSI: session %p started tx thread %u at %lu\n", session, tx_pid, jiffies);
    }
    else {
        printk("iSCSI: session %p failed to start tx thread, terminating session\n", session);
        atomic_dec(&session->refcount); /* the thread isn't actually using it */
        iscsi_terminate_session(session);
        drop_reference(session);
        return -EAGAIN;
    }
    
    /* start an rx thread */
    atomic_inc(&session->refcount);
    rx_pid = kernel_thread(iscsi_rx_thread, (void *)session, CLONE_VM| CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
    if (rx_pid > 0) {
        DEBUG_INIT3("iSCSI: session %p started rx thread %u at %lu\n", session, rx_pid, jiffies);
    }
    else {
        printk("iSCSI: session %p failed to start rx thread, terminating session\n", session);
        atomic_dec(&session->refcount); /* the thread isn't actually using it */
        iscsi_terminate_session(session);
        drop_reference(session);
        return -EAGAIN;
    }

    /* wait for the threads to start */
    while ((session->tx_pid == 0) || (session->tx_pid == 0)) {
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));
        if (signal_pending(current)) {
            iscsi_terminate_session(session);
            break;
        }
    }
    
    /* wait for the session login to complete before probing LUNs */
    wait_for_session(session, FALSE);
    if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
        printk("iSCSI: session %p terminating, ioctl returning at %lu\n", session, jiffies);
        goto done;
    }
    else if (signal_pending(current)) {
        iscsi_terminate_session(session);
        printk("iSCSI: session %p ioctl terminated, returning at %lu\n", session, jiffies);
        goto done;
    }

    /* tell the daemon what host_no and channel this target is on */
    ioctld->host_number = session->host_no;
    ioctld->channel = session->channel;
    memcpy(ioctld->TargetAlias, session->TargetAlias, MIN(sizeof(ioctld->TargetAlias), sizeof(session->TargetAlias)));
    ioctld->TargetAlias[sizeof(ioctld->TargetAlias) - 1] = '\0';
    
    /* first figure out what LUNs actually exist */
    iscsi_detect_luns(session);

    if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
        printk("iSCSI: session %p terminating, ioctl returning at %lu\n", session, jiffies);
        goto done;
    }
    else if (signal_pending(current)) {
        iscsi_terminate_session(session);
        printk("iSCSI: session %p ioctl terminated, returning at %lu\n", session, jiffies);
        goto done;
    }

    /* and then try to probe the intersection of the allowed and detected LUNs */
    iscsi_probe_luns(session, ioctld->lun_bitmap, &ioctld->device_info);

    if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
        printk("iSCSI: session %p terminating, ioctl returning at %lu\n", session, jiffies);
        goto done;
    }
    else if (signal_pending(current)) {
        iscsi_terminate_session(session);
        printk("iSCSI: session %p ioctl terminated, returning at %lu\n", session, jiffies);
        goto done;
    }
    
    /* ioctl success */
    ret = 1;

 done:
    drop_reference(session);

    if (signal_pending(current))
        return -EINTR;
    else
        return ret;
}


/* do timer processing for one session, and return the length of time
 * (in jiffies) til this session needs to be checked again.
 */
static unsigned long check_session_timeouts(iscsi_session_t *session)
{
    unsigned long timeout;
    unsigned long session_timeout = 0;

    if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
        /* check login phase timeouts */
        if (session->login_phase_timer) {
            if (time_before_eq(session->login_phase_timer, jiffies)) {
                printk("iSCSI: login phase for session %p (rx %d, tx %d) timed out at %lu, timeout was set for %lu\n", 
                       session, session->rx_pid, session->tx_pid, jiffies, session->login_phase_timer);
                session->login_phase_timer = 0;
                wmb();
                iscsi_drop_session(session);
            }
            else
                session_timeout = session->login_phase_timer;
        }
#if MULTIPATH_SUPPORT
        if (session->replacement_timeout && session->session_drop_time) {
            timeout = session->session_drop_time + (session->replacement_timeout * HZ);
            if (time_before_eq(timeout, jiffies)) {
                /* FIXME: if the session replacement timeout has expired, start completing
                 * commands with an error code, so that any multipath I/O drivers get an
                 * indication that the path has failed, and using a different path would
                 * be a good idea.
                 *
                 * The tricky part is that the locking order is io_request_lock, and then hba->session_lock,
                 * so either we need to acquire the io_request_lock even though we may not need it,
                 * or we need to queue up dropped sessions somewhere to process later, or we need
                 * to queue up failing commands somewhere for completion later.  Queueing
                 * up the commands themselves for completion later needs to ensure the cmnds
                 * don't get accidently reordered if the session comes up, and then drops again.
                 * We can't have 2 threads doing mass completions and maintain ordering.
                 *
                 * Another option: redo the locking order, to move the hba->session_lock before
                 * the io_request_lock/host_lock.  Doable, but must reevaluate all the code for
                 * deadlocks.
                 */
                set_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits);
                printk("iSCSI: session %p replacement timed out at %lu\n", session, jiffies);
            }
            else {
                if ((session_timeout == 0) || (time_before_eq(timeout, session_timeout)))
                    session_timeout = timeout;
            }
        }
#endif
    }
    else {
        /* check full-feature phase timeouts. */

        if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
            /* we're waiting for tasks to complete before logging out.
             * no need to check the CmdSN window, since we won't be starting any more tasks.
             */
            if (time_before_eq(session->logout_response_deadline, jiffies)) {
                /* passed the deadline for a logout response, just drop the session */
                printk("iSCSI: session %p logout response timeout at %lu, dropping session\n", session, jiffies);
                session->logout_response_deadline = 0;
                session->logout_deadline = 0;
                wmb();
                iscsi_drop_session(session);
            }
            else if (time_before_eq(session->logout_deadline, jiffies)) {
                /* send a logout */
                DEBUG_INIT2("iSCSI: session %p logout deadline reached at %lu\n", session, jiffies);
                session->logout_deadline = 0;
                wmb();
                wake_tx_thread(TX_LOGOUT, session);
            }
        }
        else if (test_bit(SESSION_WINDOW_CLOSED, &session->control_bits) &&
            time_before_eq(session->last_window_check + HZ, jiffies)) {
            /* command window closed, ping once a second to ensure we find out
             * when it re-opens.  Target should send us an update when it does,
             * but we're not very trusting of target correctness.
             */
            session->last_window_check = jiffies;
            printk("iSCSI: session %p command window closed, ExpCmdSN %u, MaxCmdSN %u\n", 
                   session, session->ExpCmdSn, session->MaxCmdSn);
            
            /* request a window update from the target with Nops */
            wake_tx_thread(TX_PING, session);
        }
        
        if (atomic_read(&session->num_active_tasks))
            timeout = session->active_timeout;
        else
            timeout = session->idle_timeout;
        
        if (timeout) {
            if (session->ping_timeout && 
                time_before_eq(session->last_rx + (timeout * HZ) + (session->ping_timeout * HZ), jiffies)) {
                /* should have received something by now, kill the connection */
                if ((session->last_kill == 0) || time_before_eq(session->last_kill + HZ, jiffies)) {
                    
                    session->last_kill = jiffies;
                    
                    printk("iSCSI: %lu second timeout expired for session %p, rx %lu, ping %lu, now %lu\n", 
                           timeout + session->ping_timeout, session, session->last_rx, session->last_ping, jiffies);
                    
                    iscsi_drop_session(session);
                    
                    session_timeout = jiffies + HZ;
                }
                else
                    session_timeout = 0;
            }
            else if (time_before_eq(session->last_rx + (timeout * HZ), jiffies)) {
                
                if (time_before_eq(session->last_ping, session->last_rx)) {
                    /* send a ping to try to provoke some traffic */
                    DEBUG_FLOW4("iSCSI: timer queuing ping for session %p, rx %lu, ping %lu, now %lu\n", 
                                session, session->last_rx, session->last_ping, jiffies);
                    session->last_ping = jiffies - 1;
                    
                    wake_tx_thread(TX_PING, session);
                }
                session_timeout = session->last_rx + (timeout * HZ) + (session->ping_timeout * HZ);
            }
            else {
                if (atomic_read(&session->num_active_tasks)) {
                    session_timeout = session->last_rx + (session->active_timeout * HZ);
                }
                else {
                    unsigned long active_timeout, idle_timeout;
                    
                    /* session is idle, but may become active without the timer being notified,
                     * so use smaller of (now + active_timeout, last_rx + idle_timeout)
                     */
                    idle_timeout = session->last_rx + (session->idle_timeout * HZ);
                    active_timeout = jiffies + (session->active_timeout * HZ);
                    if (time_before_eq(idle_timeout, active_timeout)) {
                        session_timeout = idle_timeout;
                    }
                    else {
                        session_timeout = active_timeout;
                    }
                }
            }
        }
    }
    
    return session_timeout;
}

/*
 *  FIXME: it'd probably be cleaner to move the timeout logic to the rx thread.
 *         The only danger is if the rx thread somehow blocks indefinately.
 *         Doing timeouts here makes sure the timeouts get checked, at the
 *         cost of having this code constantly loop.
 */
static int iscsi_timer_thread(void *vtaskp) { 
    iscsi_session_t *session;
    iscsi_hba_t *hba;

    /* become a child of init, and abandon any user space resources */
    sprintf(current->comm, "iscsi-timer");
    iscsi_daemonize(); 

    iscsi_timer_pid = current->pid;
    mb();
    DEBUG_INIT2("iSCSI: timer pid %d starting at %lu\n", iscsi_timer_pid, jiffies);

    /* Block all signals except SIGKILL */
    spin_lock_irq(&current->sig->siglock);
    siginitsetinv(&current->blocked, sigmask(SIGKILL));
    RECALC_PENDING_SIGNALS;
    spin_unlock_irq(&current->sig->siglock);                                                            

    /* wait for the module to initialize */
    while (test_bit(0, &init_module_complete) == 0) {
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));
        if (signal_pending(current)) {
            iscsi_timer_running = 0;
            mb();
            return 0;
        }
    }

    DEBUG_INIT1("iSCSI: timer waiting for HBA at %lu\n", jiffies);
    while (!signal_pending(current)) {
        spin_lock(&iscsi_hba_list_lock);
        hba = iscsi_hba_list;
        spin_unlock(&iscsi_hba_list_lock);

        if (hba)
            break;

        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));
    }

    DEBUG_INIT1("iSCSI: timer looping over HBAs at %lu\n", jiffies);

    while (!signal_pending(current)) {
        unsigned long next_timeout = jiffies + (5 * HZ);
#if (ISCSI_MIN_CANQUEUE != ISCSI_MAX_CANQUEUE)
        int can_queue = 0;
#endif

        spin_lock(&iscsi_hba_list_lock);
        hba = iscsi_hba_list;
        while (hba) {
            DECLARE_NOQUEUE_FLAGS;

            SPIN_LOCK_NOQUEUE(&hba->session_lock);
            session = hba->session_list_head;
            while (session) {
                unsigned long session_timeout = 0;

                if (LOG_ENABLED(ISCSI_LOG_ALLOC))
                    printk("iSCSI: session %p, rx %5u, tx %5u, %u luns, %3u u, %3u i, %3u t, bits 0x%08lx at %lu\n",
                           session, session->rx_pid, session->tx_pid, session->num_luns, 
                           atomic_read(&session->num_cmnds), atomic_read(&session->num_ignored_cmnds),
                           atomic_read(&session->num_active_tasks), 
                           session->control_bits, jiffies);
                
#if (ISCSI_MIN_CANQUEUE != ISCSI_MAX_CANQUEUE)
                if (!sna_lt(session->MaxCmdSn, session->CmdSn)) {
                    /* record how many more commands we can send on this session */
                    can_queue += max_tasks_for_session(session);
                }
#endif
                session_timeout = check_session_timeouts(session);

                /* find the earliest timeout that might occur, so that we know how long to sleep */
                if (session_timeout && time_before_eq(session_timeout, jiffies)) 
                    printk("iSCSI: ignoring session timeout %lu at %lu, last rx %lu, for session %p\n", 
                           session_timeout, jiffies, session->last_rx, session);
                else if (session_timeout && time_before(session_timeout, next_timeout))
                    next_timeout = session_timeout;

                session = session->next;
            }
            SPIN_UNLOCK_NOQUEUE(&hba->session_lock);

#if (ISCSI_MIN_CANQUEUE != ISCSI_MAX_CANQUEUE)
            /* dynamically adjust the number of commands the HBA will accept, based
             * on each session's CmdSN window.
             */
            if (can_queue > ISCSI_MAX_CANQUEUE) {
                /* to avoid exhausting system resources, clamp the maximum number of commands
                 * the driver will accept.  This hopefully fixes the stalls seen when sessions drop
                 * and the daemon can't get a new session up because it's blocked on something.
                 */
                hba->host->can_queue = ISCSI_MAX_CANQUEUE;
                mb();
            }
# if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
            else if (can_queue > (NR_REQUEST / 2)) {
                /* we run into problems if we exhaust every struct request in the system.
                 * limit ourselves to half of the number of requests, so that other reads and
                 * writes can occur to non-iSCSI devices.  Linux 2.2 tends to reserve the top 1/3 
                 * of the requests for reads, so we could use at most 2/3 of the requests
                 * if we want to allow writes to non-iSCSI devices.  1/2 seems like a reasonable
                 * compromise.  Linux 2.4 uses separate request queues per device.
                 */
                hba->host->can_queue = NR_REQUEST / 2;
                mb();
            }
# endif
            else if (can_queue > ISCSI_MIN_CANQUEUE) {
                hba->host->can_queue = can_queue;
                mb();
            }
            else {
                hba->host->can_queue = ISCSI_MIN_CANQUEUE;
                mb();
            }
#endif

            if (LOG_ENABLED(ISCSI_LOG_ALLOC))
                printk("iSCSI: timer - host %d can_queue %d, used %u, free %u, at %lu\n", 
                       hba->host->host_no, hba->host->can_queue, 
                       atomic_read(&hba->num_used_tasks), atomic_read(&hba->num_free_tasks),
                       jiffies);
            
            /* check every 3 minutes to see if we should free tasks from the HBA's pool back to the kernel */
            if (time_before_eq(hba->last_kfree_check + (3 * 60 * HZ), jiffies)) {
                spin_lock(&hba->free_task_lock);
                DEBUG_ALLOC2("iSCSI: checking free tasks at %lu, min_free %u\n", jiffies, hba->min_free_tasks);
                hba->last_kfree_check = jiffies;
                /* always keep some tasks pre-allocated */
                while (hba->min_free_tasks > ISCSI_PREALLOCATED_TASKS) {
                    iscsi_task_t *task;
                    
                    if ((task = pop_task(&hba->free_tasks))) {
                        atomic_dec(&hba->num_free_tasks);
                        hba->min_free_tasks--;
                        kfree(task);
                        //    DEBUG_ALLOC1("iSCSI: kfree task %p\n", task);
                    }
                    else {
                        printk("iSCSI: bug - min_free_tasks %u, free_tasks %u, but couldn't pop a task\n",
                               hba->min_free_tasks, atomic_read(&hba->num_free_tasks));
                        atomic_set(&hba->num_free_tasks, 0);
                    }
                }
                hba->min_free_tasks = atomic_read(&hba->num_free_tasks);
                spin_unlock(&hba->free_task_lock);
            }

            hba = hba->next;
        }
        spin_unlock(&iscsi_hba_list_lock);

        /* possibly start LUN probing */
        if (iscsi_lun_probe_start) {
            if (time_before_eq(iscsi_lun_probe_start, jiffies)) {
                iscsi_possibly_start_lun_probing();
            }
            else if (time_before_eq(iscsi_lun_probe_start, next_timeout)) {
                next_timeout = iscsi_lun_probe_start;
            }
        }

        /* sleep for a while */
        if (time_before(jiffies, next_timeout)) {
            unsigned long sleep;

            /* sleep til the next time a timeout might occur, and handle jiffies wrapping */
            if (next_timeout < jiffies)
                sleep = (ULONG_MAX - jiffies + next_timeout);
            else
                sleep = (next_timeout - jiffies);
            DEBUG_FLOW4("iSCSI: timer sleeping for %lu jiffies, now %lu, next %lu, HZ %u\n", 
                        sleep, jiffies, next_timeout, HZ);

            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(sleep);
            if (signal_pending(current))
                goto finished;
        }
        else {
            /* this should never happen, but make sure we block for at least a little while
             * if it does somehow, otherwise it'll lock up the machine and be impossible
             * to debug what went wrong.
             */
            DEBUG_FLOW3("iSCSI: timer forced to sleep, now %lu, next %lu, HZ %u\n", 
                        jiffies, next_timeout, HZ);
            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(1);
            if (signal_pending(current))
                goto finished;
        }
    }

 finished:
    /* timer finished */
    DEBUG_INIT1("iSCSI: timer leaving kernel at %lu\n", jiffies);
    
    set_current_state(TASK_RUNNING);

    iscsi_timer_running = 0;
    iscsi_timer_pid = 0;
    mb();
    
    return 0;
}

/* shutdown every session on the HBA */
static int iscsi_shutdown_hba(iscsi_hba_t *hba)
{
    int num_sessions = 0;
    iscsi_session_t *session;
    DECLARE_NOQUEUE_FLAGS;

    /* FIXME: we lose info on LUNs probed when this happens.  After
     * this, the kernel module must be reloaded in order for another
     * LUN probe to work correctly.  Just restarting the daemon causes
     * LUN probe attempts, but the kernel's scsi.c will detect that
     * the device is already on the HBA's device list and error out 
     * the add-single-device.  
     */
    
    /* ensure no more sessions get added to the HBA while we're trying to shut it down */
    set_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);

    do {
        num_sessions = 0;
        
        SPIN_LOCK_NOQUEUE(&hba->session_lock);
        for (session = hba->session_list_head; session; session = session->next) {
            num_sessions++;
            set_bit(SESSION_TERMINATING, &session->control_bits);
            if ((session->last_kill == 0) || time_before_eq(session->last_kill + (5 * HZ), jiffies)) {
                session->last_kill = jiffies;
                iscsi_drop_session(session);
            }
        }
        SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
    
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));

    } while (num_sessions);

    /* let sessions get added again */
    clear_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);
    
    return 1;
}

static int iscsi_shutdown(void)
{
    iscsi_hba_t *hba;
    iscsi_session_t *session;
    int num_sessions = 0;
    pid_t pid;
    DECLARE_NOQUEUE_FLAGS;
    
    /* terminate every session on every HBA */
    printk("iSCSI: driver shutdown killing all sessions\n");

    do {
        num_sessions = 0;

        spin_lock(&iscsi_hba_list_lock);
        for (hba = iscsi_hba_list; hba; hba = hba->next) {
            set_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);
            SPIN_LOCK_NOQUEUE(&hba->session_lock);
            for (session = hba->session_list_head; session; session = session->next) {
                num_sessions++;
                set_bit(SESSION_TERMINATING, &session->control_bits);
                if ((session->last_kill == 0) || time_before_eq(session->last_kill + (5 * HZ), jiffies)) {
                    printk("iSCSI: shutdown killing session %p with refcount %u\n", 
                           session, atomic_read(&session->refcount));
                    session->last_kill = jiffies;
                    iscsi_drop_session(session);
                }
            }
            SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
        }
        spin_unlock(&iscsi_hba_list_lock);

        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));
    } while (num_sessions);
    
    /* kill the timer */
    if ((pid = iscsi_timer_pid)) {
        printk("iSCSI: driver shutdown killing timer %d\n", pid);
        kill_proc(pid, SIGKILL, 1);
    }
    printk("iSCSI: driver shutdown waiting for timer to terminate\n");
    while (test_bit(0, &iscsi_timer_running)) {
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));
        if (signal_pending(current))
            return 0;
    }

    /* reset LUN probing */
    iscsi_reset_lun_probing();

    /* let sessions get added again later */
    spin_lock(&iscsi_hba_list_lock);
    for (hba = iscsi_hba_list; hba; hba = hba->next) {
        clear_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);
    }
    spin_unlock(&iscsi_hba_list_lock);

    printk("iSCSI: driver shutdown complete at %lu\n", jiffies);

    return 1;
}

static void iscsi_select_queue_depths(struct Scsi_Host *host, Scsi_Device *device_list)
{
    Scsi_Device *dev;

    DEBUG_INIT1("iSCSI: selecting queue depths for host #%u\n", host->host_no);

    for (dev = device_list; dev; dev = dev->next) {
        if (dev->host != host) continue;                                                                              
    
        if (force_tcq) {
            /* the driver always used to do this, assuming the target would deal
             * with multiple untagged commands.  Now you have to ask for it
             * with the module parameter force_tcq.
             */
            dev->queue_depth = ISCSI_CMDS_PER_LUN;
            if (dev->tagged_supported) {
                if (dev->tagged_queue == 0)
                    printk("iSCSI: enabled tagged command queueing for (%u %u %u %u), type 0x%x, depth %d\n",
                           host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
                else
                    DEBUG_INIT6("iSCSI: enabled tagged command queueing for (%u %u %u %u), type 0x%x, depth %d\n",
                                host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
            }
            else {
                if (dev->tagged_queue == 0)
                    printk("iSCSI: forced tagged command queueing for (%u %u %u %u), type 0x%x, depth %d\n",
                           host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
                else
                    DEBUG_INIT6("iSCSI: forced tagged command queueing for (%u %u %u %u), type 0x%x, depth %d\n",
                                host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
            }
            dev->tagged_queue = 1;
        }
        else if (dev->tagged_supported) {
            dev->tagged_queue = 1;
            dev->queue_depth = ISCSI_CMDS_PER_LUN;
            DEBUG_INIT6("iSCSI: enabled tagged command queueing for (%u %u %u %u), type 0x%x, depth %d\n",
                        host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
        }
        else {
            dev->tagged_queue = 0;
            if (dev->queue_depth != untagged_queue_depth)
                printk("iSCSI: tagged command queueing not supported for (%u %u %u %u), type 0x%x, depth %d\n",
                       host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
            else
                DEBUG_INIT6("iSCSI: tagged command queueing not supported for (%u %u %u %u), type 0x%x, depth %d\n",
                            host->host_no, dev->channel, dev->id, dev->lun, dev->type, dev->queue_depth);
            dev->queue_depth = untagged_queue_depth;
        }
    }
}



int iscsi_detect( Scsi_Host_Template *sht )
{
    struct Scsi_Host *sh;
    iscsi_hba_t *hba;
    int num_tasks;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    sht->proc_dir = &proc_dir_iscsi;
#else
    sht->proc_name = "iscsi";
#endif

    sh = scsi_register( sht, sizeof(iscsi_hba_t) );
    if (!sh ) {
        printk("iSCSI: Unable to register iSCSI HBA\n");
        return 0;
    }

    /* zero these now to disable the scan done during scsi_register_host.
     * iscsi_probe_luns will set them later.
     */
    sh->max_id = 0;
    sh->max_lun = 0;
    sh->max_channel = 0;

    sh->select_queue_depths = iscsi_select_queue_depths;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    /* indicate the maximum CDB length we can accept */
    sh->max_cmd_len = ISCSI_MAX_CMD_LEN;
#endif

    hba = (iscsi_hba_t *)sh->hostdata;
    memset( hba, 0, sizeof(iscsi_hba_t) );

    hba->next = NULL;
    hba->host = sh;

    /* list of sessions on this HBA */
    spin_lock_init(&hba->session_lock);
    hba->session_list_head = NULL;
    hba->session_list_tail = NULL;
    atomic_set(&hba->num_sessions, 0);

    /* pool of free iscsi tasks */
    spin_lock_init(&hba->free_task_lock);
    hba->free_tasks.head = NULL;
    hba->free_tasks.tail = NULL;
    for (num_tasks = 0; num_tasks < ISCSI_PREALLOCATED_TASKS; num_tasks++) {
        iscsi_task_t *task = (iscsi_task_t *)kmalloc(sizeof(*task), GFP_ATOMIC);

        if (task) {
            memset(task, 0x0, sizeof(*task));
            init_waitqueue_head(&task->wait_q);
            add_task(&hba->free_tasks, task);
        }
        else {
            printk("iSCSI: failed to allocate task for HBA %p, host number %d", hba, sh->host_no);
        }
    }

    atomic_set(&hba->num_free_tasks, num_tasks);
    atomic_set(&hba->num_used_tasks, 0);
    hba->min_free_tasks = 0;
    hba->last_kfree_check = jiffies;

    set_bit(ISCSI_HBA_ACTIVE, &hba->flags);
    clear_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);

    hba->host_no = sh->host_no;

    /* for now, there's just one iSCSI HBA */
    mb();
    iscsi_hba_list = hba;
    mb();
    printk("iSCSI: detected HBA %p, host #%d\n", hba, sh->host_no);
    return 1;
}


/* cleanup before unloading the module */
int iscsi_release(struct Scsi_Host *sh)
{
    iscsi_hba_t *hba;
    iscsi_task_t *task;
    
    hba = (iscsi_hba_t *)sh->hostdata;
    if ( ! hba ) {
        return FALSE;
    }

    printk("iSCSI: releasing HBA %p, host #%d\n", hba, hba->host->host_no);
    
    /* remove all sessions on this HBA, and prevent any from being added */
    if (!iscsi_shutdown_hba(hba)) {
        printk("iSCSI: can't release HBA %p, host #%u failed to shutdown\n", hba, sh->host_no);
        return FALSE;
    }

    /* free this HBA's tasks */
#if DEBUG_ALLOC
    if (atomic_read(&hba->num_free_tasks))
        printk("iSCSI: HBA %p freeing kernel memory for %u tasks\n", hba, atomic_read(&hba->num_free_tasks));
#endif
        
    spin_lock(&hba->free_task_lock);
    while ((task = pop_task(&hba->free_tasks))) {
        /* DEBUG_ALLOC1("iSCSI: freeing kernel memory for task %p\n", task); */
        kfree(task);
    }
    spin_unlock(&hba->free_task_lock);

    /* remove from the iSCSI HBA list */
    spin_lock(&iscsi_hba_list_lock);
    if (hba == iscsi_hba_list) {
        iscsi_hba_list = iscsi_hba_list->next;
    }
    else {
        iscsi_hba_t *prior = iscsi_hba_list;

        while (prior && prior->next != hba)
            prior = prior->next;
        if (prior && prior->next == hba)
            prior->next = hba->next;
    }
    spin_unlock(&iscsi_hba_list_lock);

    memset(hba, 0, sizeof(iscsi_hba_t));
    scsi_unregister( sh );

    printk("iSCSI: released HBA %p\n", hba);
    return TRUE;
}

/* remove a Scsi_Cmnd from a singly linked list joined by the host_scribble pointers. */
static int remove_cmnd(Scsi_Cmnd *sc, Scsi_Cmnd **head, Scsi_Cmnd **tail)
{
    if (!sc || !head || !tail) {
        printk("iSCSI: bug - remove_cmnd %p, head %p, tail %p\n", sc, head, tail);
        return 0;
    }

    if (sc == *head) {
        /* it's the head, remove it */
        *head = (Scsi_Cmnd *)sc->host_scribble; /* next */
        if (*head == NULL)
            *tail = NULL;
        sc->host_scribble = NULL;
        return 1;
    }
    else if (*head) {
        Scsi_Cmnd *prior, *next;
        
        /* try find the command prior to sc */
        prior = *head;
        next = (Scsi_Cmnd *)prior->host_scribble;
        while (next && (next != sc)) {
            prior = next;
            next = (Scsi_Cmnd *)prior->host_scribble; /* next command */
        }
        if (prior && (next == sc)) {
            /* remove the command */
            prior->host_scribble = sc->host_scribble;
            if (*tail == sc)
                *tail = prior;
            sc->host_scribble = NULL;
            return 1;
        }
    }

    return 0;
}

int iscsi_abort_task(iscsi_session_t *session, iscsi_task_t *task)
{
    iscsi_task_t *aborted_task = NULL;
    unsigned long now;
    unsigned long timeout;
    wait_queue_t waitq;
    int aborted = 0;
    
    timeout = jiffies + (session->abort_timeout * HZ);
    
    /* add ourselves to the task's wait queue */
    init_waitqueue_entry(&waitq, current);
    add_wait_queue(&task->wait_q, &waitq);
    
    while (!aborted) {

        spin_lock(&session->task_lock);

        /* make sure we have a session to send the abort on */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            if (test_bit(ISCSI_HBA_IN_EH_STRATEGY, &session->hba->flags)) {
                /* we don't really need a session if we're using eh_strategy,
                 * since a TUR success is not required.  The session drop
                 * will cause all outstanding commands on the target to abort,
                 * so we can just clear out the task and be done.
                 */
                DEBUG_EH4("iSCSI: session %p drop aborted itt %u task %p at %lu\n", 
                          session, task->itt, task, jiffies);
                remove_session_task(session, task);
                aborted = 1;
                task->mgmt_itt = RSVD_TASK_TAG;
                spin_unlock(&session->task_lock);
                goto done;
            }
            else {
                spin_unlock(&session->task_lock);
                
                printk("iSCSI: iscsi_abort_task %p waiting for session %p at %lu\n", task, session, jiffies);
                if (!wait_for_session(session, TRUE)) {
                    printk("iSCSI: iscsi_abort_task failed waiting for session %p at %lu\n", session, jiffies);
                    goto done;
                }
                /* reset the timeout for this abort */
                timeout = jiffies + (session->abort_timeout * HZ);

                spin_lock(&session->task_lock);
            }
        }
        
        /* prepare for a wake up before we queue the abort */
        set_current_state(TASK_INTERRUPTIBLE);
        
        /* queue an abort */
        task->mgmt_itt = 0;
        add_task(&session->tx_abort_tasks, task);
        clear_bit(ISCSI_TASK_REJECTED, &task->flags);

        spin_unlock(&session->task_lock);

        /* send the queued abort */
        DEBUG_EH4("iSCSI: session %p trying to abort task %p, itt %u at %lu\n",
                  session, task, task->itt, jiffies);
        wake_tx_thread(TX_ABORT, session);
        
        /* wait for a wake-up from either a task mgmt response, an immediate PDU reject, 
         * a timeout, or a session drop.
         */
        now = jiffies;
        if (timeout >= now)
            schedule_timeout(timeout - now);
        else
            schedule_timeout(ULONG_MAX - now + timeout);
        
        /* remove the task from whatever abort task collection it's in */
        spin_lock(&session->task_lock);
        if (test_and_clear_bit(ISCSI_TASK_REJECTED, &task->flags)) {
            DEBUG_EH5("iSCSI: session %p task %p itt %u mgmt %u rejected at %lu\n", 
                      session, task, task->itt, task->mgmt_itt, jiffies);
        }
        if ((aborted_task = remove_mgmt_task(&session->aborted_tasks, task->mgmt_itt))) {
#if TEST_LUN_RESETS
            Scsi_Cmnd *sc = task->scsi_cmnd;

            /* ignore some successful aborts, to test the other error handlers */
            if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc &&
                (sc->cmnd[0] != TEST_UNIT_READY) &&
                (aborted_task->cmdsn >= (LUN_RESET_FREQUENCY * ABORT_FREQUENCY)) && 
                ((aborted_task->cmdsn % (LUN_RESET_FREQUENCY * ABORT_FREQUENCY)) >= 0) && 
                ((aborted_task->cmdsn % (LUN_RESET_FREQUENCY * ABORT_FREQUENCY)) < LUN_RESET_COUNT))
            {
                printk("iSCSI: ignoring successful abort of command %p, task %p, (%u %u %u %u), Cmd 0x%x\n",
                       sc, aborted_task, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
                /* put the task back in rx_tasks, so that we can find it in the other eh_*_handlers */
                add_task(&session->rx_tasks, aborted_task);
                aborted = 0;
                spin_unlock(&session->task_lock);
                goto done;
            }
#endif
#if TEST_ERROR_RECOVERY_SESSION_DROP
            /* test a session dropping during error recovery */
            if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && task->scsi_cmnd &&
                (task->scsi_cmnd->cmnd[0] != TEST_UNIT_READY) && (task->mgmt_itt) &&
                (task->mgmt_itt % 3)) {
                aborted = 0;
                printk("iSCSI: session %p testing drop while aborting task %p, itt %u, mgmt %u\n",
                   session, task, task->itt, task->mgmt_itt);
                spin_unlock(&session->task_lock);
                iscsi_drop_session(session);
                continue;
            }
#endif

            remove_session_task(session, task);
            aborted = 1;
            task->mgmt_itt = RSVD_TASK_TAG;
        }
        else if ((aborted_task = remove_mgmt_task(&session->rx_abort_tasks, task->mgmt_itt))) {
            /* nothing to do, just needed it out of the queue */
        }
        else if ((aborted_task = remove_mgmt_task(&session->tx_abort_tasks, task->mgmt_itt))) {
            /* nothing to do, just needed it out of the queue */
        }
        spin_unlock(&session->task_lock);

        /* if there's no session, wait for one and try again */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            aborted = 0;
            continue;
        }
        
        /* if the task aborted, we're done */
        if (aborted) {
            DEBUG_EH4("iSCSI: session %p itt %u task %p aborted at %lu\n", session, task->itt, task, jiffies);
            goto done;
        }

        /* if the timeout expired, fail */
        if (time_before_eq(timeout, jiffies)) {
            printk("iSCSI: session %p abort of itt %u, task %p, mgmt %u timed out at %lu\n", 
                   session, task->itt, task, task->mgmt_itt, jiffies);
            goto fail;
        }
        
        /* this shouldn't happen, except possibly if the system is shutting down. */
        if (signal_pending(current)) {
            printk("iSCSI: session %p abort of itt %u, task %p, mgmt %u killed at %lu\n", 
                   session, task->itt, task, task->mgmt_itt, jiffies);
            goto fail;
        }

        /* wait a little bit and then try again.  For now, retry 50 times per second */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(20));
    }

    if (!aborted) {
    fail:
        /* put the task into rx_abort_tasks, so that it can be found later. */
        spin_lock(&session->task_lock);
        add_task(&session->rx_tasks, aborted_task);
        spin_unlock(&session->task_lock);
    }
    
 done:
    /* remove ourselves from the task's wait queue */
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&task->wait_q, &waitq);

    return aborted;
}


int iscsi_abort_unsent_cmnd(iscsi_session_t *session, Scsi_Cmnd *sc)
{
    int ret = 0;
    DECLARE_NOQUEUE_FLAGS;

    /* check if the cmnd is unsent, or was deliberately held so that we could abort it. */
    SPIN_LOCK_NOQUEUE(&session->scsi_cmnd_lock);
    if (remove_cmnd(sc, &session->scsi_cmnd_head, &session->scsi_cmnd_tail)) {
        atomic_dec(&session->num_cmnds);
        printk("iSCSI: session %p aborted unsent command %p to (%u %u %u %u), Cmd 0x%x\n",
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        ret = 1;
    }
    else if (remove_cmnd(sc, &session->ignored_cmnd_head, &session->ignored_cmnd_tail)) {
        atomic_dec(&session->num_ignored_cmnds);
        printk("iSCSI: session %p aborted command %p to (%u %u %u %u), Cmd 0x%x\n",
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        ret = 1;
    }
    SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);

    return ret;
}

/* unconditionally remove the cmnd from all driver data structures 
 * The probing code uses this when cmnds time out or the probe is killed.
 * It aborts the command on our side, but doesn't inform the target.
 * Since the cmnd is either INQUIRY or REPORT_LUNs, the target should
 * always complete the command, and we just discard the response if
 * it's already been removed from our data structures.
 */
int iscsi_squash_cmnd(iscsi_session_t *session, Scsi_Cmnd *sc)
{
    iscsi_task_t *task;
    int ret = 0;

    if (iscsi_abort_unsent_cmnd(session, sc))
        return 1;

    /* remove any task for this cmnd */
    spin_lock(&session->task_lock);
    if ((task = remove_task_for_cmnd(&session->rx_tasks, sc))) {
        /* it's waiting for data, an R2T, or a command response */
        atomic_inc(&task->refcount);
    }
    else if ((task = remove_task_for_cmnd(&session->tx_tasks, sc))) {
        /* it's received an R2T, and is queued to have data sent */
        atomic_inc(&task->refcount);
    }
    else if ((task = remove_task_for_cmnd(&session->completing_tasks, sc))) {
        atomic_inc(&task->refcount);
    }
    /* it should be in one of those collections, since eh_strategy won't see it,
     * as it only finds commands that originated in the midlayer.
     */
    if (task) {
        DEBUG_EH3("iSCSI: session %p squashing task %p, itt %u\n", session, task, task->itt);
        remove_session_task(session, task);

        while (atomic_read(&task->refcount) > 1) {
            /* wait for the driver to quit using the task */
            set_current_state(TASK_UNINTERRUPTIBLE);
            schedule_timeout(MSECS_TO_JIFFIES(10));
        }

        ret = 1;
        /* free the task */
        spin_lock(&session->hba->free_task_lock);
        free_task(task);
        spin_unlock(&session->hba->free_task_lock);
        
        DEBUG_EH3("iSCSI: session %p squashed task %p, itt %u\n", session, task, task->itt);
    }
    spin_unlock(&session->task_lock);

    if (ret == 0) {
        printk("iSCSI: session %p couldn't squash cmnd %p\n", session, sc);
    }
    return ret;
}

int iscsi_abort_cmnd(iscsi_session_t *session, Scsi_Cmnd *sc)
{
    iscsi_task_t *task = NULL;
    iscsi_task_t *aborted_task = NULL;
    unsigned long last_log = 0;
    int refcount;
    int ret = 0;

    /* 
     * check tasks (commands already sent to the target) 
     */
    spin_lock(&session->task_lock);
    if ((task = remove_task_for_cmnd(&session->rx_tasks, sc))) {
        /* it's waiting for data, an R2T, or a command response */
        task->mgmt_itt = 0;
        atomic_inc(&task->refcount);
        set_bit(ISCSI_TASK_ABORTING, &task->flags);
    }
    else if ((task = remove_task_for_cmnd(&session->tx_tasks, sc))) {
        /* it's received an R2T, and is queued to have data sent */
        task->mgmt_itt = 0;
        atomic_inc(&task->refcount);
        set_bit(ISCSI_TASK_ABORTING, &task->flags);
    }
    else if ((task = remove_task_for_cmnd(&session->completing_tasks, sc))) {
        /* already received command completion, no point in sending an abort */
        task->mgmt_itt = 0;
        atomic_inc(&task->refcount);
        set_bit(ISCSI_TASK_ABORTING, &task->flags);
        remove_session_task(session, task);
        aborted_task = task;
        DEBUG_EH4("iSCSI: session %p has no need to send abort for itt %u, task %p, cmnd %p completed\n",
                  session, task->itt, task, sc);
    }
    else {
#if DEBUG_EH
        if (LOG_ENABLED(ISCSI_LOG_EH)) {
            printk("iSCSI: session %p couldn't find task to abort for command %p to (%u %u %u %u), owner 0x%x, "
                   "%u unsent cmnds, %u ignored cmnds, tasks: active %u, used %u, free %u\n", 
                   session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->owner,
                   atomic_read(&session->num_cmnds), atomic_read(&session->num_ignored_cmnds), 
                   atomic_read(&session->num_active_tasks), 
                   atomic_read(&session->hba->num_used_tasks), atomic_read(&session->hba->num_free_tasks));
        }
#endif
    }
    spin_unlock(&session->task_lock);
    
    if (task) {
        if (aborted_task == NULL) {
            printk("iSCSI: session %p aborting itt %u, task %p, command %p to (%u %u %u %u), Cmd 0x%x at %lu\n",
                   session, task->itt, task, sc, 
                   sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], jiffies);
    
            /* try to abort the task */
            if (iscsi_abort_task(session, task)) {
                aborted_task = task;
            }
            else {
                /* the abort failed, give up */
                clear_bit(ISCSI_TASK_ABORTING, &task->flags);
                atomic_dec(&task->refcount);
                return 0;
            }
        }

        while (aborted_task) {
            /* iscsi_abort_task removed the task from the session, so now 
             * we just need to wait for the task refcount to be 1 (just us)
             * before we free it.  This should happen eventually, since
             * if nothing else the timer thread will drop the session,
             * and the tx and rx threads will quit using the task then.
             */
            if ((refcount = atomic_read(&aborted_task->refcount)) <= 1) {
                printk("iSCSI: session %p aborted itt %u, task %p, command %p to (%u %u %u %u), Cmd 0x%x at %lu\n",
                       session, aborted_task->itt, aborted_task, sc, 
                       sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], jiffies);
                atomic_dec(&aborted_task->refcount);
                spin_lock(&session->hba->free_task_lock);
                free_task(aborted_task);
                spin_unlock(&session->hba->free_task_lock);
                return 1;
            }
            
            /* log once per second indicating that eh_abort is waiting */
            if ((last_log == 0) || time_before_eq(last_log + HZ, jiffies)) {
                last_log = jiffies;
                DEBUG_EH4("iSCSI: session %p abort waiting for task %p, command %p, refcount %u before returning\n",
                          session, aborted_task, sc, refcount);
            }

            /* wait a bit and check again */
            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(MSECS_TO_JIFFIES(10));
        }
    }

    return ret;
}

/*
 * The abort handlers will wait for a session to be established, so
 * that we can actually send aborts, and we don't progress through the
 * error recovery functions while we can't do anything useful.  If we
 * don't have a session, we don't want to be doing resets, because:
 *
 * 1) the SCSI layer doesn't handle device resets of multiple commands
 * correctly in any Linux kernel available as of 2/10/2002, so we
 * almost always go on to a bus reset, since sessions rarely have
 * only one command outstanding when the session drops (they'll
 * typically either have 0 or more than 1).
 *
 * 2) resets tend to cause the next command after a reset to
 * return sense data with a deferred error "device reset".  For
 * kernels 2.2.16-2.2.20, 2.4.1, and 2.4.2, if this happens on the TUR
 * following the reset, the device will get marked offline
 * inappropriately, due to code in scsi_error.c mapping NEEDS_RETRY to
 * FAILED for no apparent reason.  For kernels with that mapping
 * removed, the sense data causes the SCSI layer to try to partially
 * complete the sectors of an IO request that have been finished, and
 * retry the rest.  However, sometimes an IO error gets sent to the
 * application.
 *
 * In short, the Linux SCSI layer is sufficiently broken that we'd
 * prefer to just pretend all the commands eventually got aborted,
 * which will hopefully retry everything without reporting errors to
 * applications, since they often can't handle errors in any useful
 * way.  We use a timeout to notice cases where to session doesn't
 * appear to be coming back up, so that we do eventually exit the
 * error recovery no matter what.  Since the error recovery process
 * blocks all commands to the iSCSI HBA, no other iSCSI targets can do
 * IO will any target is in error recovery.  The timeout ensures that
 * eventually the targets with sessions will start getting commands
 * again.  
 */
int iscsi_eh_abort( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_session_t *session = NULL;
    int ret = FAILED;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't abort cmnd %p, no session for command\n", sc);
        return FAILED;
    }

    RELEASE_MIDLAYER_LOCK(host);

    printk("iSCSI: session %p eh_abort at %lu for command %p to (%u %u %u %u), Cmd 0x%x, session %p, rx %u, tx %u, host_failed %u\n",
           session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], 
           session, session->rx_pid, session->tx_pid, sc->host->host_failed);

    /* try to wait til we have a session, since even if it's an unsent
     * command that doesn't require a session itself, the next thing
     * after a successful abort is a TUR, which requires a session to
     * succeed, and we really want it succeed, to avoid the later
     * stages of error recovery, which Linux doesn't handle very well.
     */
    if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
        if (in_interrupt()) {
            printk("iSCSI: session %p eh_abort failed, in interrupt with no session at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
                   session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
            ret = FAILED;
            goto relock;
        }
        printk("iSCSI: session %p eh_abort %p by pid %d waiting for session %p to be established at %lu\n", 
               session, sc, current->pid, session, jiffies);

        if (!wait_for_session(session, TRUE)) {
            printk("iSCSI: eh_abort failed waiting for session %p at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
                   session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
            ret = FAILED;
            goto relock;
        }

        printk("iSCSI: session %p established, eh_abort %p by pid %d wait ended at %lu\n", 
               session, sc, current->pid, jiffies);
    }


#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    if (sc->done_late) {
        /* midlayer ought to check for this race itself */
        ret = SUCCESS;
        goto done;
    }
#endif

    if (iscsi_abort_unsent_cmnd(session, sc)) {
        ret = SUCCESS;
        goto done;
    }

    if (in_interrupt()) {
        printk("iSCSI: session %p eh_abort failed, in interrupt at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
               session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        ret = FAILED;
        goto relock;
    }

    if (iscsi_abort_cmnd(session, sc)) {
        ret = SUCCESS;
    }
    else {
        /* FIXME: consider lying here, and returning SUCCESS.  We may have been called on to abort
         * a task that has already completed, due to a race in the SCSI midlayer code.  We'd prefer
         * to always return true, since the reset code in Linux tends to be buggy.
         */
        ret = FAILED;
        printk("iSCSI: session %p eh_abort couldn't find task for sc %p to (%u %u %u %u), %u unsent cmnds, %u ignored cmnds, tasks: active %u, used %u, free %u\n", 
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, 
               atomic_read(&session->num_cmnds), atomic_read(&session->num_ignored_cmnds), 
               atomic_read(&session->num_active_tasks), atomic_read(&hba->num_used_tasks), atomic_read(&hba->num_free_tasks));

#ifdef DEBUG
        {
            DECLARE_NOQUEUE_FLAGS;

            /* print the session's queued Cmnds */
            SPIN_LOCK_NOQUEUE(&session->scsi_cmnd_lock);
            print_session_cmnds(session);
            SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);
            
            /* print the session's tasks */
            spin_lock(&session->task_lock);
            print_session_tasks(session);
            spin_unlock(&session->task_lock);
        }
#endif
    }

 done:
    if (ret == SUCCESS) {
        /* scsi_unjam_host will send a TUR with a 10 second timeout.
         * try to make sure we have a session to send it on. 
         */
        wait_for_session(session, TRUE);
    }
    else {
        printk("iSCSI: session %p eh_abort failed for command %p, (%u %u %u %u), Cmd 0x%x\n",
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    }
 relock:
    drop_reference(session);
    REACQUIRE_MIDLAYER_LOCK(host);
    return ret;
}


/* trigger a warm reset */
static int warm_reset_target(iscsi_session_t *session)
{
    unsigned long timeout = jiffies + (session->reset_timeout * HZ);
    unsigned long last_log = 0;
    wait_queue_t waitq;
    int reset = 0;
    
    printk("iSCSI: warm target reset starting at %lu, timeout at %lu (%d seconds)\n",
           jiffies, timeout, session->reset_timeout);

    if (in_interrupt()) {
        printk("iSCSI: warm_reset_target in interrupt, failing\n");
        return 0;
    }

    /* make sure there is only one oustanding reset, 
     * and prevent queued commands from being sent.
     */
    if (test_and_set_bit(SESSION_RESETTING, &session->control_bits)) {
        /* already resetting */
        printk("iSCSI: session %p to %s already resetting, reset_target failed\n",
               session, session->log_name);
        return 0;
    }
    
    /* add ourselves to the session's reset wait queue */
    init_waitqueue_entry(&waitq, current);
    add_wait_queue(&session->reset_wait_q, &waitq);
    
    while (!reset) {
        /* make sure we have a session to send the reset on */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            printk("iSCSI: warm_reset_target waiting for session %p at %lu\n", session, jiffies);
            if (!wait_for_session(session, TRUE)) {
                printk("iSCSI: warm_reset_target failed waiting for session %p at %lu\n", session, jiffies);
                break;
            }
            /* reset the timeout */
            timeout = jiffies + (session->reset_timeout * HZ);
        }

        set_current_state(TASK_INTERRUPTIBLE);

        /* queue up a reset for the target */
        set_bit(SESSION_RESETTING, &session->control_bits);
        DEBUG_EH2("iSCSI: session %p trying warm target reset at %lu\n", session, jiffies);
        wake_tx_thread(TX_WARM_TARGET_RESET, session);

        /* wait for either a task mgmt response, a PDU reject, a timeout, or a session drop */
        if (timeout >= jiffies)
            schedule_timeout(timeout - jiffies);
        else
            schedule_timeout(ULONG_MAX - jiffies + timeout);

        /* if the reset is confirmed, we're done */
        if (test_and_clear_bit(SESSION_RESET, &session->control_bits)) {
#if TEST_COLD_RESETS
            printk("iSCSI: session %p ignoring warm target reset at %lu\n", session, jiffies);
            reset = 0;
#else
            printk("iSCSI: session %p confirmed warm target reset at %lu\n", session, jiffies);
            reset = 1;
#endif
            break;
        }

        /* if the session has dropped, wait for another */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            reset = 0;
            continue;
        }

        /* if the timeout expired, fail */
        if (time_before_eq(timeout, jiffies)) {
            clear_bit(TX_WARM_TARGET_RESET, &session->control_bits);
            printk("iSCSI: warm target reset timed out at %lu (%d seconds) for session %p to %s\n",
                   jiffies, session->reset_timeout, session, session->log_name);
            break;
        }
        
        /* this shouldn't happen, except possibly if the system is shutting down. */
        if (signal_pending(current)) {
            clear_bit(TX_WARM_TARGET_RESET, &session->control_bits);
            printk("iSCSI: warm target reset killed at %lu for session %p to %s\n", 
                   jiffies, session, session->log_name);
            break;
        }

        /* wait a bit and then try again */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(5));
    }

    /* remove ourselves from the session's reset wait queue */
    remove_wait_queue(&session->reset_wait_q, &waitq);

    if (reset) {
        iscsi_task_t *task;
        int nonzero_refcount = 1;

        printk("iSCSI: completing commands after warm target reset for session %p to %s\n", 
               session, session->log_name);
        
        /* make sure the refcount of all tasks is zero before continuing.
         * if it doesn't happen before the timeout expires, try killing the session,
         * to force the refcount to zero.
         */
        while (nonzero_refcount) {
            spin_lock(&session->task_lock);
            nonzero_refcount = 0;
            task = first_iscsi_task(&session->arrival_order);
            while (!nonzero_refcount && task) {
                if (time_before(timeout, jiffies)) {
                    printk("iSCSI: reset timeout expired while waiting for task %p refcount %u, killing session\n",
                           task, atomic_read(&task->refcount));
                    /* try killing the session */
                    iscsi_drop_session(session);
                    /* and waiting a bit longer */
                    timeout = jiffies + (session->reset_timeout * HZ);
                    break;
                }
                if (atomic_read(&task->refcount) == 0) {
                    task = order_next_task(&session->arrival_order, task);
                }
                else {
                    nonzero_refcount = 1;
                }
            }
            spin_unlock(&session->task_lock);
             
            if (nonzero_refcount) {
                /* log once per second indicating that we're waiting */
                if ((last_log == 0) || time_before_eq(last_log + HZ, jiffies)) {
                    last_log = jiffies;
                    printk("iSCSI: warm reset waiting for task %p refcount %u to reach 0\n", 
                           task, atomic_read(&task->refcount));
                }
                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(MSECS_TO_JIFFIES(10));
            }
        }

        if (!test_bit(ISCSI_HBA_IN_EH_STRATEGY, &session->hba->flags)) {
            printk("iSCSI: warm target reset, completing SCSI commands for session %p to %s with DID_RESET\n", 
                   session, session->log_name);
        }

        complete_all_cmnds(session, COMMAND_RESET);

        printk("iSCSI: warm target reset succeeded at %lu for session %p to %s\n", 
               jiffies, session, session->log_name);
    }

    /* let commands start flowing again */
    clear_bit(SESSION_RESETTING, &session->control_bits);
    wake_tx_thread(TX_SCSI_COMMAND, session);

    return reset;
}

/* trigger a cold reset */
static int cold_reset_target(iscsi_session_t *session)
{
    unsigned long timeout = jiffies + (session->reset_timeout * HZ);
    wait_queue_t waitq;
    int reset = 0;
    
    printk("iSCSI: cold target reset starting at %lu, timeout at %lu (%d seconds)\n",
           jiffies, timeout, session->reset_timeout);

    if (in_interrupt()) {
        printk("iSCSI: cold_reset_target in interrupt, failing\n");
        return 0;
    }

    /* make sure there is only one oustanding reset, 
     * and prevent queued commands from being sent.
     */
    if (test_and_set_bit(SESSION_RESETTING, &session->control_bits)) {
        /* already resetting */
        printk("iSCSI: session %p to %s already resetting, cold_reset_target failed\n",
               session, session->log_name);
        return 0;
    }
    
    /* add ourselves to the session's reset wait queue */
    init_waitqueue_entry(&waitq, current);
    add_wait_queue(&session->reset_wait_q, &waitq);
    
    while (!reset) {
        /* make sure we have a session to send the reset on */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            printk("iSCSI: cold_reset_target waiting for session %p at %lu\n", session, jiffies);
            if (!wait_for_session(session, TRUE)) {
                printk("iSCSI: cold_reset_target failed waiting for session %p at %lu\n", session, jiffies);
                break;
            }
            /* reset the timeout */
            timeout = jiffies + (session->abort_timeout * HZ);
        }

        set_current_state(TASK_INTERRUPTIBLE);

        /* queue up a reset for the target */
        set_bit(SESSION_RESETTING, &session->control_bits);
        DEBUG_EH2("iSCSI: session %p trying cold target reset at %lu\n", session, jiffies);
        wake_tx_thread(TX_COLD_TARGET_RESET, session);

        /* wait for either a task mgmt response, a PDU reject, a timeout, or a session drop */
        if (timeout >= jiffies)
            schedule_timeout(timeout - jiffies);
        else
            schedule_timeout(ULONG_MAX - jiffies + timeout);

        /* if the reset is confirmed, we're done */
        if (test_and_clear_bit(SESSION_RESET, &session->control_bits)) {
            printk("iSCSI: session %p confirmed cold target reset at %lu\n", session, jiffies);
            reset = 1;
            break;
        }

        /* if the session has dropped, wait for another */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            continue;
        }

        /* if the timeout expired, fail */
        if (time_before_eq(timeout, jiffies)) {
            clear_bit(TX_WARM_TARGET_RESET, &session->control_bits);
            printk("iSCSI: cold target reset timed out at %lu (%d seconds) for session %p to %s\n", 
                   jiffies, session->reset_timeout, session, session->log_name);
            break;
        }
        
        /* this shouldn't happen, except possibly if the system is shutting down. */
        if (signal_pending(current)) {
            clear_bit(TX_WARM_TARGET_RESET, &session->control_bits);
            printk("iSCSI: cold target reset killed at %lu for session %p to %s\n", 
                   jiffies, session, session->log_name);
            break;
        }

        /* wait a bit and then try again */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(5));
    }

    /* remove ourselves from the session's reset wait queue */
    remove_wait_queue(&session->reset_wait_q, &waitq);

    if (reset) {
        /* the session drop already completed all the commands */
        printk("iSCSI: cold target reset succeeded at %lu for session %p to %s\n", 
               jiffies, session, session->log_name);
    }

    /* let commands start flowing again */
    clear_bit(SESSION_RESETTING, &session->control_bits);
    wake_tx_thread(TX_SCSI_COMMAND, session);
    
    return reset;
}

int iscsi_reset_lun(iscsi_session_t *session, iscsi_task_t *task)
{
    iscsi_task_collection_t tasks_for_lun;
    iscsi_task_t *reset_task = NULL;
    Scsi_Cmnd *unsent_cmnds = NULL;
    Scsi_Cmnd *sc = task->scsi_cmnd;
    wait_queue_t waitq;
    unsigned long last_log = 0;
    unsigned long timeout;
    int reset = 0;
    DECLARE_MIDLAYER_FLAGS;
    DECLARE_NOQUEUE_FLAGS;
    
    if (task->itt) {
        printk("iSCSI: session %p attempting LUN reset for itt %u, task %p, command %p to (%u %u %u %u), Cmd 0x%x at %lu\n",
               session, task->itt, task, sc, 
               sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], jiffies);
    }
    else {
        printk("iSCSI: session %p attempting LUN reset for command %p to (%u %u %u %u), Cmd 0x%x at %lu\n",
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], jiffies);
    }

    tasks_for_lun.head = tasks_for_lun.tail = NULL;
    timeout = jiffies + (session->reset_timeout * HZ);
    
    /* add ourselves to the task's wait queue */
    init_waitqueue_entry(&waitq, current);
    add_wait_queue(&task->wait_q, &waitq);
    
    while (!reset) {
        /* make sure we have a session to send the reset on */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            printk("iSCSI: iscsi_reset_lun waiting for session %p at %lu\n", session, jiffies);
            if (!wait_for_session(session, TRUE)) {
                printk("iSCSI: iscsi_reset_lun failed waiting for session %p at %lu\n", session, jiffies);
                break;
            }
            /* reset the timeout */
            timeout = jiffies + (session->abort_timeout * HZ);
        }

        set_current_state(TASK_INTERRUPTIBLE);

        /* send a reset */
        spin_lock(&session->task_lock);
        clear_bit(ISCSI_TASK_REJECTED, &task->flags);
        add_task(&session->tx_lun_reset_tasks, task);
        spin_unlock(&session->task_lock);
        wake_tx_thread(TX_LUN_RESET, session);
        
        /* wait for a wake-up from either a task mgmt response, an immediate PDU reject, or a timeout */
        if (timeout >= jiffies)
            schedule_timeout(timeout - jiffies);
        else
            schedule_timeout(ULONG_MAX - jiffies + timeout);
        
        /* remove the task from whatever task collection it's in */
        spin_lock(&session->task_lock);
        if (test_and_clear_bit(ISCSI_TASK_REJECTED, &task->flags)) {
            DEBUG_EH4("iSCSI: session %p task %p itt %u mgmt %u rejected\n", session, task, task->itt, task->mgmt_itt);
        }
        if ((reset_task = remove_mgmt_task(&session->lun_reset_tasks, task->mgmt_itt))) {
#if TEST_WARM_RESETS
            if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc &&
                (sc->cmnd[0] != TEST_UNIT_READY) &&
                (task->cmdsn >= (WARM_RESET_FREQUENCY * LUN_RESET_FREQUENCY * ABORT_FREQUENCY)) &&
                ((task->cmdsn % (WARM_RESET_FREQUENCY * LUN_RESET_FREQUENCY * ABORT_FREQUENCY)) >= 0) &&
                ((task->cmdsn % (WARM_RESET_FREQUENCY * LUN_RESET_FREQUENCY * ABORT_FREQUENCY)) < WARM_RESET_COUNT))
            {
                Scsi_Cmnd *sc = task->scsi_cmnd;

                printk("iSCSI: ignoring successful LUN reset, task %p, sc %p, (%u %u %u %u), Cmd 0x%x\n",
                       task, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
                if (task->itt != 0) {
                    /* make sure the task can be found somewhere */
                    add_task(&session->rx_tasks, task);
                }
                spin_unlock(&session->task_lock);
                reset = 0;
                break;
            }
#endif
#if TEST_ERROR_RECOVERY_SESSION_DROP
            /* test a session dropping during error recovery */
            if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && task->scsi_cmnd &&
                (task->scsi_cmnd->cmnd[0] != TEST_UNIT_READY) && (task->mgmt_itt) &&
                (task->mgmt_itt % 3)) 
            {
                reset = 0;
                printk("iSCSI: session %p testing drop during lun reset of task %p, itt %u, mgmt %u\n",
                       session, task, task->itt, task->mgmt_itt);
                spin_unlock(&session->task_lock);
                iscsi_drop_session(session);
                continue;
            }
#endif
            /* grab all tasks and commands for this LUN */
            remove_tasks_for_lun(&session->rx_tasks, sc->lun);
            remove_tasks_for_lun(&session->tx_tasks, sc->lun);
            remove_tasks_for_lun(&session->completing_tasks, sc->lun);
            remove_tasks_for_lun(&session->rx_abort_tasks, sc->lun);
            remove_tasks_for_lun(&session->tx_abort_tasks, sc->lun);
            remove_tasks_for_lun(&session->aborted_tasks, sc->lun);
            remove_tasks_for_lun(&session->rx_lun_reset_tasks, sc->lun);
            remove_tasks_for_lun(&session->tx_lun_reset_tasks, sc->lun);
            remove_tasks_for_lun(&session->lun_reset_tasks, sc->lun);
#if DEBUG_EH
            if (LOG_ENABLED(ISCSI_LOG_EH))
                printk("iSCSI: session %p LUN reset (%u %u %u %u) grabbing tasks, arrival head %p, tail %p, num %u\n", 
                       session, sc->host->host_no, sc->channel, sc->target, sc->lun, 
                       session->arrival_order.head, session->arrival_order.tail, atomic_read(&session->num_active_tasks));
#endif
            move_session_tasks_for_lun(&tasks_for_lun, session, sc->lun);
#if DEBUG_EH
            if (LOG_ENABLED(ISCSI_LOG_EH))
                printk("iSCSI: session %p LUN reset (%u %u %u %u) done grabbing tasks, arrival head %p, tail %p, num %u, lun tasks %p, %p\n", 
                       session, sc->host->host_no, sc->channel, sc->target, sc->lun, 
                       session->arrival_order.head, session->arrival_order.tail, atomic_read(&session->num_active_tasks),
                       tasks_for_lun.head, tasks_for_lun.tail);
#endif                
            /* grab all unsent cmnds for this LUN as well */
            SPIN_LOCK_NOQUEUE(&session->scsi_cmnd_lock);
            unsent_cmnds = remove_session_cmnds_for_lun(session, sc->lun);
            SPIN_UNLOCK_NOQUEUE(&session->scsi_cmnd_lock);
            
            /* we're now committed to completing these tasks and commands,
             * and returning SUCCESS.
             */
            reset = 1;
            task->mgmt_itt = 0;
        }
        else if ((reset_task = remove_mgmt_task(&session->rx_lun_reset_tasks, task->mgmt_itt))) {
            task->mgmt_itt = 0;
        }
        else if ((reset_task = remove_mgmt_task(&session->tx_lun_reset_tasks, task->mgmt_itt))) { 
            task->mgmt_itt = 0;
        }
        spin_unlock(&session->task_lock);

        /* if there's no session, wait for one and try again */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
            reset = 0;
            continue;
        }

        /* if the reset worked, we're done */
        if (reset) {
            DEBUG_EH5("iSCSI: session %p itt %u task %p LUN %d reset successful at %lu\n", 
                      session, task->itt, task, sc->lun, jiffies);
            break;
        }
        
        /* if the timeout expired, fail */
        if (time_before_eq(timeout, jiffies)) {
            printk("iSCSI: session %p reset of LUN %d timed out at %lu\n", session, sc->lun, jiffies);
            break;
        }
        
        /* this shouldn't happen, except possibly if the system is shutting down. */
        if (signal_pending(current)) {
            printk("iSCSI: session %p reset of LUN %d killed\n", session, sc->lun);
            break;
        }

        /* if the session dropped, just keep retrying.  
         * We need a session back up before we can return anyway 
         */
        
        /* wait a bit and then try again */
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(5));
    }
    
    /* remove ourselves from the task's wait queue */
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&task->wait_q, &waitq);

    if (reset) {
        /* we need to complete tasks and commands */
        unsigned int num_tasks = 0, num_cmnds = 0;
        Scsi_Cmnd *cmnd, *task_cmnd_head = NULL, *task_cmnd_tail = NULL;

        /* wait for all the task refcounts to hit zero */
        task = first_iscsi_task(&tasks_for_lun);
        while (task) {
            if (time_before(timeout, jiffies)) {
                printk("iSCSI: session %p LUN reset timeout expired while waiting for task %p refcount %u, killing session %p\n",
                       session, task, atomic_read(&task->refcount), session);
                /* try killing the session */
                iscsi_drop_session(session);
                /* and waiting a bit longer */
                timeout = jiffies + (10 * HZ);
            }
            if (atomic_read(&task->refcount) == 0) {
                task = next_iscsi_task(&tasks_for_lun, task);
            }
            else {
                /* log once per second indicating that we're waiting */
                if ((last_log == 0) || time_before_eq(last_log + HZ, jiffies)) {
                    last_log = jiffies;
                    DEBUG_EH4("iSCSI: session %p lun %d reset waiting for task %p refcount %u to reach 0\n", 
                              session, sc->lun, task, atomic_read(&task->refcount));
                }
                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(MSECS_TO_JIFFIES(10));
            }
        }
        
        /* free all the tasks, and then complete the cmnds corresponding to each task freed */
        spin_lock(&session->hba->free_task_lock);
        while ((task = pop_task(&tasks_for_lun))) {
            /* don't complete or free tasks allocated just for the reset */
            if (task->itt == 0)
                continue;

            if (task->scsi_cmnd && task->scsi_cmnd->scsi_done) {
                cmnd = task->scsi_cmnd;
                
                num_tasks++;
                if (task_cmnd_head) {
                    task_cmnd_tail->host_scribble = (void *)cmnd;
                    task_cmnd_tail = cmnd;
                }
                else {
                    task_cmnd_head = task_cmnd_tail = cmnd;
                }
                cmnd->host_scribble = NULL;

                task->scsi_cmnd = NULL;
            }
            free_task(task);
        }
        spin_unlock(&session->hba->free_task_lock);

        printk("iSCSI: session %p LUN reset succeeded for command %p, (%u %u %u %u), Cmd 0x%x at %lu\n",
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], jiffies);

        if (test_bit(ISCSI_HBA_IN_EH_STRATEGY, &session->hba->flags)) {
            /* we're done.  eh_strategy will take care of the commands */
            return reset;
        }

        LOCK_MIDLAYER_LOCK(session->hba->host);

        /* complete all the tasks for this LUN */
        while ((cmnd = task_cmnd_head)) {
            task_cmnd_head = (Scsi_Cmnd *)cmnd->host_scribble;
            
            if (cmnd->scsi_done) {
                DEBUG_EH4("iSCSI: session %p LUN reset completing sent task %p, command %p at %lu\n", 
                          session, (void *)cmnd->host_scribble, cmnd, jiffies);
                cmnd->host_scribble = NULL;
                cmnd->result = HOST_BYTE(DID_RESET);
                cmnd->scsi_done(cmnd);
            }
        }
        
        /* complete all unsent cmnds for this LUN */
        while ((cmnd = unsent_cmnds)) {
            unsent_cmnds = (Scsi_Cmnd *)cmnd->host_scribble;

            cmnd->host_scribble = NULL;
            if (cmnd->scsi_done) {
                num_cmnds++;
                DEBUG_EH3("iSCSI: session %p LUN reset completing unsent cmnd %p at %lu\n", session, cmnd, jiffies);
                cmnd->result = HOST_BYTE(DID_RESET);
                cmnd->scsi_done(cmnd);
            }
        }

        UNLOCK_MIDLAYER_LOCK(session->hba->host);
        
        DEBUG_EH7("iSCSI: session %p LUN reset (%u %u %u %u) completed %u tasks and %u cmnds\n",
                  session, sc->host->host_no, sc->channel, sc->target, sc->lun, num_tasks, num_cmnds);
    }
    else {
        printk("iSCSI: session %p LUN reset failed for command %p, (%u %u %u %u), Cmd 0x%x at %lu\n",
               session, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], jiffies);
    }

    return reset;
}


/*
 * All the docs say we're supposed to reset the device and complete
 * all commands for it back to the SCSI layer.  However, the SCSI
 * layer doesn't actually count how many commands are completed back
 * to it after a device reset, but rather just assumes only 1 command,
 * with a comment saying it should be fixed to handle the case where
 * there are multiple commands.  
 *
 * If there are multiple commands, the SCSI layer will blindly
 * continue on to the next stage of error recovery, even if we
 * complete all the failed commands back to it after a device reset.
 * Hopefully the Linux SCSI layer will be fixed to handle this
 * corectly someday.  In the meantime, we do the right thing here, and
 * make sure the other reset handlers can deal with the case where
 * they get called with a command that has already been completed back
 * to the SCSI layer by a device reset.
 *   
 */
int
iscsi_eh_device_reset( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_session_t *session = NULL;
    iscsi_task_t *task = NULL;
    unsigned long current_gen;
    int ret = FAILED;

    if ( ! sc ) {
        printk("iSCSI: device reset, no SCSI command\n");
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: device reset, no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: device reset, no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't reset device for cmnd %p, no session\n", sc);
        return FAILED;
    }

    printk("iSCSI: eh_device_reset at %lu for command %p to (%u %u %u %u), Cmd 0x%x, session %p, rx %u, tx %u, host_failed %u\n",
           jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], 
           session, session->rx_pid, session->tx_pid, sc->host->host_failed);

    if (in_interrupt()) {
        printk("iSCSI: eh_device_reset in interrupt at %lu for command %p for (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        drop_reference(session);
        return FAILED;
    }
    
    RELEASE_MIDLAYER_LOCK(host);

    if (!wait_for_session(session, TRUE)) {
        printk("iSCSI: eh_device_reset failed waiting for session %p at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
               session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        goto done;
    }

    /* see if we have a task for this command */
    spin_lock(&session->task_lock);
    current_gen = session->generation;
    if ((task = remove_task_for_cmnd(&session->aborted_tasks, sc))) {
        /* had an abort sent and confirmed already */
        task->mgmt_itt = 0;
    }
    else if ((task = remove_task_for_cmnd(&session->rx_abort_tasks, sc))) {
        /* had an abort sent already */
        task->mgmt_itt = 0;
    }
    else if ((task = remove_task_for_cmnd(&session->tx_abort_tasks, sc))) {
        /* waiting to have an abort sent */
        task->mgmt_itt = 0;
    }
    else if ((task = remove_task_for_cmnd(&session->rx_tasks, sc))) {
        /* it's waiting for data, an R2T, or a command response */
        task->mgmt_itt = 0;
    }
    else if ((task = remove_task_for_cmnd(&session->tx_tasks, sc))) {
        /* it's received an R2T, and is queued to have data sent */
        task->mgmt_itt = 0;
    }
    else if ((task = remove_task_for_cmnd(&session->completing_tasks, sc))) {
        /* already received command completion */
        task->mgmt_itt = 0;
    }
    spin_unlock(&session->task_lock);
    
    printk("iSCSI: eh_device_reset at %lu starting target reset for command %p for (%u %u %u %u), Cmd 0x%x\n",
           jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);

    if (warm_reset_target(session)) {
        printk("iSCSI: eh_device_reset succeeded at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);

        ret = SUCCESS;

        /* wait for a session if we don't have one, since we need one for
         * the midlayer to get a TUR response, which it must get next,
         * or else the code in scsi_error.c will treat our SUCCESS as FAILED.
         */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits))
            wait_for_session(session, TRUE);
    }
    else {
        printk("iSCSI: eh_device_reset failed at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    }

 done:
    drop_reference(session);
    REACQUIRE_MIDLAYER_LOCK(host);
    return ret;
}

/* NOTE: due to bugs in the linux SCSI layer (scsi_unjam_host), it's
 * possible for this handler to be called even if the device_reset
 * handler completed all the failed commands back to the SCSI layer
 * with DID_RESET and returned SUCCESS.  To compensate for this, we
 * must ensure that this reset handler doesn't actually care whether
 * the command is still in the driver.  Just find the session
 * associated with the command, and reset it.  
 */
int iscsi_eh_bus_reset( Scsi_Cmnd *sc ) 
{ 
    struct Scsi_Host *host = NULL; 
    iscsi_hba_t *hba = NULL; 
    iscsi_session_t *session;
    int ret = FAILED;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: bus reset, no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: bus reset, no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't reset device for cmnd %p, no session\n", sc);
        return FAILED;
    }

    /* FIXME: The SCSI midlayer isn't very smart about error handling,
     * and may ask for a reset of the same target once for every
     * failed command, even though only one reset is necessary.  We do
     * want to allow resets with no commands outstanding though, as
     * someday they may be needed to clear reservations.  Find a way
     * to determine if we're in error recovery (host blocked?), and if
     * we are, make sure we do at most one reset of this type if no
     * commands are outstanding.
     */

    printk("iSCSI: eh_bus_reset about to reset target at %lu for command %p to (%u %u %u %u), Cmd 0x%x, "
           "session %p, rx %u, tx %u host_failed %u\n",
           jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], 
           session, session->rx_pid, session->tx_pid, sc->host->host_failed);
    
#if TEST_HOST_RESETS
    /* just always fail it. occasionally failing it is too much work to code */
    printk("iSCSI: ignoring attempt to bus reset, command %p to (%u %u %u %u), Cmd 0x%x, host_failed %u\n",
           sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], sc->host->host_failed);
    drop_reference(session);
    return FAILED;
#endif

    if (in_interrupt()) {
        printk("iSCSI: eh_bus_reset failing, in interrupt at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        drop_reference(session);
        return FAILED;
    }

    RELEASE_MIDLAYER_LOCK(host);

    if (!wait_for_session(session, TRUE)) {
        printk("iSCSI: eh_bus_reset failed waiting for session %p at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
               session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        goto done;
    }

    if (warm_reset_target(session)) {
        printk("iSCSI: eh_bus_reset succeeded at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);

        ret = SUCCESS;

        /* wait for a session if we don't have one, since we need one for
         * the midlayer to get a TUR response, which it must get next,
         * or else the code in scsi_error.c will treat our SUCCESS as FAILED.
         */
        if (!test_bit(SESSION_ESTABLISHED, &session->control_bits))
            wait_for_session(session, TRUE);
    }
    else {
        printk("iSCSI: eh_bus_reset failed at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    }

 done:
    drop_reference(session);
    REACQUIRE_MIDLAYER_LOCK(host);
    return ret;
}



int
iscsi_eh_host_reset( Scsi_Cmnd *sc )
{
    struct Scsi_Host *host = NULL;
    iscsi_hba_t *hba = NULL;
    iscsi_session_t *session;

    if ( ! sc ) {
        return FAILED;
    }
    host = sc->host;
    if (! host) {
        printk("iSCSI: host reset, no host for SCSI command %p\n", sc);
        return FAILED;
    }
    hba = (iscsi_hba_t *)host->hostdata;
    if (!hba) {
        printk("iSCSI: host reset, no iSCSI HBA associated with SCSI command %p\n", sc);
        return FAILED;
    }

    /* find the appropriate session for the command */
    session = find_session_for_cmnd(sc);
    if (!session) {
        printk("iSCSI: can't reset device for cmnd %p, no session\n", sc);
        return FAILED;
    }

    /* FIXME: The SCSI midlayer isn't very smart about error handling,
     * and may ask for a reset of the same target once for every
     * failed command, even though only one reset is necessary.  We do
     * want to allow resets with no commands outstanding though, as
     * someday they may be needed to clear reservations.  Find a way
     * to determine if we're in error recovery (host blocked?), and if
     * we are, make sure we do at most one reset of this type if no
     * commands are outstanding.
     */

    printk("iSCSI: eh_host_reset about to reset target at %lu for command %p for (%u %u %u %u), Cmd 0x%x, "
           "session %p, rx %u, tx %u, host_failed %u\n",
           jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], 
           session, session->rx_pid, session->tx_pid, sc->host->host_failed);

    if (in_interrupt()) {
        printk("iSCSI: eh_host_reset failing, in interrupt at %lu for command %p to (%u %u %u %u), Cmd 0x%x, host_failed %u\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], sc->host->host_failed);
        drop_reference(session);
        return FAILED;
    }

    RELEASE_MIDLAYER_LOCK(host);

    if (!wait_for_session(session, TRUE)) {
        printk("iSCSI: eh_host_reset failed waiting for session %p at %lu for command %p to (%u %u %u %u), Cmd 0x%x\n",
               session, jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        drop_reference(session);
        REACQUIRE_MIDLAYER_LOCK(host);
        return FAILED;
    }

    /* ok, warm reset must not have worked, try a cold reset */
    if (cold_reset_target(session)) {
        printk("iSCSI: eh_host_reset succeeded at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
               jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        drop_reference(session);
        REACQUIRE_MIDLAYER_LOCK(host);
        return SUCCESS;
    }

    printk("iSCSI: eh_host_reset failed at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
           jiffies, sc, sc->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
    drop_reference(session);
    REACQUIRE_MIDLAYER_LOCK(host);
    return FAILED;
}


void null_timeout_callback(Scsi_Cmnd * SCpnt)
{
}


/*
 * Replacement for the midlayer's scsi_unjam_host function, which
 * assumes parallel SCSI 2 (no LUN resets), and tends not to handle
 * multiple commands to the same device erroring out at the same time,
 * which is easy to cause with an iSCSI connection drop.
 *
 */
int iscsi_eh_strategy(struct Scsi_Host *host)
{
    Scsi_Cmnd *cmnd, *related;
    Scsi_Device *device;
    Scsi_Cmnd *error_head = NULL;
    Scsi_Cmnd *error_tail = NULL;
    int num_cmnds = 0;
    int devices_failed = 0;
    int recovered;
    iscsi_hba_t *hba = (iscsi_hba_t *)host->hostdata;
    iscsi_session_t *session;
    
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    /* on 2.4 kernels, the io_request_lock won't be held when we get called,
     * but we'll need to acquire it later in order to call scsi_done
     */
    DECLARE_MIDLAYER_FLAGS;
#else
    /* on 2.2 kernels, the eh_thread will be holding the io_request_lock. 
     * Release it, since we have our own locking, and we need to call the scheduler
     * if we're going to make any real attempts to handle errors.
     */
    RELEASE_MIDLAYER_LOCK(host);
#endif

    set_bit(ISCSI_HBA_IN_EH_STRATEGY, &hba->flags);

    /* Go through the list of commands and figure out which ones we
     * need to deal with, similar to how scsi_unjam_host does it. 
     */
    for (device = host->host_queue; device; device = device->next) {
        int device_error = 0;
        
        for (cmnd = device->device_queue; cmnd; cmnd = cmnd->next) {
            if ((cmnd->state == SCSI_STATE_FAILED) || (cmnd->state == SCSI_STATE_TIMEOUT)) {

                DEBUG_EH7("iSCSI: eh_strategy cmnd %p has owner 0x%x, state 0x%x, serial %lu, at_timeout %lu, "
                          "scribble %p, error stage 0x%x\n", 
                          cmnd, cmnd->owner, cmnd->state, cmnd->serial_number, cmnd->serial_number_at_timeout,
                          cmnd->host_scribble, CMND_ERROR_STAGE(cmnd));

                /* build a list of cmnds that have failed or timed out
                 * We could probably use bh_next and eh_state, but
                 * don't, to avoid making too many assumptions about
                 * how the midlayer works.  We're supposed to have
                 * free use of the Scsi_Pointer in each command, so we
                 * use that instead.
                 */
                if (error_head) {
                    error_tail->SCp.ptr = (char *)cmnd;
                    error_tail = cmnd;
                }
                else {
                    cmnd->SCp.ptr = NULL;
                    error_head = error_tail = cmnd;
                }

#if FAIL_ERROR_RECOVERY
                CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_OFFLINE;
                cmnd->device->online = 0;
                cmnd->host->host_failed--;
#endif
                
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
                if (cmnd->done_late) {
                    /* scsi_done has already been called once, but it did nothing,
                     * since the timer won the race.
                     */
                    printk("iSCSI: eh_strategy cmnd %p already done (late), result 0x%x, considering recovered\n", 
                           cmnd, cmnd->result);
                    CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_RECOVERED;
                    cmnd->host->host_failed--;
                }
                else
#endif
                if (cmnd->serial_number == 0) {
                    /* scsi_done has already been called, but the timer won the race */
                    /* FIXME: should we still complete this command later? */
                    printk("iSCSI: eh_strategy cmnd %p already done (serial number 0), result 0x%x, considering recovered\n", 
                           cmnd, cmnd->result);
                    CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_RECOVERED;
                    cmnd->host->host_failed--;
                }
                else {
                    /* start with aborts */
                    num_cmnds++;
                    device_error++;
                    CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_ABORT;
                }
            }
        }
        if (device_error > 0) {
            devices_failed++;
        }
    }

    printk("iSCSI: eh_strategy - error recovery in progress for %d commands on %d devices at %lu, host_failed %d\n",
           num_cmnds, devices_failed, jiffies, host->host_failed);
    
    if (num_cmnds == 0)
        goto done;

    /* try to abort every command that has failed or timed out */
    for (cmnd = error_head; cmnd; cmnd = (Scsi_Cmnd *)cmnd->SCp.ptr) {

        if (num_cmnds == 0) goto done;

        DEBUG_EH4("iSCSI: eh_strategy checking cmnd %p stage %d, owner 0x%x, state 0x%x, for abort\n",
                  cmnd, CMND_ERROR_STAGE(cmnd), cmnd->owner, cmnd->state);

        if (CMND_ERROR_STAGE(cmnd) != ISCSI_ERROR_STAGE_ABORT)
            continue;

        session = find_session_for_cmnd(cmnd);
        if (!session || test_bit(SESSION_TERMINATING, &session->control_bits)) {
            /* session terminated already, no point trying more error recovery */
            if (session) {
                printk("iSCSI: eh_strategy found terminating session %p for cmnd %p\n", session, cmnd);
                drop_reference(session);
            }
            else {
                printk("iSCSI: eh_strategy couldn't find a session for cmnd %p\n", cmnd);
            }
            
            for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
                if ((related->channel == cmnd->channel) &&
                    (related->target == cmnd->target) &&
                    (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
                {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_OFFLINE;
                    num_cmnds--;
                    host->host_failed--;
                }
            }
            
            continue;
        }

        if (iscsi_abort_unsent_cmnd(session, cmnd) || iscsi_abort_cmnd(session, cmnd)) {
            DEBUG_EH2("iSCSI: eh_strategy aborted command %p at %lu\n", cmnd, jiffies);
            num_cmnds--;
            cmnd->host->host_failed--;
            CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_RECOVERED;
        }
        else {
            printk("iSCSI: eh_strategy failed to abort command %p with scribble %p, result 0x%x, "
                   "session %p, %u unsent cmnds, %u ignored cmnds, %u tasks\n",
                   cmnd, (void *)cmnd->host_scribble, cmnd->result, 
                   session, atomic_read(&session->num_cmnds), atomic_read(&session->num_ignored_cmnds), 
                   atomic_read(&session->num_active_tasks));

            CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_LUN_RESET;
        }

        drop_reference(session);
    }

    if (num_cmnds == 0)
        goto done;

    printk("iSCSI: eh_strategy trying LUN resets at %lu, %d cmnds, host_failed %d\n",
           jiffies, num_cmnds, host->host_failed);

    /* if commands weren't aborted, try LUN resets of the LUN the command was sent to */
    for (cmnd = error_head; cmnd; cmnd = (Scsi_Cmnd *)cmnd->SCp.ptr) {
        iscsi_task_t *task;

        if (num_cmnds == 0) goto done;

        DEBUG_EH2("iSCSI: eh_strategy checking cmnd %p stage %d for LUN reset\n",
                  cmnd, CMND_ERROR_STAGE(cmnd));

        if (CMND_ERROR_STAGE(cmnd) != ISCSI_ERROR_STAGE_LUN_RESET)
            continue;

        session = find_session_for_cmnd(cmnd);
        if (!session || test_bit(SESSION_TERMINATING, &session->control_bits)) {
            /* session terminated already, no point trying more error recovery */
            if (session) {
                printk("iSCSI: eh_strategy found terminating session %p for cmnd %p\n", session, cmnd);
                drop_reference(session);
            }
            else {
                printk("iSCSI: eh_strategy couldn't find a session for cmnd %p\n", cmnd);
            }
            
            for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
                if ((related->channel == cmnd->channel) &&
                    (related->target == cmnd->target) &&
                    (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
                {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_OFFLINE;
                    num_cmnds--;
                    related->host->host_failed--;
                }
            }
            
            continue;
        }

        /* try to find the task for the cmnd */
        spin_lock(&session->task_lock);
        if ((task = remove_task_for_cmnd(&session->aborted_tasks, cmnd))) {
            /* had an abort sent and confirmed already */
            task->mgmt_itt = 0;
        }
        else if ((task = remove_task_for_cmnd(&session->rx_abort_tasks, cmnd))) {
            /* had an abort sent already */
            task->mgmt_itt = 0;
        }
        else if ((task = remove_task_for_cmnd(&session->tx_abort_tasks, cmnd))) {
            /* waiting to have an abort sent */
            task->mgmt_itt = 0;
        }
        else if ((task = remove_task_for_cmnd(&session->rx_tasks, cmnd))) {
            /* it's waiting for data, an R2T, or a command response */
            task->mgmt_itt = 0;
        }
        else if ((task = remove_task_for_cmnd(&session->tx_tasks, cmnd))) {
            /* it's received an R2T, and is queued to have data sent */
            task->mgmt_itt = 0;
        }
        else if ((task = remove_task_for_cmnd(&session->completing_tasks, cmnd))) {
            /* already received command completion */
            task->mgmt_itt = 0;
        }
        spin_unlock(&session->task_lock);
        
        if (task == NULL) {
            /* blast, we don't have one.  try to get one */
            if ((task = alloc_task(session))) {
                task->scsi_cmnd = cmnd;
                task->itt = 0;
                task->mgmt_itt = 0;
                task->cmdsn = 0;
                wmb();
                DEBUG_EH2("iSCSI: session %p allocated task %p for LUN reset\n", session, task);
            }
        }

        if (task) {
            recovered = iscsi_reset_lun(session, task);
        }
        else {
            printk("iSCSI: session %p couldn't allocate task for LUN reset at %lu\n", session, jiffies);
            recovered = 0;
        }
        drop_reference(session);

        /* if we allocated a task just for the reset, free it */
        if (task && (task->itt == 0) && (task->cmdsn == 0)) {
            spin_lock(&hba->free_task_lock);
            free_task(task);
            spin_unlock(&hba->free_task_lock);
        }


        if (recovered) {
            /* mark remaining unhandled commands to this lun as recovered */
            for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
                if ((related->channel == cmnd->channel) &&
                    (related->target == cmnd->target) &&
                    (related->lun == cmnd->lun) &&
                    (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
                {
                    /* FIXME: can we safely assume this, or should we
                     * check which commands actually were returned to
                     * us by iscsi_reset_lun?
                     */
                    num_cmnds--;
                    related->host->host_failed--;
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_RECOVERED;
                    DEBUG_EH3("iSCSI: eh_strategy LUN reset recovered cmnd %p, num_cmnds %d, host_failed %d\n", 
                              related, num_cmnds, host->host_failed);
                }
            }

            /* tell all devices attached to this LUN that a reset occured */
            for (device = host->host_queue; device; device = device->next) {
                if ((device->channel == cmnd->channel) &&
                    (device->id == cmnd->target) &&
                    (device->lun == cmnd->lun))
                {
                    device->was_reset = 1;
                    device->expecting_cc_ua = 1;
                }
            }
        }
        else {
            /* mark every remaining unhandled command to this target
             * as needing a warm reset.  There's no point doing
             * additional LUN resets to this target, since we're going
             * to send a target reset later anyway.  Some LUN resets
             * to this target may have already occured, and succeeded,
             * so we only mark commands we haven't already marked
             * recovered by a LUN reset, so that we don't screw up the
             * count of recovered commands.  
             */
            for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
                if ((related->channel == cmnd->channel) &&
                    (related->target == cmnd->target) &&
                    (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
                {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_WARM_RESET;
                }
            }
        }
    }

    if (num_cmnds == 0)
        goto done;

    printk("iSCSI: eh_strategy trying warm target resets at %lu, %d cmnds, host_failed %d\n",
           jiffies, num_cmnds, host->host_failed);

    /* if the LUN reset failed, try warm target resets of the target the command was sent to */
    for (cmnd = error_head; cmnd; cmnd = (Scsi_Cmnd *)cmnd->SCp.ptr) {

        if (num_cmnds == 0) goto done;

        if (CMND_ERROR_STAGE(cmnd) != ISCSI_ERROR_STAGE_WARM_RESET)
            continue;

        session = find_session_for_cmnd(cmnd);
        if (!session || test_bit(SESSION_TERMINATING, &session->control_bits)) {
            /* session terminated already, no point trying more error recovery */
            if (session) {
                DEBUG_EH2("iSCSI: eh_strategy found terminating session %p for cmnd %p\n", session, cmnd);
                drop_reference(session);
            }
            else {
                DEBUG_EH1("iSCSI: eh_strategy couldn't find a session for cmnd %p\n", cmnd);
            }
            
            for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
                if ((related->channel == cmnd->channel) &&
                    (related->target == cmnd->target) &&
                    (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
                {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_OFFLINE;
                    num_cmnds--;
                    related->host->host_failed--;
                }
            }
            
            continue;
        }
        
        recovered = warm_reset_target(session);
        drop_reference(session);

        if (recovered) {
            printk("iSCSI: session %p warm target reset succeeded at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
                   session, jiffies, cmnd, cmnd->host->host_no, cmnd->channel, cmnd->target, cmnd->lun, cmnd->cmnd[0]);

            /* tell all devices attached to this target that a reset occured */
            for (device = host->host_queue; device; device = device->next) {
                if ((device->channel == cmnd->channel) &&
                    (device->id == cmnd->target))
                {
                    device->was_reset = 1;
                    device->expecting_cc_ua = 1;
                }
            }
        }

        /* mark every remaining unhandled command to this target one way or the other */
        for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
            if ((related->channel == cmnd->channel) &&
                (related->target == cmnd->target) &&
                (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
            {
                if (recovered) {
                    num_cmnds--;
                    related->host->host_failed--;
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_RECOVERED;
                    DEBUG_EH3("iSCSI: eh_strategy warm target reset recovered cmnd %p, num_cmnds %d, host_failed %d\n", 
                              related, num_cmnds, host->host_failed);
                }
                else {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_COLD_RESET;
                }
            }
        }
    }

    if (num_cmnds == 0)
        goto done;

    printk("iSCSI: eh_strategy trying cold target resets at %lu, %d cmnds, host_failed %d\n",
           jiffies, num_cmnds, host->host_failed);

    /* if the warm target reset failed, try cold target resets of the target the command was sent to */
    for (cmnd = error_head; cmnd; cmnd = (Scsi_Cmnd *)cmnd->SCp.ptr) {
        if (num_cmnds == 0) goto done;

        if (CMND_ERROR_STAGE(cmnd) != ISCSI_ERROR_STAGE_COLD_RESET)
            continue;
        
        session = find_session_for_cmnd(cmnd);
        if (!session || test_bit(SESSION_TERMINATING, &session->control_bits)) {
            /* session terminated already, no point trying more error recovery */
            if (session) {
                DEBUG_EH2("iSCSI: eh_strategy found terminating session %p for cmnd %p\n", session, cmnd);
                drop_reference(session);
            }
            else {
                DEBUG_EH1("iSCSI: eh_strategy couldn't find a session for cmnd %p\n", cmnd);
            }
            
            for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
                if ((related->channel == cmnd->channel) &&
                    (related->target == cmnd->target) &&
                    (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
                {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_OFFLINE;
                    num_cmnds--;
                    related->host->host_failed--;
                }
            }
            
            continue;
        }
        
        recovered = cold_reset_target(session);
        drop_reference(session);

        if (recovered) {
            printk("iSCSI: session %p cold target reset succeeded at %lu for command %p, (%u %u %u %u), Cmd 0x%x\n",
                   session, jiffies, cmnd, cmnd->host->host_no, cmnd->channel, cmnd->target, cmnd->lun, cmnd->cmnd[0]);

            /* tell all devices attached to this target that a reset occured */
            for (device = host->host_queue; device; device = device->next) {
                if ((device->channel == cmnd->channel) &&
                    (device->id == cmnd->target))
                {
                    device->was_reset = 1;
                    device->expecting_cc_ua = 1;
                }
            }
        }

        /* mark every remaining unhandled command to this target one way or the other */
        for (related = cmnd; related; related = (Scsi_Cmnd *)related->SCp.ptr) {
            if ((related->channel == cmnd->channel) &&
                (related->target == cmnd->target) &&
                (CMND_ERROR_STAGE(related) < ISCSI_ERROR_STAGE_OFFLINE))
            {
                if (recovered) {
                    num_cmnds--;
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_RECOVERED;
                    DEBUG_EH3("iSCSI: eh_strategy cold target reset recovered cmnd %p, num_cmnds %d, host_failed %d\n", 
                              related, num_cmnds, host->host_failed);
                }
                else {
                    CMND_ERROR_STAGE(related) = ISCSI_ERROR_STAGE_OFFLINE;
                }
                
                /* either way it's been handled, though offline is really a shot to the head */
                related->host->host_failed--;
            }
        }
    }

 done:
    /* done with error recovery */
    host->in_recovery = 0;
    clear_bit(ISCSI_HBA_IN_EH_STRATEGY, &hba->flags);
    wmb();

    /* FIXME: acquire or reacquire io_request_lock?  */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
    /* on 2.2 kernels, we need to reacquire the lock */
    REACQUIRE_MIDLAYER_LOCK(host);
#else
    /* on 2.4 kernels, the io_request_lock won't be held. 
     * lock it so that we can call the completion callback for the cmnds.
     */
    LOCK_MIDLAYER_LOCK(host);
#endif

    cmnd = error_head; 
    while (cmnd) {
        Scsi_Cmnd *next = (Scsi_Cmnd *)cmnd->SCp.ptr;

        /* decide how to complete it */
        if (CMND_ERROR_STAGE(cmnd) == ISCSI_ERROR_STAGE_RECOVERED) {
            int result;
            
            /* FIXME: this relies heavily on the internals of the
             * midlayer, and may break if the kernel's SCSI midlayer
             * code changes.  It'd be nice if Linux had a
             * well-designed SCSI-3 stack with pluggable transports,
             * instead of a hacked up SCSI-2 stack that assumes
             * parallel SCSI.
             */

            /* since we don't retry the command ourselves like scsi_unjam_host,
             * let the midlayer do an extra retry.  We can't just use DID_SOFT_ERROR,
             * since some 2.4 kernels have started checking the retry count for
             * DID_SOFT_ERROR, unlike the older kernels where it was an unconditional
             * retry.
             */
            result = HOST_BYTE(DID_BUS_BUSY) | STATUS_BYTE(cmnd->result);
            cmnd->result = result;
            cmnd->retries--;

            /* when we complete the command, scsi_done will queue it up for the bottom-half handler.
             * the bottom-half handler will check it with scsi_decide_disposition, and either
             * retry it, or fail it if the retry count is exceeded.
             */
            DEBUG_EH2("iSCSI: eh_strategy retrying command %p at %lu\n", cmnd, jiffies);
        }
        else if (CMND_ERROR_STAGE(cmnd) == ISCSI_ERROR_STAGE_OFFLINE) {
            /* fail the command */
            cmnd->result = HOST_BYTE(DID_NO_CONNECT);
            set_lun_comm_failure(cmnd);

            /* when we complete the command, scsi_done will queue it up for the bottom-half handler.
             * the bottom-half handler will check it with scsi_decide_disposition, and fail it
             * back to the high-level driver.
             */
            DEBUG_EH2("iSCSI: eh_strategy failing command %p at %lu\n", cmnd, jiffies);
        }
        else {
            printk("iSCSI: bug - eh_strategy cmnd %p to (%u %u %u %u) in error stage %d\n", 
                   cmnd, cmnd->host->host_no, cmnd->channel, cmnd->target, cmnd->lun, CMND_ERROR_STAGE(cmnd));
        }
        
        /* pop it off the list */
        cmnd->SCp.ptr = NULL;
        CMND_ERROR_STAGE(cmnd) = ISCSI_ERROR_STAGE_UNKNOWN;

        /* reset the command's state so that scsi_done won't choke on SCSI_STATE_TIMEOUT */
        cmnd->state = SCSI_STATE_BHQUEUE;
        cmnd->eh_state = SUCCESS;

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
        /* use the same vile hack scsi_unjam_host does to make scsi_done happy 
         * with the cmnd. Unless it has a timer running, scsi_done assumes
         * it lost the race with the timer, and does nothing besides set done_late.
         */
        scsi_add_timer(cmnd, 100, null_timeout_callback);
#endif

        /* send it back to the midlayer for further processing */
        cmnd->scsi_done(cmnd);

        /* go on to the next */
        cmnd = next;
    }
    
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    /* done with the lock */
    UNLOCK_MIDLAYER_LOCK(host);
#endif

    if (num_cmnds == 0) {
        printk("iSCSI: eh_strategy recovered from all errors at %lu, host_failed %d\n",
               jiffies, host->host_failed);
    }
    else {
        printk("iSCSI: eh_strategy returning at %lu with num_cmnds %d, host_failed %d\n",
               jiffies, num_cmnds, host->host_failed);
    }

    /* return value currently doesn't mean anything, but Eric
     * Youngdale's docs say to return 0 in case it acquires meaning
     * later.  
     */
    return 0;
}

int iscsi_queue(Scsi_Cmnd *sc, void (*done)(Scsi_Cmnd *))
{
    iscsi_hba_t *hba;
    iscsi_session_t *session = NULL;
    struct Scsi_Host *host;
    DECLARE_NOQUEUE_FLAGS;
        
    host = sc->host;
    if (host == NULL) {
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        printk("iSCSI: queuecommand but no Scsi_Host\n");
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        set_lun_comm_failure(sc);
        done(sc);
        return 0;
    }

    hba = (iscsi_hba_t *)sc->host->hostdata;
    if ( (!hba) || (!test_bit(ISCSI_HBA_ACTIVE, &hba->flags))) {
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        printk("iSCSI: queuecommand but no HBA\n");
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        set_lun_comm_failure(sc);
        done(sc);
        return 0;
    }

    if ( ! iscsi_timer_running ) {
        /* iSCSI coming up or going down, fail the command */
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        DEBUG_QUEUE6("iSCSI: no timer, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
                     sc, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        done(sc);
        return 0;
    }

    if (sc->target >= ISCSI_MAX_TARGET_IDS_PER_CHANNEL) {
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        printk("iSCSI: invalid target id %u, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->target, sc, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        set_lun_comm_failure(sc);
        done(sc);
        return 0;
    }
    if (sc->lun >= ISCSI_MAX_LUNS_PER_TARGET) {
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        printk("iSCSI: invalid LUN %u, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->lun, sc, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        set_lun_comm_failure(sc);
        done(sc);
        return 0;
    }
    /* CDBs larger than 16 bytes require additional header segments, not yet implemented */
    if (sc->cmd_len > ISCSI_MAX_CMD_LEN) {
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        printk("iSCSI: cmd_len %u too large, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->cmd_len, sc, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        set_lun_comm_failure(sc);
        done(sc);
        return 0;
    }
    /* make sure our SG_TABLESIZE limit was respected */
    if (sc->use_sg > ISCSI_MAX_SG) {
        ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
        printk("iSCSI: use_sg %u too large, failing to queue %p to (%u %u %u %u), Cmd 0x%x\n",
               sc->use_sg, sc, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0]);
        sc->result = HOST_BYTE(DID_NO_CONNECT);
        set_lun_comm_failure(sc);
        done(sc);
        return 0;
    }
#ifdef DEBUG
    if (sc->use_sg) {
        int index;
        struct scatterlist *sglist = (struct scatterlist *)sc->request_buffer;
        unsigned int length = 0;
        int bogus = 0;

        /* sanity check the sglist, to make sure the segments have at least bufflen */
        for (index = 0; index < sc->use_sg; index++) {
            length += sglist[index].length;
            if (sglist[index].length == 0)
                bogus = 1;
        }

        if (bogus || (length < sc->request_bufflen)) {
            printk("iSCSI: attempted to queue %p at %lu to (%u %u %u %u), Cmd 0x%x, corrupt sglist, sg length %u, buflen %u\n",
                   sc, jiffies, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0],
                   length, sc->request_bufflen);
            print_cmnd(sc);
            ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->timeout_per_command);
            sc->result = HOST_BYTE(DID_NO_CONNECT);
            set_data_path_failure(sc);
            done(sc);
            return 0;
        }
    }
#endif

#if DEBUG_EH
    if (LOG_ENABLED(ISCSI_LOG_EH) && (sc->cmnd[0] == TEST_UNIT_READY)) {
        printk("iSCSI: queueing TUR %p at %lu to (%u %u %u %u), Cmd 0x%x, cpu%d\n",
               sc, jiffies, hba->host->host_no, sc->channel, sc->target, sc->lun, sc->cmnd[0], smp_processor_id());
    }
    else
#endif
    {
#if DEBUG_QUEUE
        if (LOG_ENABLED(ISCSI_LOG_QUEUE))
            printk("iSCSI: queueing %p to (%u %u %u %u) at %lu, Cmd 0x%x, cpu%d\n",
                   sc, hba->host->host_no, sc->channel, sc->target, sc->lun, jiffies, sc->cmnd[0], smp_processor_id());
#endif
    }
    
    if (hba) {
        SPIN_LOCK_NOQUEUE(&hba->session_lock);

        /* find the right session */
        session = hba->session_list_head;
        while (session) {
            if ((session->channel == sc->channel) && (session->target_id == sc->target)) {

                /* refuse any new commands while a session is terminating */
                if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
                    printk("iSCSI: session %p terminating, failing to queue %p to (%u %u %u %u), Cmd 0x%x to %s\n",
                           session, sc, 
                           session->host_no, sc->channel, sc->target, sc->lun, 
                           sc->cmnd[0], session->log_name);
                    SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
                    ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
                    sc->result = HOST_BYTE(DID_NO_CONNECT);
                    set_lun_comm_failure(sc);
                    done(sc);
                    return 0;
                }


                spin_lock(&session->scsi_cmnd_lock);

#if MULTIPATH_SUPPORT
                /* if the session replacement timeout has expired, fail the command immediately */
                /* FIXME: but first make sure all other commands and tasks for this session
                 * have already been completed with errors, to maintain ordering.
                 */
                if (test_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits)) {
                    printk("iSCSI: session %p replacement timed out, failing to queue %p to (%u %u %u %u), Cmd 0x%x to %s\n",
                           session, sc, 
                           session->host_no, sc->channel, sc->target, sc->lun, 
                           sc->cmnd[0], session->log_name);
                    SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
                    ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
                    sc->result = HOST_BYTE(DID_NO_CONNECT);
                    set_lun_comm_failure(sc);
                    done(sc);
                    return 0;
                }
#endif

#ifdef DEBUG
                /* make sure the command hasn't already been queued */
                {
                    Scsi_Cmnd *search = session->scsi_cmnd_head;
                    while (search) {
                        if (search == sc) {
                            printk("iSCSI: bug - cmnd %p, state %x, eh_state %x, scribble %p is already queued to session %p\n",
                                   sc, sc->state, sc->eh_state, sc->host_scribble, session);
                            print_session_cmnds(session);
                            spin_unlock(&session->scsi_cmnd_lock);
                            SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
                            ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
                            return 1;
                        }
                        search = (Scsi_Cmnd *)search->host_scribble;
                    }
                }
#endif

                /* make sure we can complete it properly later */
                sc->scsi_done = done;
                sc->result = 0;
                
                /* initialize Scsi_Pointer fields that we might use later */
                sc->SCp.ptr = NULL;
                sc->SCp.Status = 0;
                CMND_ERROR_STAGE(sc) = ISCSI_ERROR_STAGE_UNKNOWN;
                sc->host_scribble = NULL;

                /* add it to the session's command queue so the tx thread will send it */
                if (session->scsi_cmnd_head) {
                    /* append at the tail */
                    session->scsi_cmnd_tail->host_scribble = (unsigned char *)sc;
                    session->scsi_cmnd_tail = sc;
                }
                else {
                    /* make it the head */
                    session->scsi_cmnd_head = session->scsi_cmnd_tail = sc;
                }
                atomic_inc(&session->num_cmnds);
#if DEBUG_EH
                if (LOG_ENABLED(ISCSI_LOG_EH) && (sc->cmnd[0] == TEST_UNIT_READY)) {
                    printk("iSCSI: queued TUR %p at %lu to session %p\n", sc, jiffies, session);
                }
                else
#endif
                {
#if DEBUG_SMP || DEBUG_QUEUE
                    if (LOG_ENABLED(ISCSI_LOG_SMP) || LOG_ENABLED(ISCSI_LOG_QUEUE))
                        printk("iSCSI: queued %p to session %p at %lu, %u cmnds, head %p, tail %p\n", 
                               sc, session, jiffies, atomic_read(&session->num_cmnds), 
                               session->scsi_cmnd_head, session->scsi_cmnd_tail);
#endif
                }
                spin_unlock(&session->scsi_cmnd_lock);
                    
                /* unlock, wake the tx thread, and return */
                SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
                ISCSI_TRACE(ISCSI_TRACE_Qd, sc, NULL, sc->retries, sc->timeout_per_command);
                wake_tx_thread(TX_SCSI_COMMAND, session);
                return 0;
            }
            
            session = session->next;
        }
        SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
    }

    /* it's normal for there to be no session while the module is initializing. */
    if (test_bit(0, &init_module_complete)) {
        printk("iSCSI: queuecommand %p failed to find a session for HBA %p, host %d channel %d id %d lun %d\n",
               sc, hba, hba->host->host_no, sc->channel, sc->target, sc->lun);
    }

    /* Complete with an error.  A comment says we shouldn't, but
     * it's actually ok to call done() as long as we're using
     * the eh_ style error handlers.
     */
    ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries, sc->allowed);
    DEBUG_QUEUE1("iSCSI: queuecommand completing %p with DID_NO_CONNECT\n", sc);
    sc->result = HOST_BYTE(DID_NO_CONNECT);

    done(sc);

    return 0;
}

int iscsi_biosparam( Disk *disk, kdev_t dev, int geom[] )
{
#if 0
    /* FIXME: lk 2.5 kdev_t changes */
    DEBUG_INIT3("iSCSI: BiosParam, capacity %d, major %d, minor %d\n",
                disk->capacity, MAJOR(dev), MINOR(dev));
#endif
    geom[0] = 64;                                   /* heads */
    geom[1] = 32;                                   /* sectors */
    geom[2] = disk->capacity / (64*32);     /* cylinders */
    return 1;
}

const char *iscsi_info( struct Scsi_Host *sh )
{
    iscsi_hba_t *hba;
    static char buffer[256];
    char *build_str = BUILD_STR;

    DEBUG_INIT0("iSCSI: Info\n");
    hba = (iscsi_hba_t *)sh->hostdata;
    if ( ! hba ) {
        return NULL;
    }

    memset(buffer, 0, sizeof(buffer) );

    if (build_str) {
        /* developer-built variant of a 4-digit internal release */
        sprintf(buffer, "iSCSI %d.%d.%d.%d%s variant (%s)",
                DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
                DRIVER_EXTRAVERSION, ISCSI_DATE);
    }
    else if (DRIVER_INTERNAL_VERSION > 0) {
        /* 4-digit release */
        sprintf(buffer, "iSCSI %d.%d.%d.%d%s (%s)",
                DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
                DRIVER_EXTRAVERSION, ISCSI_DATE);
    }
    else {
        /* 3-digit release */
        sprintf(buffer, "iSCSI %d.%d.%d%s (%s)",
                DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION,
                DRIVER_EXTRAVERSION, ISCSI_DATE);
    }

    return buffer;
}


static void right_justify(char *buffer, int fieldsize) 
{
    char *src = buffer;
    int length = strlen(buffer);
    int shift;
    char *dst;

    if (length > fieldsize) {
        printk("iSCSI: can't right justify, length %d, fieldsize %d\n", length, fieldsize);
        return;
    }

    shift = (fieldsize - length);

    if ((length > 0) && (shift > 0)) {
        /* memmove it to the right, assuming the buffer is at least one byte longer than fieldsize */
        dst = buffer + fieldsize;
        src = buffer + length;

        /* have to copy 1 byte at a time from the end, to avoid clobbering the source */
        while (src >= buffer) {
            *dst-- = *src--;
        }

        /* fill spaces on the left */
        for (dst = buffer; shift; shift--)
            *dst++ = ' ';
    }
}

/* if current_offset is in [start..finish), add str to the buffer */
static int add_proc_buffer(char *str, off_t start, off_t finish, off_t *current_offset, char **buffer)
{
    int length = strlen(str);
    int leading = 0;

    if (*current_offset + length <= start) {
        /* don't need any of it */
        *current_offset += length;
        return 1;
    }
    else if (*current_offset + length <= finish) {
        /* need everything at or above start */
        if (*current_offset < start)
            leading = start - *current_offset;
        
        strcpy(*buffer, str + leading);
        *buffer += (length - leading);
        *current_offset += length;
        return 1;
    }
    else {
        /* need everything at or above start, and before finish */
        if (*current_offset < start)
            leading = start - *current_offset;

        strncpy(*buffer, str + leading, finish - start);
        *buffer += (finish - start);
        *current_offset = finish;
        return 0; /* no more */
    }
}


/* Each LUN line has:
 * 2 spaces + ((3-digit field + space) * 4) + 4 spaces + 15 char IP address + 2 spaces + 5-digit port + 2 spaces 
 * 2 + 16 + 4 + 15 + 2 + 5 + 2 = fixed length 46 + TargetName + newline
 * the TargetName ends the line, and has a variable length.
 */
static int add_lun_line(iscsi_session_t *session, int lun,
                             off_t start,  off_t finish,  off_t *current_offset, char **buffer)
{
    char str[32]; 

    DEBUG_FLOW6("iSCSI: add_lun_line %p, %d, %lu, %lu, %lu, %p\n",
                session, lun, start, finish, *current_offset, *buffer);

    /* 2 spaces + (4 * (3-digit field + space)) + 4 spaces = 22 chars */
    if (lun >= 0)
        sprintf(str, "  %3.1u %3.1u %3.1u %3.1u    ",
                session->host_no, session->channel, session->target_id, lun);
    else
        sprintf(str, "  %3.1u %3.1u %3.1u   ?    ",
                session->host_no, session->channel, session->target_id);

    if (!add_proc_buffer(str, start, finish, current_offset, buffer))
        return 0;
    
    if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
        /* up to 15 char IP address + 2 spaces = 17 chars */
        sprintf(str, "%u.%u.%u.%u  ",
                session->ip_address[0], session->ip_address[1], session->ip_address[2], session->ip_address[3]);
        right_justify(str, 17);
        
        if (!add_proc_buffer(str, start, finish, current_offset, buffer))
            return 0;
        
        /* 5-digit port + 2 spaces = 7 chars */
        sprintf(str,  "%d  ", session->port);
        right_justify(str, 7);

        if (!add_proc_buffer(str, start, finish, current_offset, buffer))
            return 0;
    }
    else {
        /* fill in '?' for each empty field, so that /proc/scsi/iscsi can easily 
         * be processed by tools such as awk and perl.
         */
        sprintf(str, "              ?      ?  ");
        if (!add_proc_buffer(str, start, finish, current_offset, buffer))
            return 0;
    }

    /* the TargetName, up to 255 chars */
    if (strlen(session->TargetName) > 255)
        printk("iSCSI: /proc read, session %p TargetName too long, %d\n", session, strlen(session->TargetName));

    if (!add_proc_buffer(session->TargetName, start, finish, current_offset, buffer))
        return 0;

    /* and a newline */
    sprintf(str, "\n");
    if (!add_proc_buffer(str, start, finish, current_offset, buffer))
        return 0;
    
    /* keep going */
    return 1;
}

/* Show LUNs for every session on the HBA.  The parameters tell us
 * what part of the /proc "file" we're supposed to put in the buffer.
 * This is somewhat broken, since the data may change in between the
 * multiple calls to this function, since we can't keep holding the
 * locks.  Implemented by throwing away bytes that we would have
 * written to the buffer until we reach <start>, and then putting
 * everything before <finish> in the buffer.  Return when we run
 * out of data, or out of buffer.  We avoid oddities in the output,
 * we must ensure that the size of the output doesn't vary while /proc
 * is being read.  We can easily make each line a fixed size, but
 * if the number of LUNs or sessions varies during a /proc read,
 * the user loses.
 */
static int show_session_luns(iscsi_session_t *session,
                             off_t start,  off_t finish,  off_t *current_offset, char **buffer)
{
    /* if we've already found LUNs, show them all */
    int lfound = 0;
    int l;

    /* FIXME: IPv6 */
    if (session->ip_length != 4)
        return 1;

    for (l=0; l<ISCSI_MAX_LUN; l++) {
        if (test_bit(l, session->luns_activated)) {
            lfound += 1;
            
            if (!add_lun_line(session, l, start, finish, current_offset, buffer)) {
                DEBUG_FLOW2("iSCSI: show session luns returning 0 with current offset %lu, buffer %p\n", 
                       *current_offset, buffer);
                return 0;
            }
        }
    }

    /* if we haven't found any LUNs, use ? for a LUN number */
    if ( ! lfound ) {
        if (!add_lun_line(session, -1, start, finish, current_offset, buffer)) {
            DEBUG_FLOW2("iSCSI: show session luns returning 0 with current offset %lu, buffer %p\n", 
                        *current_offset, buffer);
            return 0;
        }
    }
    else {
        DEBUG_FLOW2("iSCSI: show session luns returning 1 with current offset %lu, buffer %p\n", 
                    *current_offset, buffer);
    }

    return 1;
}



/* returns number of bytes matched */
static int find_keyword(char *start, char *end, char *key)
{
    char *ptr = start;
    int key_len = strlen(key);

    /* skip leading whitespace */
    while ((ptr < end) && is_space(*ptr))
        ptr++;

    /* compare */
    if (((end - ptr) == key_len) && !memcmp(key, ptr, key_len)) {
        return (ptr - start) + key_len;
    }
    else if (((end - ptr) > key_len) && !memcmp(key, ptr, key_len) && is_space(ptr[key_len])) {
        return (ptr - start) + key_len;
    }
    else {
        return 0;
    }
}

static int find_number(char *start, char *end, int *number) 
{
    char *ptr = start;
    int found = 0;
    int acc = 0;

    /* skip leading whitespace */
    while ((ptr < end) && is_space(*ptr))
        ptr++;
    
    while (ptr < end) {
        if (is_space(*ptr)) {
            break;
        }
        else if (is_digit(*ptr)) {
            found = 1;
            acc = (acc * 10) + (*ptr - '0');
        }
        else {
            /* something bogus */
            return 0;
        }
        ptr++;
    }

    if (found) {
        if (number)
            *number = acc;
        return (ptr - start);
    }
    else
        return 0;
}

/*
 * *buffer: I/O buffer
 * **start: for user reads, driver can report where valid data starts in the buffer
 * offset: current offset into a /proc/scsi/iscsi/[0-9]* file
 * length: length of buffer
 * hostno: Scsi_Host host_no
 * write: TRUE - user is writing; FALSE - user is reading
 *
 * Return the number of bytes read from or written to a
 * /proc/scsi/iscsi/[0-9]* file.
 */

int iscsi_proc_info( char *buffer,
                     char **start,
                     off_t offset,
                     int length,
                     int hostno,
                     int write)
{
    char *bp = buffer;
    iscsi_hba_t *hba;
    iscsi_session_t *session;
    unsigned int host = 0, channel = 0, target = 0;
    DECLARE_NOQUEUE_FLAGS;

    if (!buffer)
        return -EINVAL;

    if (write) {
        int cmd_len;
        char *end = buffer + length;

        if ((cmd_len = find_keyword(bp, end, "log"))) {
            unsigned int log_setting = 0;
            
            bp += cmd_len;
            
            if ((cmd_len = find_keyword(bp, end, "all")) != 0) {
                iscsi_log_settings = 0xFFFFFFFF;
                printk("iSCSI: all logging enabled\n");
            }
            else if ((cmd_len = find_keyword(bp, end, "none")) != 0) {
                iscsi_log_settings = 0;
                printk("iSCSI: all logging disabled\n");
            }
            else if ((cmd_len = find_keyword(bp, end, "sense")) != 0) {
                log_setting = ISCSI_LOG_SENSE;
                bp += cmd_len;
                
                if ((cmd_len = find_keyword(bp, end, "always")) != 0) {
                    iscsi_log_settings |= LOG_SET(log_setting);
                    printk("iSCSI: log sense always\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "on")) != 0) {
                    iscsi_log_settings |= LOG_SET(log_setting);
                    printk("iSCSI: log sense yes\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "yes")) != 0) {
                    iscsi_log_settings |= LOG_SET(log_setting);
                    printk("iSCSI: log sense yes\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "1")) != 0) {
                    iscsi_log_settings |= LOG_SET(log_setting);
                    printk("iSCSI: log sense 1\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "minimal")) != 0) {
                    iscsi_log_settings &= ~LOG_SET(log_setting);
                    printk("iSCSI: log sense off\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "off")) != 0) {
                    iscsi_log_settings &= ~LOG_SET(log_setting);
                    printk("iSCSI: log sense off\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "no")) != 0) {
                    iscsi_log_settings &= ~LOG_SET(log_setting);
                    printk("iSCSI: log sense no\n");
                }
                else if ((cmd_len = find_keyword(bp, end, "0")) != 0) {
                    iscsi_log_settings &= ~LOG_SET(log_setting);
                    printk("iSCSI: log sense 0\n");
                }
            }
            else {
                if ((cmd_len = find_keyword(bp, end, "login")) != 0) {
                    log_setting = ISCSI_LOG_LOGIN;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "init")) != 0) {
                    log_setting = ISCSI_LOG_INIT;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "queue")) != 0) {
                    log_setting = ISCSI_LOG_QUEUE;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "alloc")) != 0) {
                    log_setting = ISCSI_LOG_ALLOC;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "flow")) != 0) {
                    log_setting = ISCSI_LOG_FLOW;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "error")) != 0) {
                    log_setting = ISCSI_LOG_ERR;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "eh")) != 0) {
                    log_setting = ISCSI_LOG_EH;
                    bp += cmd_len;
                }
                else if ((cmd_len = find_keyword(bp, end, "smp")) != 0) {
                    log_setting = ISCSI_LOG_SMP;
                    bp += cmd_len;
                }
                
                if (log_setting) {
                    if ((cmd_len = find_keyword(bp, end, "on")) != 0) {
                        iscsi_log_settings |= LOG_SET(log_setting);
                    }
                    else if ((cmd_len = find_keyword(bp, end, "yes")) != 0) {
                        iscsi_log_settings |= LOG_SET(log_setting);
                    }
                    else if ((cmd_len = find_keyword(bp, end, "1")) != 0) {
                        iscsi_log_settings |= LOG_SET(log_setting);
                    }
                    else if ((cmd_len = find_keyword(bp, end, "off")) != 0) {
                        iscsi_log_settings &= ~LOG_SET(log_setting);
                    }
                    else if ((cmd_len = find_keyword(bp, end, "no")) != 0) {
                        iscsi_log_settings &= ~LOG_SET(log_setting);
                    }
                    else if ((cmd_len = find_keyword(bp, end, "0")) != 0) {
                        iscsi_log_settings &= ~LOG_SET(log_setting);
                    }
                }
            }

            printk("iSCSI: log settings %8x\n", iscsi_log_settings);
            mb();
        }
        else if ((cmd_len = find_keyword(bp, end, "shutdown"))) {
            /* try to shutdown the driver */
            if (!iscsi_shutdown()) {
                printk("iSCSI: driver shutdown failed\n");
            }
        }
        else if ((cmd_len = find_keyword(bp, end, "target"))) {
            bp += cmd_len;

            if ((cmd_len = find_number(bp, end, &host)) == 0) {
                printk("iSCSI: /proc/scsi/iscsi couldn't determine host number of session\n");
                return length;
            }
            bp += cmd_len;

            if ((cmd_len = find_number(bp, end, &channel)) == 0) {
                printk("iSCSI: /proc/scsi/iscsi couldn't determine channel number of session\n");
                return length;
            }
            bp += cmd_len;

            if ((cmd_len = find_number(bp, end, &target)) == 0) {
                printk("iSCSI: /proc/scsi/iscsi couldn't determine target id number of session\n");
                return length;
            }
            bp += cmd_len;
            
            session = find_session_by_channel(host, channel, target);
            if (session) {
                if ((cmd_len = find_keyword(bp, end, "nop"))) {
                    unsigned int data_length = 0;

                    bp += cmd_len;
                    if ((cmd_len = find_number(bp, end, &data_length)) == 0) {
                        printk("iSCSI: session %p Nop test couldn't determine data length\n", session);
                        return length;
                    }
                    bp += cmd_len;

                    if (data_length) {
                        printk("iSCSI: session %p for (%u %u %u *), %d byte Nop data test requested at %lu\n", 
                               session, host, channel, target, data_length, jiffies);
                        iscsi_ping_test_session(session, data_length);
                    }
                    else {
                        printk("iSCSI: session %p for (%u %u %u *), 0 byte Nop data test ignored at %lu\n", 
                               session, host, channel, target, jiffies);
                    }
                }
                else if ((cmd_len = find_keyword(bp, end, "portal"))) {
                    unsigned int portal = 0;

                    bp += cmd_len;
                    if ((cmd_len = find_number(bp, end, &portal)) == 0) {
                        printk("iSCSI: /proc/scsi/iscsi session %p for (%u %u %u *) no portal specified\n",
                               session, host, channel, target);
                        return length;
                    }
                    bp += cmd_len;

                    spin_lock(&session->portal_lock);
                    if (portal < session->num_portals) {
                        session->requested_portal = portal;
                        printk("iSCSI: /proc/scsi/iscsi session %p for (%u %u %u *) requesting switch to portal %u at %lu\n", 
                               session, host, channel, target, portal, jiffies);
                        
                        /* request a logout for the current session */
                        spin_lock(&session->task_lock);
                        iscsi_request_logout(session, session->active_timeout, session->active_timeout);
                        spin_unlock(&session->task_lock);
                    }
                    else {
                        printk("iSCSI: /proc/scsi/iscsi session %p for (%u %u %u *) can't switch to portal %u, only %d portals\n",
                               session, host, channel, target, portal, session->num_portals);
                    }
                    spin_unlock(&session->portal_lock);
                }
                else if ((cmd_len = find_keyword(bp, end, "logout"))) {
                    unsigned int deadline = 0;
                    unsigned int response_deadline = 0;

                    bp += cmd_len;
                    if ((cmd_len = find_number(bp, end, &deadline)) == 0) {
                        deadline = session->active_timeout;
                    }
                    bp += cmd_len;

                    if ((cmd_len = find_number(bp, end, &response_deadline)) == 0) {
                        response_deadline = session->active_timeout;
                    }
                    bp += cmd_len;

                    printk("iSCSI: /proc/scsi/iscsi session %p for (%u %u %u *) requesting logout at %lu, "
                           "logout deadline %u seconds, response deadline %u seconds\n",
                           session, host, channel, target, jiffies, deadline, response_deadline);
                    
                    /* request a logout for the current session */
                    spin_lock(&session->task_lock);
                    iscsi_request_logout(session, session->active_timeout, session->active_timeout);
                    spin_unlock(&session->task_lock);
                }
                else if ((cmd_len = find_keyword(bp, end, "drop"))) {
                    printk("iSCSI: /proc/scsi/iscsi dropping session %p for (%u %u %u *) at %lu\n", 
                           session, host, channel, target, jiffies);
                    iscsi_drop_session(session);
                }
                else if ((cmd_len = find_keyword(bp, end, "terminate"))) {
                    printk("iSCSI: /proc/scsi/iscsi terminating session %p for (%u %u %u *) at %lu\n", 
                           session, host, channel, target, jiffies);
                    iscsi_terminate_session(session);
                }
                else {
                    printk("iSCSI: /proc/scsi/iscsi session %p (%u %u %u *) unknown command\n",
                           session, host, channel, target);
                }

                /* done with the session */
                drop_reference(session);
            }
            else {
                printk("iSCSI: /proc/scsi/iscsi failed to find session (%u %u %u *)\n", host, channel, target);
            }
        }
        
        /* FIXME: some SCSI read and write tests would be useful as
         * well.  Allow user to specify read6, write6, read10,
         * write10, specify the command queue size (max outstanding),
         * total number of commands, command buffer size, starting
         * block offset, and block increment per command.  This should
         * let us do sequential or fixed offset I/O tests, and try to
         * determine throughput without having do worry about what the
         * SCSI layer or applications above us are capable of.  Also,
         * consider a flag that controls writing/validating a data
         * pattern.  We don't always want to do it, but it may be
         * useful sometimes.
         */
        return length;
    }
    else {
        /* it's a read */
        char *build_str = BUILD_STR;
        off_t current_offset = 0;
        off_t finish = offset + length;

        DEBUG_FLOW5("iSCSI: /proc read, buffer %p, start %p, offset %lu, length %d, hostno %d\n",
                    buffer,  start,  offset, length, hostno);
        
        /* comment header with version number */
        /* FIXME: we assume the buffer length is always large enough for our header */
        if (build_str) {
            /* developer-built variant of a 4-digit internal release */
            bp += sprintf(bp, "# iSCSI driver version: %d.%d.%d.%d%s variant (%s)\n#\n",
                          DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
                          DRIVER_EXTRAVERSION, ISCSI_DATE);
        }
        else if (DRIVER_INTERNAL_VERSION > 0) {
            /* 4-digit release */
            bp += sprintf(bp, "# iSCSI driver version: %d.%d.%d.%d%s (%s)\n#\n",
                          DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
                          DRIVER_EXTRAVERSION, ISCSI_DATE);
        }
        else {
            /* 3-digit release */
            bp += sprintf(bp, "# iSCSI driver version: %d.%d.%d%s (%s)\n#\n",
                          DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION,
                          DRIVER_EXTRAVERSION, ISCSI_DATE);
        }
        bp += sprintf(bp,"# SCSI:                   iSCSI:\n");
        bp += sprintf(bp,"# Hst Chn Tgt Lun         IP address   Port  TargetName\n");
        
        *start = buffer;
        current_offset = bp - buffer;

        if (offset >= current_offset) {
            bp = buffer; /* don't need any of that header, toss it all */
            DEBUG_FLOW2("iSCSI: /proc skipping header, current offset %lu, buffer %p\n", current_offset, bp);
        }
        else if (offset != 0) {
            /* need only some of the header */
            char *src = buffer + offset;
            char *dst = buffer;
            
            /* memmove what we need to the beginning of the buffer, since we may have
             * to return the whole buffer length to avoid prematurely indicating EOF.
             */
            while ((*dst++ = *src++))
                ;
            
            bp = dst;

            DEBUG_FLOW2("iSCSI: /proc partial header, current offset %lu, buffer %p\n", current_offset, bp);
        }
        else {
            DEBUG_FLOW2("iSCSI: /proc full header, current offset %lu, buffer %p\n", current_offset, bp);
        }

        /* find the HBA corresponding to hostno */
        spin_lock(&iscsi_hba_list_lock);
        hba = iscsi_hba_list;
        while (hba && hba->host->host_no != hostno)
            hba = hba->next;
        spin_unlock(&iscsi_hba_list_lock);
        
        if (hba) {
            SPIN_LOCK_NOQUEUE(&hba->session_lock);
            session = hba->session_list_head;
            while (session) {
                if (!show_session_luns(session, offset, finish, &current_offset, &bp))
                    break;
                session = session->next;
            }
            SPIN_UNLOCK_NOQUEUE(&hba->session_lock);
        }
        else {
            printk("iSCSI: /proc read couldn't find HBA #%d\n", hostno);
        }

        /* tell the caller about the output */
        if (current_offset <= offset) {
            /* return no valid data if the desired offset is beyond the total "file" length */
            DEBUG_FLOW2("iSCSI: /proc read returning 0 of %ld (EOF), start %p\n", finish - offset, *start);
            return 0;
        }
        
        /* return how much valid data is in the buffer */
        DEBUG_FLOW3("iSCSI: /proc read returning %ld of %ld, start %p\n", current_offset - offset, finish - offset, *start);
        return (current_offset - offset);
    }
}

/*
 * We cannot include scsi_module.c because the daemon has not got a connection
 * up yet.
 */
static int
ctl_open( struct inode *inode, struct file *file )
{
    MOD_INC_USE_COUNT;
    return 0;
}

static int
ctl_close( struct inode *inode, struct file *file )
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static int
ctl_ioctl( struct inode *inode,
           struct file *file,
           unsigned int cmd,
           unsigned long arg)
{
    int rc;

    if (cmd == ISCSI_ESTABLISH_SESSION) {
        iscsi_session_ioctl_t *ioctld = kmalloc(sizeof(*ioctld), GFP_KERNEL);
        iscsi_session_t *session;
        iscsi_portal_info_t *portals = NULL;
        
        if (!ioctld) {
            printk("iSCSI: couldn't allocate space for session ioctl data\n");
            return -ENOMEM;
        }
        if (copy_from_user(ioctld, (void *)arg, sizeof(*ioctld))) {
            printk("iSCSI: Cannot copy session ioctl data\n");
            kfree(ioctld);
            return -EFAULT;
        }
        if (ioctld->ioctl_size != sizeof(iscsi_session_ioctl_t)) {
            printk("iSCSI: ioctl size %u incorrect, expecting %u\n", ioctld->ioctl_size, sizeof(*ioctld));
            kfree(ioctld);
            return -EINVAL;
        }
        if (ioctld->ioctl_version != ISCSI_SESSION_IOCTL_VERSION) {
            printk("iSCSI: ioctl version %u incorrect, expecting %u\n", ioctld->ioctl_version, ISCSI_SESSION_IOCTL_VERSION);
            kfree(ioctld);
            return -EINVAL;
        }
        if (ioctld->portal_info_size != sizeof(iscsi_portal_info_t)) {
            printk("iSCSI: ioctl portal info size %u incorrect, expecting %u\n", ioctld->portal_info_size, sizeof(*portals));
            kfree(ioctld);
            return -EINVAL;
        }
        if (ioctld->num_portals == 0) {
            printk("iSCSI: ioctl has no portals\n");
            kfree(ioctld);
            return -EINVAL;
        }
        
        /* allocate the portals */
        portals = (iscsi_portal_info_t *)kmalloc(ioctld->num_portals * sizeof(*portals), GFP_KERNEL);
        if (portals == NULL) {
            printk("iSCSI: cannot allocate %d portals\n", ioctld->num_portals);
            kfree(ioctld);
            return -ENOMEM;
        }
        DEBUG_INIT5("iSCSI: allocated portals %p (size %u) for bus %d target %d at %lu\n", 
                    portals, ioctld->num_portals * sizeof(*portals), ioctld->iscsi_bus, ioctld->target_id, jiffies);
        memset(portals, 0, ioctld->num_portals * sizeof(*portals));
        
        /* copy the portal info from the user ioctl structure */
        if (copy_from_user(portals, (void *)arg + offsetof(struct iscsi_session_ioctl, portals), 
                           ioctld->num_portals * sizeof(*portals)))
        {
            printk("iSCSI: cannot copy portal info, ioctl %p, size %d, portals %p\n",
                   ioctld, sizeof(*portals), (void *)arg + offsetof(struct iscsi_session_ioctl, portals));
            kfree(ioctld);
            kfree(portals);
            return -EFAULT;
        }
        else {
            DEBUG_ALLOC3("iSCSI: copied %d bytes of portal info from %p to %p\n",
                         ioctld->num_portals * sizeof(*portals), (void *)arg + offsetof(struct iscsi_session_ioctl, portals), portals);
        }

        /* allocate a session structure */
        session = (iscsi_session_t *)kmalloc(sizeof(*session), GFP_KERNEL);
        if (session == NULL) {
            printk("iSCSI: cannot allocate new session\n");
            kfree(ioctld);
            kfree(portals);
            return -ENOMEM;
        }
        memset(session, 0, sizeof(*session));
        atomic_set(&session->refcount, 1);
        DEBUG_INIT5("iSCSI: allocated session %p (size %u) for bus %d target %d at %lu\n", 
                    session, sizeof(*session), ioctld->iscsi_bus, ioctld->target_id, jiffies);

        /* unless we already have one, start a timer thread */
        if (!test_and_set_bit(0, &iscsi_timer_running)) {
            DEBUG_INIT1("iSCSI: starting timer thread at %lu\n", jiffies);
            kernel_thread(iscsi_timer_thread, NULL, 0);
        }
        
        /* iscsi_session will dec the refcount when it finishes with the session */
        rc = iscsi_session(session, ioctld, portals);

        if (copy_to_user((void *)arg, ioctld, sizeof(*ioctld)) ) {
            printk("iSCSI: failed to copy ioctl data back to user mode for session %p\n", session);
        }
        kfree(ioctld);

        return rc;
    }
    else if ( cmd == ISCSI_SHUTDOWN ) {
        return iscsi_shutdown();
    }
    else if ( cmd == ISCSI_RESET_PROBING ) {
        return iscsi_reset_lun_probing();
    }
    else if ( cmd == ISCSI_PROBE_LUNS ) {
        iscsi_session_t *session = NULL;
        iscsi_probe_luns_ioctl_t *ioctld = kmalloc(sizeof(*ioctld), GFP_KERNEL);
        
        if (!ioctld) {
            printk("iSCSI: couldn't allocate space for probe ioctl data\n");
            return -ENOMEM;
        }
        if (copy_from_user(ioctld, (void *)arg, sizeof(*ioctld)) ) {
            printk("iSCSI: Cannot copy session ioctl data\n");
            kfree(ioctld);
            return -EFAULT;
        }
        if (ioctld->ioctl_size != sizeof(*ioctld)) {
            printk("iSCSI: ioctl size %u incorrect, expecting %u\n", ioctld->ioctl_size, sizeof(*ioctld));
            kfree(ioctld);
            return -EINVAL;
        }
        if (ioctld->ioctl_version != ISCSI_PROBE_LUNS_IOCTL_VERSION) {
            printk("iSCSI: ioctl version %u incorrect, expecting %u\n", ioctld->ioctl_version, ISCSI_PROBE_LUNS_IOCTL_VERSION);
            kfree(ioctld);
            return -EINVAL;
        }
        
        rc = 0;
        
        /* find the session */
        session = find_session_by_bus(ioctld->iscsi_bus, ioctld->target_id);
        if (session == NULL) {
            printk("iSCSI: ioctl probe LUNs (bus %d, target %d) failed, no session\n", 
                   ioctld->iscsi_bus, ioctld->target_id);
        }

        if (signal_pending(current))
            goto done;
        if (test_bit(SESSION_TERMINATING, &session->control_bits))
            goto done;

        DEBUG_INIT3("iSCSI: ioctl triggering LUN probe for session %p (bus %d, target %d)\n", 
                    session, ioctld->iscsi_bus, ioctld->target_id);
        
        iscsi_detect_luns(session);

        if (signal_pending(current))
            goto done;
        if (test_bit(SESSION_TERMINATING, &session->control_bits))
            goto done;
        
        iscsi_probe_luns(session, ioctld->lun_bitmap, &ioctld->device_info);
        if (signal_pending(current))
            goto done;
        if (test_bit(SESSION_TERMINATING, &session->control_bits))
            goto done;

        rc = 1;

    done:
        if (session)
            drop_reference(session);
        kfree(ioctld);
        return rc;
    }
    else if ( cmd == ISCSI_TERMINATE_SESSION ) {
        iscsi_terminate_session_ioctl_t ioctld;
        iscsi_session_t *session = NULL;

        if ( copy_from_user(&ioctld, (void *)arg, sizeof(ioctld)) ) {
            printk("iSCSI: Cannot copy session ioctl data\n");
            return -EFAULT;
        }
        if (ioctld.ioctl_size != sizeof(ioctld)) {
            printk("iSCSI: ioctl size %u incorrect, expecting %u\n", ioctld.ioctl_size, sizeof(ioctld));
            return -EINVAL;
        }
        if (ioctld.ioctl_version != ISCSI_TERMINATE_SESSION_IOCTL_VERSION) {
            printk("iSCSI: ioctl version %u incorrect, expecting %u\n", ioctld.ioctl_version, ISCSI_TERMINATE_SESSION_IOCTL_VERSION);
            return -EINVAL;
        }
        
        /* find the session */
        session = find_session_by_bus(ioctld.iscsi_bus, ioctld.target_id);
        if (session) {
            printk("iSCSI: ioctl terminating session %p (bus %d, target %d)\n", session, ioctld.iscsi_bus, ioctld.target_id);
            iscsi_terminate_session(session);
            drop_reference(session);
            return 1;
        }
        else {
            printk("iSCSI: terminate session ioctl (bus %d, target %d) failed, no session\n", ioctld.iscsi_bus, ioctld.target_id);
            return 0;
        }
    }
    else if ( cmd == ISCSI_GETTRACE ) {
#if DEBUG_TRACE
        iscsi_trace_dump_t dump;
        iscsi_trace_dump_t *user_dump;
        int rc;
        DECLARE_NOQUEUE_FLAGS;
        
        user_dump = (iscsi_trace_dump_t *)arg;
        if (copy_from_user(&dump, user_dump, sizeof(dump))) {
            printk("iSCSI: trace copy_from_user %p, %p, %u failed\n", 
                   &dump, user_dump, sizeof(dump));
            return -EFAULT;
        }

        if (dump.dump_ioctl_size != sizeof(iscsi_trace_dump_t)) {
            printk("iSCSI: trace dump ioctl size is %u, but caller uses %u\n",
                   sizeof(iscsi_trace_dump_t), dump.dump_ioctl_size);
            return -EINVAL;
        }

        if (dump.dump_version != TRACE_DUMP_VERSION) {
            printk("iSCSI: trace dump version is %u, but caller uses %u\n",
                   TRACE_DUMP_VERSION, dump.dump_version);
            return -EINVAL;
        }

        if (dump.trace_entry_size != sizeof(iscsi_trace_entry_t)) {
            printk("iSCSI: trace dump ioctl size is %u, but caller uses %u\n",
                   sizeof(iscsi_trace_dump_t), dump.dump_ioctl_size);
            return -EINVAL;
        }

        if (dump.num_entries < ISCSI_TRACE_COUNT) {
            /* tell the caller to use a bigger buffer */
            dump.num_entries = ISCSI_TRACE_COUNT;
            if (copy_to_user(user_dump, &dump, sizeof(dump)))
                return -EFAULT;
            else
                return -E2BIG;
        }

        /* the caller is responsible for zeroing the buffer before the ioctl, so
         * if the caller asks for too many entries, it should be able to tell which
         * ones actually have data.
         */

        /* only send what we've got */
        dump.num_entries = ISCSI_TRACE_COUNT;
        if (copy_to_user(user_dump, &dump, sizeof(dump))) {
            printk("iSCSI: trace copy_to_user %p, %p, %u failed\n", 
                   user_dump, &dump, sizeof(dump));
            return -EFAULT;
        }
        
        SPIN_LOCK_NOQUEUE(&iscsi_trace_lock);
        /* FIXME: copy_to_user may sleep, but we're holding a spin_lock with interrupts off */
        if (copy_to_user(user_dump->trace, &trace_table[0], dump.num_entries * sizeof(iscsi_trace_entry_t))) {
            printk("iSCSI: trace copy_to_user %p, %p, %u failed\n", 
                   user_dump->trace, &trace_table[0], dump.num_entries);
            SPIN_UNLOCK_NOQUEUE(&iscsi_trace_lock);
            return -EFAULT;
        }
        rc = trace_index;
        printk("iSCSI: copied %d trace entries to %p at %lu\n", dump.num_entries, user_dump->trace, jiffies);
        SPIN_UNLOCK_NOQUEUE(&iscsi_trace_lock);

        return rc;
#else
        printk("iSCSI: iSCSI kernel module does not implement tracing\n");
        return -ENXIO;
#endif
    }

    return -EINVAL;
}


Scsi_Host_Template iscsi_driver_template = {
    next : NULL,
    module : NULL,
    proc_dir : NULL,
    proc_info : iscsi_proc_info,
    name : NULL,
    detect : iscsi_detect,
    release : iscsi_release,
    info : iscsi_info,
    ioctl : NULL,
    command : NULL,
    queuecommand : iscsi_queue,
    eh_strategy_handler : ISCSI_EH_STRATEGY,
    eh_abort_handler : iscsi_eh_abort,
    eh_device_reset_handler : iscsi_eh_device_reset,
    eh_bus_reset_handler : iscsi_eh_bus_reset,
    eh_host_reset_handler : iscsi_eh_host_reset,
    abort : NULL,
    reset : NULL,
    slave_attach : NULL,
    bios_param : iscsi_biosparam,
    this_id: -1,
    can_queue : ISCSI_MIN_CANQUEUE,
    sg_tablesize : ISCSI_MAX_SG,
    cmd_per_lun: ISCSI_CMDS_PER_LUN,
    present : 0,
    unchecked_isa_dma : 0,
    use_clustering : ENABLE_CLUSTERING,
#if ( LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0) )
    use_new_eh_code : 1,
#endif
    emulated : 1
};

#ifdef MODULE
EXPORT_NO_SYMBOLS;

static int INIT_MODIFIER iscsi_init( void )
{
    char *build_str = BUILD_STR;
    int ret = -ENODEV;

    DEBUG_INIT0("iSCSI: init module\n");

    if (build_str) {
        /* developer-built variant of a 4-digit internal release */
        printk("iSCSI: %d.%d.%d.%d%s variant (%s) built for Linux %s\niSCSI: %s\n",
               DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
               DRIVER_EXTRAVERSION, ISCSI_DATE, UTS_RELEASE,
               build_str);
    }
    else if (DRIVER_INTERNAL_VERSION > 0) {
        /* 4-digit release */ 
        printk("iSCSI: %d.%d.%d.%d%s (%s) built for Linux %s\n",
               DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
               DRIVER_EXTRAVERSION, ISCSI_DATE, UTS_RELEASE);
    }
    else {
        /* 3-digit release */
        printk("iSCSI: %d.%d.%d%s (%s) built for Linux %s\n",
               DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION, DRIVER_PATCH_VERSION,
               DRIVER_EXTRAVERSION, ISCSI_DATE, UTS_RELEASE);
    }

    /* log any module param settings the user has changed */
    if (translate_deferred_sense)
        printk("iSCSI: will translate deferred sense to current sense on command responses\n");

    if (force_tcq)
        printk("iSCSI: will force tagged command queueing for all devices\n");

    if (untagged_queue_depth != 1)
        printk("iSCSI: untagged queue depth %d\n", untagged_queue_depth);

    control_major = register_chrdev( 0, control_name, &control_fops );
    if ( control_major < 0 ) {
        printk("iSCSI: failed to register the control device\n");
        return control_major;
    }
    printk("iSCSI: control device major number %d\n", control_major);

    iscsi_driver_template.module = THIS_MODULE;
    REGISTER_SCSI_HOST(&iscsi_driver_template);
    if (iscsi_driver_template.present ) {
        ret = 0;
    }
    else {
        DEBUG_ERR0("iSCSI: failed to register SCSI HBA driver\n");
        UNREGISTER_SCSI_HOST(&iscsi_driver_template);
    }

    set_bit(0, &init_module_complete);
    return ret;
}

static void EXIT_MODIFIER iscsi_cleanup( void )
{
    pid_t pid = 0;
    int rc;

    DEBUG_INIT0("iSCSI: cleanup module\n");
    if ( control_major > 0 ) {
        rc = unregister_chrdev( control_major, control_name );
        if ( rc ) {
            DEBUG_ERR0("iSCSI: error trying to unregister control device\n");
        } 
        else {
            control_major = 0;
        }
    }

    if ( iscsi_driver_template.present ) {
        DEBUG_INIT0("iSCSI: SCSI template present\n");
        /* this will cause the SCSI layer to call our iscsi_release function */
        UNREGISTER_SCSI_HOST(&iscsi_driver_template);
        iscsi_driver_template.present = 0;
    }
    
    /* kill the timer */
    if ((pid = iscsi_timer_pid))
        kill_proc(pid, SIGKILL, 1);

    /* wait for the timer to exit */
    while (test_bit(0, &iscsi_timer_running)) {
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(MSECS_TO_JIFFIES(10));
    } 

    DEBUG_INIT0("iSCSI: cleanup module complete\n");
    return;
}

module_init(iscsi_init);
module_exit(iscsi_cleanup);

#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4 
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
